Package: motifTestR
Title: Perform key tests for binding motifs in sequence data
Version: 1.7.0
Authors@R: person("Stevie", "Pederson", 
  email = "stephen.pederson.au@gmail.com", 
  role = c("aut", "cre"),
  comment = c(ORCID = "0000-0001-8197-3303")
  )
Description: Taking a set of sequence motifs as PWMs, test a set of sequences
  for over-representation of these motifs, as well as any positional features
  within the set of motifs. Enrichment analysis can be undertaken using multiple
  statistical approaches. The package also contains core functions to prepare 
  data for analysis, and to visualise results.
License: GPL-3
Encoding: UTF-8
URL: https://github.com/smped/motifTestR
BugReports: https://github.com/smped/motifTestR/issues
Depends: Biostrings, GenomicRanges, ggplot2 (>= 4.0.0), R (>= 4.5.0),
Imports: Seqinfo, graphics, harmonicmeanp, IRanges, matrixStats,
        methods, parallel, patchwork, rlang, S4Vectors, stats,
        universalmotif,
Suggests: AnnotationHub, BiocStyle, BSgenome.Hsapiens.UCSC.hg19,
        extraChIPs (>= 1.13.3), ggdendro, knitr, MASS, MotifDb,
        rmarkdown, rtracklayer, SimpleUpset, testthat (>= 3.0.0), VGAM
biocViews: MotifAnnotation, ChIPSeq, ChipOnChip, SequenceMatching,
        Software
LazyData: false
RoxygenNote: 7.3.3
Roxygen: list(markdown = TRUE)
Config/testthat/edition: 3
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/motifTestR
git_branch: devel
git_last_commit: 26a47a7
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-10-31
NeedsCompilation: no
Packaged: 2025-10-31 23:00:09 UTC; biocbuild
Author: Stevie Pederson [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-8197-3303>)
Maintainer: Stevie Pederson <stephen.pederson.au@gmail.com>

# Generated by roxygen2: do not edit by hand

export(clusterMotifs)
export(countClusterMatches)
export(countPwmMatches)
export(getClusterMatches)
export(getPwmMatches)
export(makeRMRanges)
export(plotMatchPos)
export(simMultiMotifs)
export(simSeq)
export(testClusterEnrich)
export(testClusterPos)
export(testMotifEnrich)
export(testMotifPos)
exportMethods(makeRMRanges)
import(Biostrings)
import(GenomicRanges)
import(ggplot2)
importClassesFrom(universalmotif,universalmotif)
importFrom(IRanges,DataFrameList)
importFrom(IRanges,IRanges)
importFrom(IRanges,IRangesList)
importFrom(IRanges,IntegerList)
importFrom(IRanges,Views)
importFrom(IRanges,findOverlaps)
importFrom(IRanges,overlapsAny)
importFrom(IRanges,width)
importFrom(S4Vectors,"mcols<-")
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,endoapply)
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,queryHits)
importFrom(S4Vectors,splitAsList)
importFrom(S4Vectors,subjectHits)
importFrom(Seqinfo,seqinfo)
importFrom(graphics,abline)
importFrom(harmonicmeanp,p.hmp)
importFrom(matrixStats,colSds)
importFrom(matrixStats,rowSums2)
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,slot)
importFrom(parallel,mclapply)
importFrom(patchwork,plot_layout)
importFrom(rlang,"!!")
importFrom(rlang,sym)
importFrom(stats,as.dendrogram)
importFrom(stats,as.dist)
importFrom(stats,as.formula)
importFrom(stats,binom.test)
importFrom(stats,cutree)
importFrom(stats,dist)
importFrom(stats,glm)
importFrom(stats,hclust)
importFrom(stats,p.adjust)
importFrom(stats,pchisq)
importFrom(stats,phyper)
importFrom(stats,poisson)
importFrom(stats,poisson.test)
importFrom(stats,quasipoisson)
importFrom(stats,setNames)
importFrom(stats,xtabs)
importFrom(universalmotif,compare_motifs)
importFrom(universalmotif,convert_type)
importFrom(universalmotif,create_motif)

#' @title Count the matches to a PWM within an XStringSet
#'
#' @description Count the matches to a PWM within an XStringSet
#'
#' @details
#' Will simply count the matches within an XStringSet and return an integer.
#' All matches are included.
#'
#' @return An integer vector
#'
#' @param pwm A Position Weight Matrix
#' @param stringset An XStringSet
#' @param rc logical(1) Also find matches using the reverse complement of pwm
#' @param min_score The minimum score to return a match
#' @param mc.cores Passed to \link[parallel]{mclapply} when analysing a list of
#' PWMs
#' @param ... Passed to \link[Biostrings]{countPWM}
#'
#' @examples
#' ## Load the example PWM
#' data("ex_pfm")
#' esr1 <- ex_pfm$ESR1
#'
#' ## Load the example Peaks
#' data("ar_er_seq")
#' countPwmMatches(esr1, ar_er_seq)
#'
#' ## Count all PWMs
#' countPwmMatches(ex_pfm, ar_er_seq)
#'
#' @importFrom parallel mclapply
#' @importFrom matrixStats rowSums2
#' @import Biostrings
#' @export
countPwmMatches <- function(
        pwm, stringset, rc = TRUE, min_score = "80%", mc.cores = 1, ...
) {

    if (!is.list(pwm)) pwm <- list(pwm)
    mc.cores <- min(mc.cores, length(pwm))
    pwm <- mclapply(pwm, .checkPWM, mc.cores = mc.cores)
    pwm <- .cleanMotifList(pwm)
    nm <- names(pwm)
    ## Append any reverse matrices to the existing list
    if (rc) pwm <- c(pwm, lapply(pwm, reverseComplement))

    # Form the entire XStringSetList into a Views object
    map <- .viewMapFromXStringset(stringset)
    views <- Views(
        unlist(stringset), start = map$start, width = map$width,
        names = map$names
    )

    ## Now the counts
    counts <- mclapply(
        pwm,
        \(x) countPWM(x, views, min.score = min_score, ...),
        mc.cores = mc.cores
    )
    out <- unlist(counts)
    if (rc) out <- rowSums2(matrix(out, ncol = 2, dimnames = list(nm, NULL)))
    names(out) <- nm
    out

}

test_that("countPwmMatches behaves as expected",{
    expect_true(countPwmMatches(esr1, stringset) == 1)
    expect_true(countPwmMatches(esr1, stringset, rc = FALSE) == 0)
})

test_that("countPwmMatches accepts a list", {
    n <- countPwmMatches(ex_pfm, stringset)
    expect_true(length(n) == 5)
    expect_equal(names(n), names(ex_pfm))
})

#' Script used to create example motifs
#' All taken from the coreA set of HOCOMOCOv11
library(MotifDb)
ex_pfm <- MotifDb |>
  subset(organism == "Hsapiens") |>
  query(andStrings = "HOCOMOCOv11-core-A", orStrings = c("ESR1", "ANDR", "FOXA1", "ZNF")) |>
  as.list()
names(ex_pfm) <- gsub(".+-core-A-([A-Z0-9]+)_HUMAN.+", "\\1", names(ex_pfm))

save(ex_pfm, file = "data/ex_pfm.RData")

#' Test positional bias motifs within a cluster
#'
#' @description
#' Test positional bias for all motifs within a given cluster
#'
#' @details
#' This is a reimplementation of \link{testMotifPos} for sets of motifs which
#' have been clustered for similarity.
#' The positions test the bias of any motifs within the cluster given that
#' overlapping matches are only counted once, and with the match retained being
#' the one with the highest relative score.
#'
#' It should also be noted that some motif clusters will contain PWMs of
#' varying length. When finding positional bias, the widest motif is taken as
#' the width for all, and any matches from narrower motifs outside of the range
#' allowed by wider motifs are discarded. This reduction in signal will make a
#' small difference in the outer bins, but is not considered to be problematic
#' for the larger analysis.
#'
#' @return
#' A data.frame with columns `start`, `end`, `centre`, `width`, `total_matches`,
#' `matches_in_region`, `expected`, `enrichment`, `prop_total`, `p`
#' and `consensus_motif`
#' The total matches represent the total number of matches within the set of
#' sequences, whilst the number observed in the final region are also given,
#' along with the proportion of the total this represents.
#' Enrichment is simply the ratio of observed to expected based on the
#' expectation of the null hypothesis
#'
#' The consensus motif across all matches is returned as a Position Frequency
#' Matrix (PFM) using \link[Biostrings]{consensusMatrix}.
#'
#' @param x A Position Weight Matrix, universalmotif object or list thereof.
#' Alternatively can be a single DataFrame or list of DataFrames as returned
#' by \link{getClusterMatches} with `best_only = TRUE`
#' @param stringset An XStringSet. Not required if matches are supplied as x
#' @param binwidth Width of bins across the range to group data into
#' @param abs Use absolute positions around zero to find symmetrical enrichment
#' @param rc logical(1) Also find matches using the reverse complement of each
#' PWM
#' @param min_score The minimum score to return a match
#' @param break_ties Choose how to resolve matches with tied scores
#' @param alt Alternative hypothesis for the binomial test
#' @param sort_by Column to sort results by
#' @param mc.cores Passed to \link[parallel]{mclapply}
#' @param ... Passed to \link[Biostrings]{matchPWM}
#'
#' @examples
#' ## Load the example PWM
#' data("ex_pfm")
#' ## Load the example sequences
#' data("ar_er_seq")
#'
#' ## Cluster the motifs
#' cl <- list(A = ex_pfm[1], B = ex_pfm[2:3])
#'
#' ## Get the best match and use this data
#' matches <- getClusterMatches(cl, ar_er_seq, best_only = TRUE)
#' ## Test for enrichment in any position
#' testClusterPos(matches)
#'
#' ## Or just pass the clustered matrices
#' ## Here we've set abs = TRUE to test absolute distance from the centre
#' testClusterPos(cl, ar_er_seq, abs = TRUE, binwidth = 10)
#'
#' @importFrom parallel mclapply
#' @importFrom stats p.adjust
#' @export
testClusterPos <- function(
        x, stringset, binwidth = 10, abs = FALSE, rc = TRUE, min_score = "80%",
        break_ties = "all", alt = c("greater", "less", "two.sided"),
        sort_by = c("p", "none"), mc.cores = 1,
        ...
) {

    cols <- c(
        "start", "end", "centre", "width", "total_matches", "matches_in_region",
        "expected", "enrichment", "prop_total", "p", "fdr", "consensus_motif"
    )
    alt <- match.arg(alt)
    sort_by <- match.arg(sort_by)

    if (is(x, "DataFrame")) x <- list(x)
    is_DF <- vapply(x, is, logical(1), "DataFrame")
    matches <- NULL
    if (all(is_DF)) {
        ## If we have a list of best matches
        check <- .checkMatches(x)
        matches <- x
    } else {
        ## Otherwise, obtain the list of best match DFs
        is_list <- vapply(x, is, logical(1), "list")
        if (!all(is_list)) stop("All clusters should contain lists of motifs")
        clust_lens <- vapply(x, length, integer(1))
        if (all(clust_lens == 1)) stop("All clusters cannot contain a single motif")
        cl <- lapply(x, .cleanMotifList)
        matches <- getClusterMatches(
            cl, stringset, rc, min_score, TRUE, break_ties, mc.cores, ...
        )
    }
    if (is.null(matches)) stop("Couldn't determine structure of input")

    out <- mclapply(
        matches, .testSingleClusterPos, binwidth = binwidth, abs = abs, rc = rc,
        min_score = min_score, break_ties = break_ties, alt = alt, ...,
        mc.cores = mc.cores
    )
    out <- do.call("rbind", out)
    out$fdr <- p.adjust(out$p, "fdr")
    o <- seq_len(nrow(out))
    if (sort_by != "none") o <- order(out[[sort_by]])
    out[o, cols]

}

#' @keywords internal
.testSingleClusterPos <- function(
        matches, binwidth, abs, rc, min_score, break_ties, alt, ...
) {

    ## This differs from the single motif case in that different width motifs
    ## may exist within a cluster. As a result, the boundary points need to be
    ## recalculated using the widest motif and all matches outside these
    ## boundary points will be excluded. This small loss of information
    ## shouldn't be profoundly deleterious to the process

    ## Setup a well formed output object for easy combining with other tests
    out_cols <- c(
        "start", "end", "centre", "width", "total_matches", "matches_in_region",
        "expected", "enrichment", "prop_total", "p", "consensus_motif"
    )
    out <- lapply(out_cols, \(x) integer())
    out$consensus_motif <- list()
    names(out) <- out_cols
    if (!nrow(matches)) return(data.frame(out))

    ## Remove those where the centre is outside of the range
    max_width <- max(width(matches$match))
    seq_width <- min(matches$seq_width)
    lim <- (seq_width - max_width) / 2
    matches <- matches[abs(matches$from_centre) <= lim, ]

    .testSingleMotifPos(
        matches, binwidth, abs, rc, min_score, break_ties, alt, ...
    )
}

## Convert ranges to DNAStringSets
test_set <- ar_er_seq
bg_set <- getSeq(genome, bg_ranges)
mcols(bg_set) <- mcols(bg_ranges)

cl <- list(A = ex_pfm[1], B = ex_pfm[2:3], C = ex_pfm[4:5])

test_that("Basic Poisson analysis works", {
    res <- testClusterEnrich(cl, test_set, bg_set, model = "poisson")
    expect_true(is(res, "data.frame"))
    expect_true(nrow(res) == 3)
    expect_true("est_bg_rate" %in% colnames(res))
})


test_that("Iteration works", {
    mcols(bg_set) <- mcols(bg_ranges)
    iter <- testClusterEnrich(cl, test_set, bg_set, model = "iter")
    expect_true(is(iter, "data.frame"))
    expect_true(nrow(iter) == 3)
    expect_true("iter_p" %in% colnames(iter))
    expect_true(all(iter$n_iter == 10))

})

test_that("Quasipoisson works", {
    mcols(bg_set) <- mcols(bg_ranges)
    iter <- testClusterEnrich(cl, test_set, bg_set, model = "quasi")
    expect_true(is(iter, "data.frame"))
    expect_true(nrow(iter) == 3)
    expect_true(all(iter$n_iter == 10))

})

test_that("GLM-Poisson works", {
    mcols(bg_set) <- mcols(bg_ranges)
    iter <- testClusterEnrich(cl, test_set, bg_set, model = "glm")
    expect_true(is(iter, "data.frame"))
    expect_true(nrow(iter) == 3)
    expect_true(all(iter$n_iter == 10))

})

test_that("Hypergeometric works", {
    mcols(bg_set) <- mcols(bg_ranges)
    hg <- testClusterEnrich(cl, test_set, bg_set, model = "hyper")
    expect_true(is(hg, "data.frame"))
    expect_true(nrow(hg) == 3)

})

test_that("clusterMotifPos also works as expected", {
    res <- testClusterPos(cl, test_set)
    expect_equal(nrow(res), length(cl))
    expect_equal(nrow(testClusterPos(cl, test_set[0])), 0)
    matches <- getClusterMatches(cl, ar_er_seq, best_only = TRUE)
    res <- testClusterPos(matches, abs = TRUE)
    expect_equal(nrow(res), length(cl))
    expect_true(all(res$start >= 0))

    expect_equal(nrow(testClusterPos(matches[[1]])), 1L)
    expect_error(testClusterPos(ex_pfm, ar_er_seq), "All clusters should.+")
    expect_error(testClusterPos(cl[1], ar_er_seq), "All clusters cannot.+")
})

#' @title Test enrichment across a cluster of motifs using a background set of sequences
#'
#' @description
#' Test for enrichment of any motif within a cluster across a set of sequences
#' using a background set to
#' derive a NULL hypothesis
#'
#' @details
#' This extends the analytic methods offered by \link{testMotifEnrich} using
#' PWMs grouped into a set of clusters.
#' As with all cluster-level approaches, hits from multiple PWMs which overlap
#' are counted as a single hit ensuring that duplicated matches are not
#' double-counted, and that only individual positions within the sequences are.
#'
#'
#' @return
#' See \link{testMotifEnrich}
#'
#' @param cl  A list of Position Weight Matrices, universalmotifs, with each
#' element representing clusters of related matrices
#' @param stringset An XStringSet with equal sequence widths
#' @param bg An XStringSet with the same sequence widths as the test XStringset
#' @param model The model used for analysis
#' @param var A column in the mcols element of bg, usually denoting an iteration
#' number
#' @param sort_by Column to sort results by
#' @param mc.cores Passed to \link[parallel]{mclapply}
#' @param prior.count Added to all counts to better manage zero counts in
#' background sequences. For analysis under QuasiPoisson models prior counts
#' are added as Poisson noise using this value as expected counts
#' @param seed Used for reproducibility when adding Poisson noise
#' @param ... Passed to \link{getPwmMatches} or \link{countPwmMatches}
#'
#' @seealso [makeRMRanges()], [getClusterMatches()], [countClusterMatches()], [testMotifEnrich()]
#'
#' @examples
#' ## Load the example peaks & the sequences
#' data("ar_er_peaks")
#' data("ar_er_seq")
#' sq <- seqinfo(ar_er_peaks)
#' ## Now sample size-matched ranges 10 times larger. In real-world analyses,
#' ## this set should be sampled as at least 1000x larger, ensuring features
#' ## are matched to your requirements. This example masks regions with known N
#' ## content, including centromeres & telomeres
#' data("hg19_mask")
#' set.seed(305)
#' bg_ranges <- makeRMRanges(
#'   ar_er_peaks, GRanges(sq)[1], exclude = hg19_mask, n_iter = 10
#' )
#'
#' ## Convert ranges to DNAStringSets
#' library(BSgenome.Hsapiens.UCSC.hg19)
#' genome <- BSgenome.Hsapiens.UCSC.hg19
#' bg_seq <- getSeq(genome, bg_ranges)
#'
#' ## Test for enrichment of clustered motifs
#' data("ex_pfm")
#' cl <- list(A = ex_pfm[1], B = ex_pfm[2:3])
#' testClusterEnrich(cl, ar_er_seq, bg_seq, model = "poisson")
#'
#'
#' @importFrom stats p.adjust
#' @export
testClusterEnrich <- function(
        cl, stringset, bg, var = "iteration",
        model = c("poisson", "hypergeometric", "quasipoisson", "glm_poisson", "iteration"),
        sort_by = c("p", "none"), mc.cores = 1, prior.count = 1, seed = 100, ...
) {

    ## Checks
    stopifnot(is(bg, "XStringSet"))
    model <- match.arg(model)
    args <- c(as.list(environment()), list(...))
    prior.count <- prior.count[[1]]
    stopifnot(prior.count >= 0)

    ## Prepare the output
    cols <- c("sequences", "matches", "expected", "enrichment", "Z", "p", "fdr")
    mod_cols <- list(
        poisson = "est_bg_rate", iteration = c("iter_p", "n_iter", "sd_bg"),
        hypergeometric = "odds_ratio", quasipoisson = c("n_iter", "sd_bg"),
        glm_poisson = c("n_iter", "sd_bg")
    )
    cols <- c(cols, mod_cols[[model]])
    if (model == "hypergeometric") cols <- setdiff(cols, "Z")
    ## Run the analysis
    cl <- lapply(cl, .cleanMotifList)
    if (model == "poisson")
        out <- .testPois(
            cl, stringset, bg, mc.cores, "cluster", prior.count, ...
        )
    if (model == "iteration")
        out <- .testIter(
            cl, stringset, bg, var, mc.cores, "cluster", prior.count, ...
        )
    if (model == "quasipoisson")
        out <- .testQuasi(
            cl, stringset, bg, var, mc.cores, "cluster", prior.count, seed, ...
        )
    if (model == "glm_poisson")
        out <- .testGlmPois(
            cl, stringset, bg, var, mc.cores, "cluster", prior.count, seed, ...
        )
    if (model == "hypergeometric")
        out <- .testHyper(
            cl, stringset, bg, mc.cores, "cluster", prior.count, ...
        )

    out$fdr <- p.adjust(out$p, "fdr")
    o <- seq_len(nrow(out))
    sort_by <- match.arg(sort_by)
    if (sort_by != "none") o <- order(out[[sort_by]])
    out[o,cols]

}

#' Script for creating example peaks.
#'
#' These represent peaks where ER, AR are found along with H3K27ac under DHT treatment


library(rtracklayer)
library(extraChIPs)

sq <- defineSeqinfo(build = "GRCh37")
genome(sq) <- "hg19"

## ER
er_url <- "https://ftp.ncbi.nlm.nih.gov/geo/samples/GSM3511nnn/GSM3511085/suppl/GSM3511085%5FER%5Fpeaks%5FED.bed.gz"
er <- import.bed(er_url)
seqlevels(er) <- seqnames(sq)
seqinfo(er) <- sq
er <- sort(er)
er

## AR
ar_url <- "https://ftp.ncbi.nlm.nih.gov/geo/samples/GSM3511nnn/GSM3511083/suppl/GSM3511083%5FAR%5Fpeaks%5FED.bed.gz"
ar <- import.bed(ar_url)
seqlevels(ar) <- seqnames(sq)
seqinfo(ar) <- sq
ar <- sort(ar)
ar

## H3K27ac
h3_url <- "https://ftp.ncbi.nlm.nih.gov/geo/samples/GSM3511nnn/GSM3511087/suppl/GSM3511087%5FH3K27ac%5Fpeaks%5FED.bed.gz"
h3 <- import.bed(h3_url)
seqlevels(h3) <- seqnames(sq)
seqinfo(h3) <- sq
h3 <- sort(h3)
h3

ar_er_peaks <- GRangesList(
  # ER = er, H3K27ac = h3
  AR = ar, ER = er, H3K27ac = h3
  # AR = ar, ER = er
) |>
  makeConsensus(p = 2/3, method = "coverage", min_width = 200) |>
  # makeConsensus(method = "coverage", min_width = 200) |>
  subset(seqnames == "chr1") |>
  resize(width = 400, fix = 'center') |>
  granges()
# ar_er_peaks

save(ar_er_peaks, file = "data/ar_er_peaks.RData")

library(BSgenome.Hsapiens.UCSC.hg19)
ar_er_seq <- getSeq(BSgenome.Hsapiens.UCSC.hg19, ar_er_peaks)
save(ar_er_seq, file = "data/ar_er_seq.RData")

#' @title Simulate sequences with multiple motifs
#'
#' @description
#' Simulate a set of sequences incorporating multiple motifs
#'
#' @details
#' Simulate a set of sequences with multiple motifs inserted using different
#' rates and distributions, as specified.
#' All shape, rate and theta parameters are recycled to match the length of the
#' supplied motif list, and can be supplied as vectors to tailor these
#' parameters to each provided element of the list of matrices
#'
#' @return
#' A DNAStringSet with mcols denoting the positions of all inserted motifs
#'
#' @param n The number of sequences to simulate
#' @param width Width of sequences to simulate
#' @param pfm List of Probability Weight/Frequency Matrices
#' @param bg Optional, pre-defined set of background sequences. Can be passed as
#' an XStringSet or character vector. All sequences must be the same width
#' @param nt Nucleotides to include
#' @param prob Sampling probabilities for each nucleotide
#' @param shape1,shape2 Passed to \link[VGAM]{rbetabinom.ab}
#' @param rate The expected rate of motifs per sequence. Is equivalent to
#' \eqn{ \lambda } in \link[stats]{rpois}. If set to NULL or NA, all sequences will
#' be simulated with a single motif, otherwise a Poisson distribution will be used
#' @param theta Overdispersion parameter passed to \link[MASS]{rnegbin}.
#' If set to NULL or NA the rate parameter will be passed to \link[stats]{rpois}.
#' However if this value is set, the rate and theta parameters are passed to
#' \link[MASS]{rnegbin} to simulate overdispersed counts
#' @param as ObjectClass to return objects as. Defaults to DNAStringSet, but
#' other viable options may include 'character', 'CharacterList' or any
#' other class from which a character vector may be coerced.
#' @param ol When randomly simulated positions overlap, choose one either at
#' random, by the first occurring PFM in the list of PFMs, or by the last.
#' @param ... Not used
#'
#' @examples
#' data("ex_pfm")
#' ## Simulate sequences including both ESR1 and ANDR, but with
#' ## ESR1 being included at a higher rate
#' seq <- simMultiMotifs(10, 100, ex_pfm[1:2], rate = c(2, 1))
#' seq
#' ## The positions of the motifs are included in the mcols
#' mcols(seq)
#'
#'
#' @export
simMultiMotifs <- function(
    n, width, pfm = NULL, bg = NULL, nt = c("A", "C", "G", "T"),
    prob = rep(0.25, 4), shape1 = 1, shape2 = shape1, rate = NA, theta = NA,
    as = "DNAStringSet", ol = c("random", "first", "last"), ...
){

  args <- c(as.list(environment()), list(...))
  args$bg <- NULL
  ## Pass to simSeq if there are no motifs
  if (is.null(pfm)) return(do.call(motifTestR::simSeq, args))
  prob <- rep_len(prob, length(nt))

  ## Setup the original (BG) set of sequences to be character(1)
  bg <- .defineBG(bg, n, width, nt, prob)
  seq_starts <- 1 + c(0, cumsum(nchar(bg)))[seq_along(bg)]
  n <- length(seq_starts)
  bg <- strsplit(paste(bg, collapse = ""), "")[[1]]
  width <- as.integer(length(bg) / n)

  ## Check the PFMs. Should be a list here! And all need to be PFMs not PWMs
  pfm <- .checkPfmList(pfm)

  ## Now expand the relevant params to match the nbr of motifs
  n_motifs <- length(pfm)
  shape1 <- rep_len(shape1, n_motifs)
  shape2 <- rep_len(shape2, n_motifs)
  if (!is.null(rate)) rate <- rep_len(rate, n_motifs)
  if (!is.null(theta)) theta <- rep_len(theta, n_motifs)
  pfm_width <- vapply(pfm, ncol, integer(1))
  max_start <- width - pfm_width
  stopifnot(all(max_start > 0))

  ## From here we step through each motif, or do we lapply the positions
  ## and sampled motifs and inject all at once? Not sure how to handle
  ## overlaps either
  pos_list <- mapply(
    .samplePos, rate = rate, theta = theta,
    shape1 = shape1, shape2 = shape2, max_start = max_start,
    MoreArgs = list(n = n, seq_starts = seq_starts), SIMPLIFY = FALSE
  )
  ## We need to ensure pfm is a named list!!!
  vec_pos <- mapply(
    \(x, y) {
      vapply(x, \(i) seq(i, length.out = y), numeric(y)) |> as.integer()
    }, x = pos_list, y = pfm_width, SIMPLIFY = FALSE
  )
  ## Now sample the new nucleotides
  new_nuc <- mapply(
    \(x, y) {
      replicate(
        length(x), apply(y, MARGIN = 2, FUN = \(p) sample(nt, 1, prob = p))
      ) |> as.character()
    }, x = pos_list, y = pfm, SIMPLIFY = FALSE
  )
  index_df <- data.frame(i = unlist(vec_pos), nt = unlist(new_nuc))

  ## Handle overlapping positions
  ol <- match.arg(ol)
  if (ol == "random") # Randomise the order
    index_df <- index_df[sample.int(nrow(index_df), nrow(index_df)),]
  if (ol == "last") # Reverse the order
    index_df <- index_df[seq(nrow(index_df), 1),]
  ## Keep the first of any duplicate rows using the above re-ordering
  index_df <- index_df[!duplicated(index_df[["i"]]),]

  ## Create the final set of sequences
  bg[index_df$i] <- index_df$nt
  seq <- apply(matrix(bg, ncol = n), MARGIN = 2, paste, collapse = "")
  seq <- as(seq, as)
  ## Now update the mcols if required
  if (is(seq, "Vector")) seq <- .addPosList(seq, pos_list, width, pfm)
  seq

}

#' @keywords internal
#' @importFrom IRanges IntegerList
#' @importFrom S4Vectors DataFrame endoapply
.addPosList <- function(seq, pos_list, width, pfm) {
  df_list <- lapply(
    pos_list,
    \(pos) {
      which_seq <- ceiling(pos / width)
      int_list <- IntegerList(vector("list", length(seq)))
      int_list[unique(which_seq)] <- split(pos %% width, f = which_seq)
      out <- endoapply(endoapply(int_list, unique), sort)
      n_motifs <- vapply(out, length, integer(1))
      if (all(n_motifs == 1)) out <- unlist(out)
      out
    }
  )
  nm <- names(pfm)
  if (is.null(nm)) nm <- paste0("X", seq_along(pfm))
  names(df_list) <- nm
  counts <- lapply(df_list, \(x) vapply(x, length, integer(1)))
  df_list$n_motifs <- rowSums(do.call("cbind", counts))
  DF <- DataFrame(df_list)
  mcols(seq) <- DF
  seq
}

#' @importFrom methods is
#' @importFrom IRanges width
.defineBG <- function(bg, n, wd, nt, prob) {
  ## Return a character vector which will be collapsed
  if (is.null(bg))
    bg <- simSeq(n, wd, nt = nt, prob = prob, as = "character")

  if (is(bg, "XStringSet")) {
    stopifnot(length(unique(width(bg))) == 1) # Fixed width check
    bg <- as.character(bg)
  }

  ## Must be a character beyond this point
  stopifnot(is.character(bg))

  ## Ensure that only the passed nucleotides are present
  pat <- paste0("[^", paste(nt, collapse = ""), "].+")
  has_non_nuc <- vapply(bg, \(x) grepl(pat, x), logical(1))
  stopifnot(all(!has_non_nuc))

  ## These need to be the same width for sampling motif positions
  w <- unique(nchar(bg))
  stopifnot(length(w) == 1)
  bg

}

#' @importFrom universalmotif create_motif
.checkPfmList <- function(pfm) {
  stopifnot(is.list(pfm))
  if (length(pfm) == 1) stop("Please use simSeq for a single motif")
  is_uni <- vapply(pfm, is, logical(1), "universalmotif")
  if (all(is_uni)) {
    nm <- vapply(pfm, slot, character(1), "name")
    pfm <- convert_type(pfm, "PPM")
    pfm <- lapply(pfm, slot, "motif")
    names(pfm) <- nm
  } else {
    col_sums <- lapply(pfm, \(x) unname(colSums(x)))
    is_pfm <- lapply(col_sums, \(x) all.equal(sum(x), length(x), tolerance = 1e-6))
    is_pfm <- unlist(is_pfm)
    if (is.character(is_pfm)) {
      ## Perform the conversion using universal motif
      nm <- names(pfm)
      pfm <- lapply(pfm, create_motif, type = "PPM")
      pfm <- lapply(pfm, slot, "motif")
      names(pfm) <- nm
    }
  }
  stopifnot(all(vapply(pfm, is.matrix, logical(1))))
  pfm
}

test_that(".checkPfmList functions as expected", {

    mat_list <- .checkPfmList(um_db)
    expect_true(all(vapply(mat_list, is.matrix, logical(1))))
    nm <- vapply(um_db, slot, character(1), "name")
    expect_equal(names(mat_list), nm)

    mat_list <- .checkPfmList(ex_pfm)
    expect_equal(names(mat_list), names(ex_pfm))

    ## Errors
    expect_error(.checkPfmList(um_db[[1]]))
    expect_error(.checkPfmList(um_db[1]), "Please use simSeq.+")
})

test_that(".defineBG functions as expected", {

    bg <- .defineBG(stringset, nt = c("A", "C", "G", "T"))
    expect_true(is.character(bg))

    bg <- .defineBG(NULL, 2, 5, c("A", "C"), c(0.5, 0.5))
    expect_true(is.character(bg))
    expect_true(length(bg) == 2)
    expect_true(all(vapply(bg, nchar, integer(1)) == 5))

    err_bg <- as(list(stringset[[1]], stringset[[1]][1:5]), "DNAStringSet")
    expect_error(.defineBG(err_bg, nt = c("A", "C", "G", "T")))

    expect_error(.defineBG(1))
    expect_error(defineBG("cat", 1, 5, nt = c("A", "C"), c(0.5, 0.5)))
})

test_that("Main function simulates data as expected", {

    pfm_list <- ex_pfm[1:2]
    ## Check character output
    seq <- simMultiMotifs(2, 100, pfm_list, as = "character")
    expect_true(is.character(seq))
    expect_true(sum(nchar(seq)) == 2 * 100)

    ## Check DNAStringSet output
    seq <- simMultiMotifs(2, 100, pfm_list, ol = "last")
    expect_true(is(seq, "DNAStringSet"))
    expect_true(sum(width(seq)) == 2 * 100)
    expect_equal(colnames(mcols(seq)), c(names(pfm_list), "n_motifs"))
    ## These should be simulated with 1 motif/sequence
    expect_true(all(vapply(mcols(seq), is.numeric, logical(1))))
    ## Re-simulate using the rate & more sequences
    set.seed(10)
    seq <- simMultiMotifs(10, 100, pfm_list, rate = c(2, 0.5))
    mc <- mcols(seq)[names(pfm_list)]
    expect_true(all(vapply(mc, \(x) is(x, "IntegerList"), logical(1))))
    counts <- lapply(mc, \(x) vapply(x, length, integer(1)))
    expect_true(mean(counts[[1]]) > mean(counts[[2]]))


})

test_that("Differences between PFM & PWMs are managed", {
    pwm_list <- um_db |>
        universalmotif::convert_type("PWM") |>
        lapply(slot, "motif")
    names(pwm_list) <- vapply(um_db, slot, character(1), "name")
    chk <- .checkPfmList(pwm_list)
    expect_true(all(names(chk) == names(pwm_list)))
    expect_true(all(vapply(chk, is.matrix, logical(1))))

})

#' Assign each motif to a cluster
#'
#' Cluster related motifs for testing as a group
#'
#' @details
#' This builds on \link[universalmotif]{compare_motifs}, enabling the
#' assignment of each PWM to a cluster, and subsequent testing of motifs as
#' a cluster, rather than returning individual results.
#'
#' Internally all matrices are converted to distance matrices and
#' \link[stats]{hclust} is used to form clusters. By default, options such as
#' "EUCL", "MAN" produce distances, whilst similarity matrices are produced
#' when choosing "PCC" and other correlation based metrics. In these cases, the
#' distance matrix is obtained by taking 1 - similarity.
#'
#' By default PWM labels are hidden (labels = FALSE), however these can be shown
#' using labels = NULL as explained in \link[stats]{plot.hclust}.
#'
#' Raising the threshold will lead to fewer, larger clusters whilst leaving
#' this value low will return a more conservative approach, with more smaller
#' clusters.
#' The final decision as the best clustering strategy is highly subjective and
#' left to the user.
#' Manual inspection of motifs within a cluster can be performed using
#' \link[universalmotif]{view_motifs}, as shown in the vignette.
#'
#' @return Named vector with numeric values representing which cluster each
#' motif has been assigned to.
#'
#' If setting return_d = TRUE, a named list will be returned with the clusters
#' as the element `cl` and a list with distance matrices for each cluster as the
#' element `d`
#'
#' @param motifs A list of universalmotifs or a list of PWMs
#' @param type Can be ICM or PPM
#' @param method The method to be used for determining similarity/distances
#' @param power Raise correlation matrices to this power before converting to a
#' distance matrix. Only applied if method is either "PCC" or "WPCC"
#' @param agglom Method to be used for agglomeration by \link[stats]{hclust}
#' @param thresh Tree heights below which motifs are formed into a cluster
#' @param return_d logical(1) Return the distance matrices for each cluster
#' @param plot Show tree produced by \link[stats]{hclust}. If requested the
#' value set by thresh will be shown as a horizontal line
#' @param labels,cex Passed to \link[stats]{plot.hclust}
#' @param linecol Passed to \link[graphics]{abline} as the argument `col`
#' @param ... passed to \link[universalmotif]{compare_motifs}
#'
#' @examples
#' # Load the example motifs
#' data("ex_pfm")
#'
#' # Return a vector with each motif assigned a cluster
#' # The default uses Pearson's Correlation Coefficient
#' clusterMotifs(ex_pfm)
#'
#' # Preview the settings noting that showing labels can clutter the plot
#' # with large numbers of motifs. The defaults for Euclidean distance
#' # show the threshold may need raising
#' clusterMotifs(ex_pfm, plot = TRUE, labels = NULL, method = "EUCL")
#'
#' @importFrom universalmotif compare_motifs create_motif
#' @importFrom stats as.dist hclust cutree
#' @importFrom graphics abline
#' @export
clusterMotifs <- function(
        motifs, type = c("PPM", "ICM"),
        method = c("PCC", "EUCL", "SW", "KL", "ALLR", "BHAT", "HELL", "SEUCL", "MAN", "ALLR_LL", "WEUCL", "WPCC"),
        power = 1, agglom = "complete", thresh = 0.2, return_d = FALSE,
        plot = FALSE, labels = FALSE, cex = 1, linecol = "red", ...
){
    # Convert to universal motif, if a list is passed
    if (all(vapply(motifs, is, logical(1), class2 = "matrix"))) {
        motifs <- lapply(
            names(motifs), \(x) create_motif(motifs[[x]], name = x)
        )
    }
    stopifnot(all(vapply(motifs, is, logical(1), class2 = "universalmotif")))
    method <- match.arg(method)
    type <- match.arg(type)
    if (type == "ICM" & method %in% c("ALLR", "ALLR_LL"))
        stop("Cannot use ICM with ALLR or ALLR_LL")
    is_dist <- method %in% c("EUCL", "KL", "HELL", "SEUCL", "MAN", "WEUCL")
    mat <- compare_motifs(motifs, use.type = type, method = method, ...)
    ## This is really only useful for correlations
    if (power != 1 & method %in% c("PCC", "WPCC")) mat <- mat^power
    ## Make a distance/dissimilarity matrix
    mat <- abs(mat) / max(abs(mat)) # Scale to be in [0,1]
    if (!is_dist) mat <- 1 - abs(mat)
    d <- as.dist(mat)
    cl <- hclust(d, method = agglom)
    if (plot) {
        plot(cl, labels = labels, cex = cex)
        abline(a = thresh, b = 0, col = linecol)
    }
    cl <- cutree(cl, h = thresh)
    if (return_d) {
        nm <- lapply(split(cl, cl), names)
        d_split <- lapply(nm, \(x) as.matrix(d)[x, x])
        return(list(cl = cl, d = d_split))
    }
    cl
}

test_that("clusterMotifs works as expected", {
    cl <- clusterMotifs(ex_pfm, power = 2, plot = TRUE)
    expect_true(is(cl, "integer"))
    expect_equal(names(cl), names(ex_pfm))
    expect_error(
        clusterMotifs(ex_pfm, type = "ICM", method = "ALLR"),
        "Cannot use ICM with ALLR or ALLR_LL"
    )
    expect_error(clusterMotifs(c("cat", "dog")))
})

test_that("returning the distance matrix works", {
    cl <- clusterMotifs(ex_pfm, return_d = TRUE)
    expect_equal(c("cl", "d"), names(cl))
    expect_equal(max(cl$cl), length(cl$d))
})

## ----setup, echo = FALSE------------------------------------------------------
knitr::opts_chunk$set(message = FALSE, crop = NULL)

## ----load-packages------------------------------------------------------------
library(motifTestR)
library(rtracklayer)
library(BSgenome.Hsapiens.UCSC.hg19)
library(parallel)
library(ggplot2)
library(patchwork)
library(universalmotif)
library(extraChIPs)
theme_set(theme_bw())
cores <- 1

## ----load-ar-er-peaks, echo = FALSE-------------------------------------------
data("ar_er_peaks")

## ----load-example-peaks-------------------------------------------------------
data("ar_er_peaks")
ar_er_peaks
sq <- seqinfo(ar_er_peaks)

## ----test-seq-----------------------------------------------------------------
test_seq <- getSeq(BSgenome.Hsapiens.UCSC.hg19, ar_er_peaks)
names(test_seq) <- as.character(ar_er_peaks)

## ----ex-pwm-------------------------------------------------------------------
data("ex_pfm")
names(ex_pfm)
ex_pfm$ESR1

## ----match-esr1---------------------------------------------------------------
score_thresh <- "70%"
getPwmMatches(ex_pfm$ESR1, test_seq, min_score = score_thresh)

## ----best-match-esr1----------------------------------------------------------
getPwmMatches(ex_pfm$ESR1, test_seq, min_score = score_thresh, best_only = TRUE)

## ----bm-all-------------------------------------------------------------------
bm_all <- getPwmMatches(
  ex_pfm, test_seq, min_score = score_thresh, best_only = TRUE, break_ties = "all",
  mc.cores = cores
)

## ----count-pwm-matches--------------------------------------------------------
countPwmMatches(ex_pfm, test_seq, min_score = score_thresh, mc.cores = cores)

## ----test-motif-pos-----------------------------------------------------------
res_pos <- testMotifPos(bm_all, mc.cores = cores)
head(res_pos)

## ----test-motif-pos-abs-------------------------------------------------------
res_abs <- testMotifPos(bm_all, abs = TRUE, mc.cores = cores) 
head(res_abs)

## ----plot-pos, fig.cap = "Distribution of motif matches around the centres of the set of peaks", fig.alt="Two panel plot with the left panel showing the probability distribution of motif matches for ESR1, ANDR and FOXA1 within the set of test sequences as coloured lines. The right panel shows the counts within each 10bp bin around the centre of the sequences with a loess curve overlaid"----
topMotifs <- res_pos |>
    subset(fdr < 0.05) |>
    rownames()
A <- plotMatchPos(bm_all[topMotifs], binwidth = 10, se = FALSE)
B <- plotMatchPos(
    bm_all[topMotifs], binwidth = 10, geom = "col", use_totals = TRUE
) +
  geom_smooth(se = FALSE, show.legend = FALSE) +
  facet_wrap(~name)
A + B + plot_annotation(tag_levels = "A") & theme(legend.position = "bottom")

## ----plot-abs-pos, fig.cap = "Distribution of motif matches shown as a distance from the centre of each sequence", fig.alt="Two panel plot with the left panel showing the distances from the sequence centres of the best matches for FOXA1, ANDR and ESR1. The right panel shows the CDF of these matches as a line plot"----
topMotifs <- res_abs |>
    subset(fdr < 0.05) |>
    rownames()
A <- plotMatchPos(bm_all[topMotifs], abs = TRUE, type = "heatmap") +
  scale_fill_viridis_c()
B <- plotMatchPos(
  bm_all[topMotifs], abs = TRUE, type = "cdf", geom = "line", binwidth = 5
)
A + B + plot_annotation(tag_levels = "A") & theme(legend.position = "bottom")

## ----zr75-enh-----------------------------------------------------------------
data("zr75_enh")
mean(overlapsAny(ar_er_peaks, zr75_enh))

## ----define-bg-ranges---------------------------------------------------------
ar_er_peaks$feature <- ifelse(
  overlapsAny(ar_er_peaks, zr75_enh), "enhancer", "other"
)
chr1 <- GRanges(sq)[1]
bg_ranges <- GRangesList(
  enhancer = zr75_enh,
  other = GenomicRanges::setdiff(chr1, zr75_enh)
)

## ----rm-ranges----------------------------------------------------------------
data("hg19_mask")
set.seed(305)
rm_ranges <- makeRMRanges(
  splitAsList(ar_er_peaks, ar_er_peaks$feature),
  bg_ranges, exclude = hg19_mask,
  n_iter = 100
)

## ----rm-seq-------------------------------------------------------------------
rm_seq <- getSeq(BSgenome.Hsapiens.UCSC.hg19, rm_ranges)
mcols(rm_seq) <- mcols(rm_ranges)

## ----enrich-res---------------------------------------------------------------
enrich_res <- testMotifEnrich(
  ex_pfm, test_seq, rm_seq, min_score = score_thresh, model = "quasi", mc.cores = cores
)
head(enrich_res)

## ----iter-res-----------------------------------------------------------------
iter_res <- testMotifEnrich(
  ex_pfm, test_seq, rm_seq, min_score = score_thresh, mc.cores = cores, model = "iteration"
)
head(iter_res)

## ----plot-overlaps, fig.cap = "Distribution of select PWM matches within sequences. Each sequence is only considered once and as such, match numbers may be below those returned during testing, which includes multiple matches within a sequence.", fig.alt = "UpSet plot showing the sequences which match one or more of the PWMs in the example dataset."----
ex_pfm |>
  getPwmMatches(test_seq, min_score = score_thresh, mc.cores = cores) |>
  lapply(\(x) x$seq) |>
  plotOverlaps(type = "upset")

## ----compare-andr-foxa1, fig.cap = "Binding Motifs for ANDR and FOXA1 overlaid for the best alignment, showing the similarity in the core regions.", fig.alt="Binding Motifs for ANDR and FOXA1 overlaid for the best alignment, showing the similarity in the core regions."----
c("ANDR", "FOXA1") |>
    lapply(
        \(x) create_motif(ex_pfm[[x]], name = x, type = "PPM")
    ) |>
    view_motifs()

## ----ex-cl, fig.cap = "Plot produced when requested by clusterMotifs showing the relationship between motifs. The red horizontal line indicates the threshold below which motifs are grouped together into a cluster. Clustering is performed using hclust on the distance matrices produced by `universalmotif::compare_motifs()`", fig.alt = "Dendrogram showing clusters formed by `clusterMotifs()` with default settings."----
cl <- clusterMotifs(ex_pfm, plot = TRUE, labels = NULL)
ex_cl <- split(ex_pfm, cl)
names(ex_cl) <- vapply(split(names(cl), cl), paste, character(1), collapse = "/")

## ----cl-matches---------------------------------------------------------------
cl_matches <- getClusterMatches(
    ex_cl, test_seq, min_score = score_thresh, best_only = TRUE
)
cl_matches

## ----test-cluster-pos---------------------------------------------------------
testClusterPos(cl_matches, test_seq, abs = TRUE)

## ----test-cluster-enrich------------------------------------------------------
testClusterEnrich(
  ex_cl, test_seq, rm_seq, min_score = score_thresh, model = "quasi", mc.cores = cores
)

## ----sessionInfo, echo=FALSE--------------------------------------------------
sessionInfo()

#' @title Test motif enrichment using a background set of sequences
#'
#' @description
#' Test for motif enrichment within a set of sequences using a background set to
#' derive a NULL hypothesis
#'
#' @details
#' This function offers four alternative models for assessing the enrichment of
#' a motif within a set of sequences, in comparison to a background set of
#' sequences.
#' Selection of the BG sequences plays an important role and, in conjunction
#' with the question being addressed, determines the most appropriate model to
#' use for testing, as described below.
#'
#' It should also be noted that the larger the BG set of sequences, the larger
#' the computational burden, and results can take far longer to return. For
#' many millions of background sequences, this may run beyond an hour
#'
#' ## Descriptions of Models and Use Cases
#'
#' ### Hypergeometric Tests
#'
#' Hypergeometric tests are best suited to the use case where the test set of
#' sequences represents a subset of a larger set, with a specific feature or
#' behaviour, whilst the BG set may be the remainder of the set without that
#' feature. For example, the test set may represent ChIP-Seq binding sites
#' where signal changes in response to treatment, whilst the BG set represents
#' the sites where no changed signal was observed. Testing is one-sided, for
#' enrichment of motifs within the test set.
#'
#' Due to these relatively smaller sized datasets, setting
#' model = "hypergeometric", will generally return results quickly
#'
#' ### Poisson Tests
#'
#' This approach requires a set of background sequences which should be much
#' larger than the test set of sequences.
#' The parameters for a Poisson model are estimated in a per-sequence manner on
#' the set of BG sequences, and the observed rate of motif-matches within the
#' test set is then tested using \link[stats]{poisson.test}.
#' Testing is two-sided.
#'
#' This approach assumes that all matches follow a Poisson distribution, which
#' is often true, but data can also be over-dispersed. Given that this model can
#' also return results relatively quickly, is it primarily suitable for data
#' exploration, such as quickly checking for expected behaviours, but not for
#' final results.
#'
#' ### Quasi-Poisson Test
#'
#' The quasipoisson model allows for over-dispersion and will return more
#' conservative results than using the standard Poisson model.
#' Under the method currently implemented here, BG sequences should be divided
#' into blocks (i.e. iterations), identical in size to the test set of sequences.
#' Model parameters are estimated per iteration across the BG set of sequences,
#' with the rate of matches in the test set being compared against these blocks.
#' This ensures more conservative results that if analysing test and bg
#' sequences as collections of individual sequences.
#'
#' It is expected that the BG set will matched for the features of interest and
#' chosen using \link{makeRMRanges} with a large number of iterations, e.g.
#' n_iter = 1000.
#' Due to this parameterisation, quasipoisson approaches can be computationally
#' time-consuming, as this is effectively an iterative approach.
#' Testing is two-sided.
#'
#' ### GLM-Poisson Test
#'
#' This follows the same approach as the Quasi-Poisson model, relying on fitting
#' iterations using [glm()].
#' For this model however, no over-dispersions are estimated and the underlying
#' family is simply the `poisson()` family
#'
#' ### Iteration
#'
#' Setting the model as "iteration" performs a non-parametric analysis, with
#' the exception of returning Z-scores under the Central Limit Theorem.
#' Mean and SD of matches is found for each iteration, and used to return Z
#' scores, with p-values returned from both a Z-test and from comparing
#' observed values directly to sampled values obtained from the BG sequences.
#' Sampled values are calculated directly and as such, are limited in precision.
#'
#' As for the QuasiPoisson model, a very large number of iterations is expected
#' to be used, to ensure the CLT holds, again making this a computationally
#' demanding test.
#' Each iteration/block is expected to be identically-sized to the test set,
#' and matched for any features as appropriate using [makeRMRanges()].
#'
#'
#'
#' @return
#' A data.frame with columns: `sequences`, `matches`, `expected`, `enrichment`,
#' and `p`, with additional columns `Z`, `est_bg_rate` (Poisson), `odds_ratio`
#' (Hypergeometric) or `Z`, `sd_bg`, `n_iter` and `iter_p` (Iterations).
#' The numbers of sequences and matches refer to the test set of sequences,
#' whilst expected is the expected number of matches under the Poisson or
#' iterative null distribution. The ratio of matches to expected is given as
#' `enrichment`, along with the Z score and p-value. Whilst the Z score is only
#' representative under the Poisson model, it is used to directly estimate
#' p-values under the iterative approach.
#' Under this latter approach, the sd of the matches found in the background
#' sequences is also given, along with the number of iterations and the p-values
#' from permutations testing the one-sided hypothesis hypothesis for enrichment.
#'
#' It may also be worth noting that if producing background sequences using
#' \link{makeRMRanges} with `replace = TRUE` and `force_ol = TRUE`, the
#' iterative model corresponds to a bootstrap, given that the test sequences
#' will overlap the background sequences and background ranges are able to be
#' sampled with replacement.
#'
#' @param pwm A Position Weight Matrix or list of PWMs
#' @param stringset An XStringSet with equal sequence widths
#' @param bg An XStringSet with the same sequence widths as the test XStringset
#' @param model The model used for analysis
#' @param var A column in the mcols element of bg, usually denoting an iteration
#' number
#' @param sort_by Column to sort results by
#' @param mc.cores Passed to \link[parallel]{mclapply}
#' @param prior.count Added to all counts to better manage zero counts in
#' background sequences. For analysis under QuasiPoisson models prior counts
#' are added as Poisson noise using this value as expected counts
#' @param seed Used for reproducibility when adding Poisson noise
#' @param ... Passed to \link{getPwmMatches} or \link{countPwmMatches}
#'
#' @seealso [makeRMRanges()], [getPwmMatches()], [countPwmMatches()]
#'
#' @examples
#' ## Load the example peaks & the sequences
#' data("ar_er_peaks")
#' data("ar_er_seq")
#' sq <- seqinfo(ar_er_peaks)
#' ## Now sample size-matched ranges 10 times larger. In real-world analyses,
#' ## this set should be sampled as at least 1000x larger, ensuring features
#' ## are matched to your requirements. This example masks regions with known N
#' ## content, including centromeres & telomeres
#' data("hg19_mask")
#' set.seed(305)
#' bg_ranges <- makeRMRanges(
#'   ar_er_peaks, GRanges(sq)[1], exclude = hg19_mask, n_iter = 10
#' )
#'
#' ## Convert ranges to DNAStringSets
#' library(BSgenome.Hsapiens.UCSC.hg19)
#' genome <- BSgenome.Hsapiens.UCSC.hg19
#' bg_seq <- getSeq(genome, bg_ranges)
#' mcols(bg_seq)$iteraton <-bg_ranges$iteration
#'
#' ## Test for enrichment of the ESR1 motif
#' data("ex_pfm")
#' esr1 <- ex_pfm$ESR1
#' testMotifEnrich(esr1, ar_er_seq, bg_seq, model = "poisson")
#'
#' ## Test all motifs
#' testMotifEnrich(ex_pfm, ar_er_seq, bg_seq, model = "poisson")
#'
#'
#' @importFrom stats p.adjust
#' @export
testMotifEnrich <- function(
        pwm, stringset, bg, var = "iteration",
        model = c("poisson", "hypergeometric", "quasipoisson", "glm_poisson", "iteration"),
        sort_by = c("p", "none"), mc.cores = 1, prior.count = 1, seed = 100, ...
) {

    ## Checks
    stopifnot(is(bg, "XStringSet"))
    model <- match.arg(model)
    args <- c(as.list(environment()), list(...))
    prior.count <- prior.count[[1]]
    stopifnot(prior.count >= 0)

    ## Prepare the output
    cols <- c("sequences", "matches", "expected", "enrichment", "Z", "p", "fdr")
    mod_cols <- list(
        poisson = "est_bg_rate", iteration = c("iter_p", "n_iter", "sd_bg"),
        hypergeometric = "odds_ratio", quasipoisson = c("n_iter", "sd_bg"),
        glm_poisson = c("n_iter", "sd_bg")
    )
    cols <- c(cols, mod_cols[[model]])
    if (model == "hypergeometric") cols <- setdiff(cols, "Z")
    ## Run the analysis
    if (is.matrix(pwm)) pwm <- list(pwm)
    pwm <- .cleanMotifList(pwm)
    if (model == "poisson")
        out <- .testPois(pwm, stringset, bg, mc.cores, "pwm", prior.count, ...)
    if (model == "iteration")
        out <- .testIter(pwm, stringset, bg, var, mc.cores, pc = prior.count, ...)
    if (model == "quasipoisson")
        out <- .testQuasi(
            pwm, stringset, bg, var, mc.cores, "pwm", prior.count, seed, ...
        )
    if (model == "glm_poisson")
        out <- .testGlmPois(
            pwm, stringset, bg, var, mc.cores, "pwm", prior.count, seed, ...
        )
    if (model == "hypergeometric")
        out <- .testHyper(pwm, stringset, bg, mc.cores, pc = prior.count, ...)

    out$fdr <- p.adjust(out$p, "fdr")
    o <- seq_len(nrow(out))
    sort_by <- match.arg(sort_by)
    if (sort_by != "none") o <- order(out[[sort_by]])
    out[o,cols]

}

#' @importFrom parallel mclapply
#' @importFrom stats phyper
.testHyper <- function(
        x, stringset, bg, mc.cores, type = c("pwm", "cluster"), pc, ...
){

    ## Check there's no overlap between the two by removing shared sequences
    cl <- class(stringset)
    ss <- as(setdiff(stringset, bg), cl)
    bg <- as(setdiff(bg, stringset), cl)
    if (length(ss) != length(stringset)) {
        msg <- paste(
            "Shared sequences found in the test and background set.",
            "These have been excluded from both"
        )
        message(msg)
        stopifnot(length(ss) > 0)
    }
    n_ss <- length(ss)
    n_bg <- length(bg)

    type <- match.arg(type)

    if (type == "pwm") {
        matches_ss <- mclapply(
            x, .hasPwmMatch, stringset = ss, ..., mc.cores = mc.cores
        )
        matches_bg <- mclapply(
            x, .hasPwmMatch, stringset = bg, ..., mc.cores = mc.cores
        )
    }
    if (type == "cluster") {
        matches_ss <- mclapply(
            x, .hasClusterMatch, stringset = ss, ..., mc.cores = mc.cores
        )
        matches_bg <- mclapply(
            x, .hasClusterMatch, stringset = bg, ..., mc.cores = mc.cores
        )
    }
    n_matches_ss <- vapply(matches_ss, sum, integer(1)) + pc
    n_matches_bg <- vapply(matches_bg, sum, integer(1)) + pc

    or_denom <- (n_ss - n_matches_ss) / (n_bg - n_matches_bg)
    or <- (n_matches_ss / n_matches_bg) / or_denom

    ## Perform a 1-sided hypergeometric test
    p_vals <- mclapply(
        seq_along(x),
        \(i) {
            q <- n_matches_ss[[i]]
            k <- q + n_matches_bg[[i]]
            phyper(q - 1, m = n_ss, n = n_bg, k = k, lower.tail = FALSE)
        },
        mc.cores = mc.cores
    )

    ## Return the output
    data.frame(
        sequences = n_ss, matches = n_matches_ss,
        expected = n_ss * n_matches_bg / n_bg,
        enrichment = (n_matches_ss / n_ss) / (n_matches_bg / n_bg),
        p = unlist(p_vals),
        odds_ratio = or
    )

}


#' @importFrom parallel mclapply
#' @importFrom stats glm quasipoisson
#' @importFrom matrixStats colSds
#' @keywords internal
.testQuasi <- function(
        x, stringset, bg, var, mc.cores, type = c("pwm", "cluster"), pc, seed,
        ...
) {

    ## Prior counts should be handled with caution here as using a fixed value
    ## will breach the Poisson distribution, e.g. adding a fixed value of 1 to
    ## a consistent run of zeros will give mean 1 with variance 0. Poisson
    ## noise may be preferable, but may profoundly reduce significance

    n <- length(stringset)
    stopifnot(var %in% colnames(mcols(bg)))
    splitbg <- split(bg, mcols(bg)[[var]])
    if (!all(vapply(splitbg, length, integer(1)) == n))
        stop("All iterations must be the same size as the test sequences")

    type <- match.arg(type)
    if (type == "pwm") {
        matches <- countPwmMatches(x, stringset, mc.cores = mc.cores, ...)
        bg_matches <- mclapply(
            splitbg, \(i) countPwmMatches(x, i, mc.cores = 1, ...),
            mc.cores = mc.cores
        )
    }
    if (type == "cluster") {
        matches <- countClusterMatches(x, stringset, mc.cores = mc.cores, ...)
        bg_matches <- mclapply(
            splitbg, \(i) countClusterMatches(x, i, mc.cores = 1, ...),
            mc.cores = mc.cores
        )
    }
    bg_mat <- do.call("rbind", bg_matches)
    ## Add Poisson prior counts
    if (pc != 0) {
        set.seed(seed)
        matches <- matches + stats::rpois(length(matches), pc)
        bg_mat <- bg_mat + stats::rpois(length(bg_mat), pc)
    }
    n_iter <- nrow(bg_mat)
    stopifnot(n_iter > 1)
    mean_bg <- colMeans(bg_mat)
    sd_bg <- colSds(bg_mat)
    Z <- (matches - mean_bg) / sd_bg
    Z[sd_bg == 0] <- NA_real_ # Handle where sd_bg == 0
    enrichment <- (matches) / colMeans(bg_mat)

    p <- vapply(
        seq_along(x),
        \(i) {
            df <- data.frame(
                x = c(matches[[i]], bg_mat[,i]),
                type = c("test", rep_len("control", n_iter))
            )
            fit <- glm(x~type, family = quasipoisson(), data = df)
            summary(fit)$coef[2, 4]
        }, numeric(1)
    )

    data.frame(
        sequences = n, matches, expected = mean_bg,
        enrichment, Z, p, n_iter, sd_bg
    )
}

#' @importFrom parallel mclapply
#' @importFrom stats glm poisson
#' @importFrom matrixStats colSds
#' @keywords internal
.testGlmPois <- function(
        x, stringset, bg, var, mc.cores, type = c("pwm", "cluster"), pc, seed,
        ...
) {

    ## Prior counts should be handled with caution here as using a fixed value
    ## will breach the Poisson distribution, e.g. adding a fixed value of 1 to
    ## a consistent run of zeros will give mean 1 with variance 0. Poisson
    ## noise may be preferable, but may profoundly reduce significance

    n <- length(stringset)
    stopifnot(var %in% colnames(mcols(bg)))
    splitbg <- split(bg, mcols(bg)[[var]])
    if (!all(vapply(splitbg, length, integer(1)) == n))
        stop("All iterations must be the same size as the test sequences")

    type <- match.arg(type)
    if (type == "pwm") {
        matches <- countPwmMatches(x, stringset, mc.cores = mc.cores, ...)
        bg_matches <- mclapply(
            splitbg, \(i) countPwmMatches(x, i, mc.cores = 1, ...),
            mc.cores = mc.cores
        )
    }
    if (type == "cluster") {
        matches <- countClusterMatches(x, stringset, mc.cores = mc.cores, ...)
        bg_matches <- mclapply(
            splitbg, \(i) countClusterMatches(x, i, mc.cores = 1, ...),
            mc.cores = mc.cores
        )
    }
    bg_mat <- do.call("rbind", bg_matches)
    ## Add Poisson prior counts
    if (pc != 0) {
        set.seed(seed)
        matches <- matches + stats::rpois(length(matches), pc)
        bg_mat <- bg_mat + stats::rpois(length(bg_mat), pc)
    }
    n_iter <- nrow(bg_mat)
    stopifnot(n_iter > 1)
    mean_bg <- colMeans(bg_mat)
    sd_bg <- colSds(bg_mat)
    Z <- (matches - mean_bg) / sd_bg
    Z[sd_bg == 0] <- NA_real_ # Handle where sd_bg == 0
    enrichment <- (matches) / colMeans(bg_mat)

    p <- vapply(
        seq_along(x),
        \(i) {
            df <- data.frame(
                x = c(matches[[i]], bg_mat[,i]),
                type = c("test", rep_len("control", n_iter))
            )
            fit <- glm(x~type, family = poisson(), data = df)
            summary(fit)$coef[2, 4]
        }, numeric(1)
    )

    data.frame(
        sequences = n, matches, expected = mean_bg,
        enrichment, Z, p, n_iter, sd_bg
    )
}

#' @importFrom parallel mclapply
#' @importFrom stats pchisq
#' @importFrom matrixStats colSds
#' @keywords internal
.testIter <- function(
        x, stringset, bg, var, mc.cores, type = c("pwm", "cluster"), pc, ...
) {

    stopifnot(var %in% colnames(mcols(bg)))
    n <- length(stringset)
    splitbg <- split(bg, mcols(bg)[[var]])
    if (!all(vapply(splitbg, length, integer(1)) == n))
        stop("All iterations must be the same size as the test sequences")
    type <- match.arg(type)
    if (type == "pwm") {
        matches <- countPwmMatches(x, stringset, mc.cores = mc.cores, ...) + pc
        bg_matches <- mclapply(
            splitbg, \(i) countPwmMatches(x, i, mc.cores = 1, ...) + pc,
            mc.cores = mc.cores
        )
    }
    if (type == "cluster") {
        matches <- countClusterMatches(x, stringset, mc.cores = mc.cores, ...) + pc
        bg_matches <- mclapply(
            splitbg, \(i) countClusterMatches(x, i, mc.cores = 1, ...) + pc,
            mc.cores = mc.cores
        )
    }
    bg_mat <- do.call("rbind", bg_matches)
    mean_bg <- colMeans(bg_mat)
    sd_bg <- colSds(bg_mat)
    n_iter <- nrow(bg_mat)
    stopifnot(n_iter > 1)
    diff <- bg_mat - matrix(
        matches, nrow = n_iter, ncol = length(matches), byrow = TRUE
    )
    iter_p <- (colSums(diff > 0) + 1) / n_iter
    Z <- (matches - mean_bg) / sd_bg
    Z[sd_bg == 0] <- NA_real_ ## Handle where sd_bg == 0
    p <- 1 - pchisq(Z^2, 1)

    data.frame(
        sequences = length(stringset), matches, expected = mean_bg,
        enrichment = matches / mean_bg, Z, p, iter_p, n_iter, sd_bg
    )

}


#' @importFrom stats poisson.test
#' @keywords internal
.testPois <- function(
        x, test_seq, bg_seq, mc.cores, type = c("pwm", "cluster"), pc, ...
){

    ## Prior counts can be added here as we have single-valued counts for each
    ## motif. This value is use to estimate lambda for mean & variance.
    ## Adding a single value, effectively increases the variance & makes the
    ## returned parameter estimates more conservative

    n_seq <- length(test_seq)
    type <- match.arg(type)
    if (type == "pwm") {
        matches <- countPwmMatches(x, test_seq, mc.cores = mc.cores, ...) + pc
        n_bg <- countPwmMatches(x, bg_seq, mc.cores = mc.cores, ...) + pc
    }
    if (type == "cluster") {
        matches <- countClusterMatches(x, test_seq, mc.cores = mc.cores, ...) + pc
        n_bg <- countClusterMatches(x, bg_seq, mc.cores = mc.cores, ...) + pc
    }
    est_bg_rate <- n_bg / length(bg_seq)
    expected <- est_bg_rate * n_seq
    ## Get the Z-scores & handle those with zero variance
    Z <- (matches - expected) / sqrt(expected)
    Z[sqrt(expected) == 0] <- NA_real_
    ## Running vapply seems faster than mclappy here
    p <- vapply(
        seq_along(x),
        \(i) poisson.test(matches[i], n_seq, est_bg_rate[i])$p.value,
        numeric(1)
    )
    data.frame(
        sequences = n_seq, matches, expected, enrichment = matches / expected,
        Z, p, est_bg_rate
    )

}

## Convert ranges to DNAStringSets
test_set <- ar_er_seq
bg_set <- getSeq(genome, bg_ranges)
mcols(bg_set) <- mcols(bg_ranges)


test_that("Basic Poisson analysis works", {
    res <- testMotifEnrich(esr1, test_set, bg_set, model = "poisson")
    expect_true(is(res, "data.frame"))
    expect_true(nrow(res) == 1)
    expect_true("est_bg_rate" %in% colnames(res))
})

test_that("Lists are tested", {
    res <- testMotifEnrich(um_db, test_set, bg_set, model = "poisson", sort_by = "none")
    expect_true(is(res, "data.frame"))
    expect_equal(rownames(res), vapply(um_db, slot, character(1), "name"))
})

test_that("Iteration works", {
    mcols(bg_set) <- mcols(bg_ranges)
    iter <- testMotifEnrich(esr1, test_set, bg_set, model = "iter")
    expect_true(is(iter, "data.frame"))
    expect_true(nrow(iter) == 1)
    expect_true("iter_p" %in% colnames(iter))
    expect_true(iter$n_iter == 10)

})

test_that("Quasipoisson works", {
    mcols(bg_set) <- mcols(bg_ranges)
    iter <- testMotifEnrich(esr1, test_set, bg_set, model = "quasi")
    expect_true(is(iter, "data.frame"))
    expect_true(nrow(iter) == 1)
    expect_true(iter$n_iter == 10)

})

test_that("GLM-Poisson works", {
    mcols(bg_set) <- mcols(bg_ranges)
    iter <- testMotifEnrich(esr1, test_set, bg_set, model = "glm")
    expect_true(is(iter, "data.frame"))
    expect_true(nrow(iter) == 1)
    expect_true(iter$n_iter == 10)

})

test_that("Hypergeometric works", {
    mcols(bg_set) <- mcols(bg_ranges)
    hg <- testMotifEnrich(esr1, test_set, bg_set, model = "hyper")
    expect_true(is(hg, "data.frame"))
    expect_true(nrow(hg) == 1)
    expect_true("odds_ratio" %in% colnames(hg))
    expect_message(
        testMotifEnrich(
            esr1, test_set, c(bg_set, test_set[1]), model = "hyper"
        ),
        "Shared.+"
    )

})

test_that("Errors are caught",{
    expect_error(.testIter(list(esr1), test_set, bg_set[-1]))
})

#' Test for a Uniform Distribution across a set of best matches
#'
#' @details
#' This function tests for an even positional spread of motif matches across a
#' set of sequences, using the assumption (i.e. H~0~) that if there is no
#' positional bias, matches will be evenly distributed across all positions
#' within a set of sequences.
#' Conversely, if there is positional bias, typically but not necessarily near
#' the centre of a range, this function intends to detect this signal, as a
#' rejection of the null hypothesis.
#'
#' Input can be provided as the output from \link{getPwmMatches} setting
#' `best_only = TRUE` if these matches have already been identified.
#' If choosing to provide this object to the argument `matches`, nothing is required
#' for the arguments `pwm`, `stringset`, `rc`, `min_score` or `break_ties`
#' Otherwise, a Position Weight Matrix (PWM) and an `XStringSet` are required,
#' along with the relevant arguments, with best matches identified within the
#' function.
#'
#' The set of best matches are then grouped into bins along the range, with the
#' central bin containing zero, and tallied.
#' Setting `abs` to `TRUE` will set all positions from the centre as
#' *absolute values*, returning counts purely as bins with distances from zero,
#' marking this as an inclusive lower bound.
#' Motif alignments are assigned into bins based on the central position of the
#' match, as provided in the column `from_centre` when calling
#' \link{getPwmMatches}.
#'
#' The \link[stats]{binom.test} is performed on each bin using the alternative
#' hypothesis, with the returned p-values across all bins combined using the
#' Harmonic Mean p-value (HMP) (See \link[harmonicmeanp]{p.hmp}).
#' All bins with raw p-values below the HMP are identified and the returned
#' values for start, end, centre, width, matches in region, expected and
#' enrichment are across this set of bins.
#' The expectation is that where a positional bias is evident, this will be a
#' narrow range containing a non-trivial proportion of the total matches.
#'
#' It should also be noted that `binom.test()` can return p-values of zero, as
#' beyond machine precision. In these instances, zero p-values are excluded from
#' calculation of the HMP. This will give a very slight conservative bias, and
#' assumes that for these extreme cases, neighbouring bins are highly likely to
#' also return extremely low p-values and no significance will be lost.
#'
#'
#' @return
#' A data.frame with columns `start`, `end`, `centre`, `width`, `total_matches`,
#' `matches_in_region`, `expected`, `enrichment`, `prop_total`, `p`
#' and `consensus_motif`
#' The total matches represent the total number of matches within the set of
#' sequences, whilst the number observed in the final region are also given,
#' along with the proportion of the total this represents.
#' Enrichment is simply the ratio of observed to expected based on the
#' expectation of the null hypothesis
#'
#' The consensus motif across all matches is returned as a Position Frequency
#' Matrix (PFM) using \link[Biostrings]{consensusMatrix}.
#'
#'
#' @param x A Position Weight Matrix, universalmotif object or list thereof.
#' Alternatively can be a single DataFrame or list of DataFrames as returned
#' by \link{getPwmMatches} with `best_only = TRUE`
#' @param stringset An XStringSet. Not required if matches are supplied as x
#' @param binwidth Width of bins across the range to group data into
#' @param abs Use absolute positions around zero to find symmetrical enrichment
#' @param rc logical(1) Also find matches using the reverse complement of each
#' PWM
#' @param min_score The minimum score to return a match
#' @param break_ties Choose how to resolve matches with tied scores
#' @param alt Alternative hypothesis for the binomial test
#' @param sort_by Column to sort results by
#' @param mc.cores Passed to \link[parallel]{mclapply}
#' @param ... Passed to \link[Biostrings]{matchPWM}
#'
#'
#' @examples
#' ## Load the example PWM
#' data("ex_pfm")
#' esr1 <- ex_pfm$ESR1
#'
#' ## Load the example sequences
#' data("ar_er_seq")
#'
#' ## Get the best match and use this data
#' matches <- getPwmMatches(esr1, ar_er_seq, best_only = TRUE)
#' ## Test for enrichment in any position
#' testMotifPos(matches)
#'
#' ## Provide a list of PWMs, testing for distance from zero
#' testMotifPos(ex_pfm, ar_er_seq, abs = TRUE, binwidth = 10)
#'
#'
#' @importFrom parallel mclapply
#' @importFrom stats p.adjust
#' @export
testMotifPos <- function(
        x, stringset, binwidth = 10, abs = FALSE, rc = TRUE, min_score = "80%",
        break_ties = "all", alt = c("greater", "less", "two.sided"),
        sort_by = c("p", "none"), mc.cores = 1,
        ...
) {

    ## Handle single objects by coercing to a list where appropriate
    valid_cl <- c("matrix", "universalmotif", "DataFrame")
    if (any(vapply(valid_cl, \(cl) is(x, cl), logical(1)))) x <- list(x)
    if (!is.list(x))
        stop("Input should be provided as a list, PWM/matrix or DF of matches")

    ## Handle a list of PWMs
    isPWM <- vapply(
        x, \(el) is(el, "matrix") | is(el, "universalmotif"), logical(1)
    )
    if (all(isPWM)) {
        ## Ensure all sequences are the same width
        if (length(unique(width(stringset))) != 1)
            stop("All sequences are required to be the same width")
        matches <- getPwmMatches(
            x, stringset, rc, min_score, best_only = TRUE, break_ties,
            mc.cores, ...
        )
    } else {
        ## Now for a list of matches
        .checkMatches(x) ## Will fail if not valid
        matches <- x
    }
    if (missing(matches)) stop("Provided input is not in a recognised format")
    alt <- match.arg(alt)

    cols <- c(
        "start", "end", "centre", "width", "total_matches", "matches_in_region",
        "expected", "enrichment", "prop_total", "p", "fdr", "consensus_motif"
    )

    out <- mclapply(
        matches, .testSingleMotifPos, binwidth = binwidth, abs = abs, rc = rc,
        min_score = min_score, break_ties = break_ties, alt = alt, ...,
        mc.cores = mc.cores
    )
    out <- do.call("rbind", out)
    out$fdr <- p.adjust(out$p, "fdr")
    o <- seq_len(nrow(out))
    sort_by <- match.arg(sort_by)
    if (sort_by != "none") o <- order(out[[sort_by]])
    out[o, cols]

}

#' @importFrom harmonicmeanp p.hmp
#' @importFrom stats binom.test
#' @keywords internal
.testSingleMotifPos <- function(
        matches, binwidth, abs, rc, min_score, break_ties, alt, ...
) {

    ## Setup a well formed output object for easy combining with other tests
    out_cols <- c(
        "start", "end", "centre", "width", "total_matches", "matches_in_region",
        "expected", "enrichment", "prop_total", "p", "consensus_motif"
    )
    out <- lapply(out_cols, \(x) integer())
    out$consensus_motif <- list()
    names(out) <- out_cols
    n_matches <- nrow(matches)
    if (n_matches == 0) return(data.frame(out))
    matches <- .makeBmBins(matches, binwidth, abs)
    bins <- levels(matches$bin)

    ## Get the counts & form a df with every bin as a row
    counts <- vapply(
        splitAsList(matches, matches$bin), \(x) sum(x$weight), numeric(1)
    )
    df <- data.frame(bin = names(counts), matches_in_bin = as.integer(counts))
    bin_coords <- do.call("rbind", strsplit(df$bin, split = ","))
    bin_coords <- apply(
        bin_coords, 2, \(x) as.integer(gsub("\\[|\\(|\\]", "", x))
    )
    df$start <- bin_coords[, 1]
    df$end <- bin_coords[, 2]
    df$bin <- factor(df$bin, levels = bins)
    ## Given matches are measured from the centre, only bases > d_from_edge
    ## away from the ends of the range are viable
    motif_width <- max(width(matches$match))
    d_from_edge <- (motif_width - 1) / 2
    valid_pos <- seq(min(df$start) + d_from_edge, max(df$end) - d_from_edge)
    if (abs) valid_pos <- seq(d_from_edge %% 1, max(df$end) - d_from_edge)
    df$valid_width <- vapply(
        seq_len(nrow(df)),
        \(i) sum(valid_pos > df$start[i] & valid_pos <= df$end[i]), numeric(1)
    )
    df$bin_prob <- df$valid_width / sum(df$valid_width)
    df$expected <- n_matches * df$bin_prob
    df$p <- vapply(
        seq_along(df$bin),
        \(i) {
            x <- df[i,]
            binom.test(x$matches_in_bin, n_matches, x$bin_prob, alt)$p.value
        }, numeric(1)
    )
    ## It is entirely possible for this to return p < Machine precision
    ## This will cause the HMP to fail so setting to the lowest Machine
    ## precision will be conservative, but a bit of a hack.
    df$p[df$p < .Machine$double.xmin] <- .Machine$double.xmin
    hmp <- p.hmp(df$p, L = length(bins))
    df <- subset(df, df$p < hmp | df$p == min(df$p))
    out <- data.frame(start = min(df$start), end = max(df$end))
    out$centre <- (out$start + out$end) / 2
    out$width <- out$end - out$start
    out$total_matches <- n_matches
    out$matches_in_region <- sum(df$matches_in_bin)
    out$expected <- sum(df$expected)
    out$enrichment <- out$matches_in_region / out$expected
    out$prop_total <- out$matches_in_region / out$total_matches
    out$p <- as.numeric(hmp)

    ## Setup the consensus motif to return
    consensus <- consensusMatrix(matches$match)[c("A", "C", "G", "T"),]
    out$consensus_motif <- list(consensus)
    out[, out_cols]
}

exp_cols <- c(
    "start", "end", "centre", "width", "total_matches", "matches_in_region",
    "expected", "enrichment", "prop_total", "p", "fdr", "consensus_motif"
)

## Bodgy up a set of matches
matches <- getPwmMatches(esr1, stringset)
matches$seq <- c(1, seq_len(nrow(matches) - 1))

test_that("testMotifPos returns symmetrical peaks as expected",{
    res <- testMotifPos(matches)
    expect_true(nrow(res) == 1)
    expect_equal(colnames(res), exp_cols)
    expect_equal(
        res[,1:6],
        data.frame(
            start = 35, end = 45, centre = 40, width = 10,
            total_matches = 1, matches_in_region = 1
        )
    )
    expect_true(is(res$consensus_motif[[1]], "matrix"))
    expect_equal(dim(res$consensus_motif[[1]]), c(4, 9))
})

test_that("setting abs = TRUE behaves as expected", {
    bins <- .makeBmBins(matches, binwidth = 10, abs = TRUE)
    expect_equal(bins$weight, 1)
    exp_bins <- c(
        "[0,10]", "(10,20]", "(20,30]", "(30,40]", "(40,50]", "(50,60]",
        "(60,70]", "(70,80]", "(80,90]", "(90,100]", "(100,110]", "(110,120]",
        "(120,130]", "(130,140]", "(140,150]", "(150,160]", "(160,170]",
        "(170,180]", "(180,190]", "(190,200]"
    )
    expect_equal(levels(bins$bin), exp_bins)

})

test_that("Passing to getPwmMatches works", {
    res <- testMotifPos(esr1, stringset, abs = TRUE, binwidth = 100)
    expect_true(nrow(res) == 1)
    expect_equal(
        res[,1:6],
        data.frame(
            start = 0, end = 100, centre = 50, width = 100,
            total_matches = 1, matches_in_region = 1
        )
    )
})

test_that("List input works", {
    res <- testMotifPos(ex_pfm, seq, sort_by = "none")
    expect_true(is(res, "data.frame"))
    expect_equal(rownames(res), names(ex_pfm))
    expect_true(all(exp_cols %in% colnames(res)))
    expect_true("fdr" %in% colnames(res))
})

test_that("testMotifPos errors correctly", {
    empty <- testMotifPos(matches[0,])
    expect_true(nrow(empty) == 0)
    expect_true(is(empty, "data.frame"))
    expect_equal(colnames(empty), exp_cols)
    expect_error(testMotifPos(esr1, stringset, binwidth = 500))
    expect_error(testMotifPos(""))
    expect_error(testMotifPos(c(list(""), ex_pfm)))
})

test_that("Same width sequences are enforced", {
    matches$seq_width[1] <- 100
    expect_error(testMotifPos(matches))
    mixed_width <- c(stringset, as(stringset[[1]][1:200], "DNAStringSet"))
    expect_error(testMotifPos(ex_pfm, mixed_width))
})

#' @title Form a set of random, matching ranges for bootstrapping or permuting
#'
#' @description
#' Form a set of ranges from y which (near) exactly match those in x for use as
#' a background set requiring matching
#'
#' @details
#' This function uses the width distribution of the 'test' ranges (i.e. `x`) to
#' randomly sample a set of ranges with matching width from the ranges provided
#' in `y`. The width distribution will clearly be exact when a set of
#' fixed-width ranges is passed to `x`, whilst random sampling may yield some
#' variability when matching ranges of variable width.
#'
#' When both x and y are GRanges objects, they are implicitly assumed to both
#' represent similar ranges, such as those overlapping a promoter or enhancer.
#' When passing two GRangesList objects, both objects are expected to contain
#' ranges annotated as belonging to key features, such that the list elements in
#' y must encompass all elements in x.
#' For example if `x` contains two elements named 'promoter' and 'intron', `y`
#' should also contain elements named 'promoter' and 'intron' and these will
#' be sampled as matching ranges for the same element in `x`. If elements of
#' `x` and `y` are not named, they are assumed to be in matching order.
#'
#' The default behaviour is to assume that randomly-generated ranges are for
#' iteration, and as such, ranges are randomly formed in multiples of the number
#' of 'test' ranges provided in `x`. The column `iteration` will be added to the
#' returned ranges.
#' Placing any number into the `n_total` argument will instead select a total
#' number of ranges as specified here. In this case, no `iteration` column will
#' be included in the returned ranges.
#'
#' Sampling is assumed to be with replacement as this is most suitable for
#' bootstrapping and related procedures, although this can be disabled by
#' setting `replace = FALSE`
#'
#' @return A GRanges or GRangesList object
#'
#'
#' @param x GRanges/GRangesList with ranges to be matched
#' @param y GRanges/GRangesList with ranges to select random matching ranges from
#' @param exclude GRanges of ranges to omit from testing
#' @param n_iter The number of times to repeat the random selection process
#' @param n_total Setting this value will over-ride anything set using n_iter.
#' Can be vector of any length, corresponding to the length of x, when x is a
#' GRangesList
#' @param replace logical(1) Sample with our without replacement when creating
#' the set of random ranges.
#' @param mc.cores Passsed to \link[parallel]{mclapply}
#' @param ... Not used
#' @param force_ol logical(1) Enforce an overlap between every site in x and y
#' @param unlist logical(1) Return as a sorted GRanges object, or leave as a
#' GRangesList
#'
#' @examples
#' ## Load the example peaks
#' data("ar_er_peaks")
#' sq <- seqinfo(ar_er_peaks)
#' ## Now sample size-matched ranges for two iterations from chr1
#' makeRMRanges(ar_er_peaks, GRanges(sq)[1], n_iter = 2)
#'
#' ## Or simply sample 100 ranges if not planning any iterative analyses
#' makeRMRanges(ar_er_peaks, GRanges(sq)[1], n_total = 100)
#'
#'
#' @rdname makeRMRanges-methods
#' @aliases makeRMRanges
#' @export
setGeneric(
    "makeRMRanges",
    function(x, y, ...) standardGeneric("makeRMRanges")
)
#' @import GenomicRanges
#' @importFrom IRanges overlapsAny width
#' @importFrom Seqinfo seqinfo
#' @rdname makeRMRanges-methods
#' @aliases makeRMRanges
#' @export
setMethod(
    "makeRMRanges", signature = signature(x = "GRanges", y = "GRanges"),
    function(
        x, y, exclude = GRanges(), n_iter = 1, n_total = NULL, replace = TRUE,
        ..., force_ol = TRUE
    ) {

        ## Check some basic features, e.g. y must be bigger than x
        stopifnot(sum(width(y)) > sum(width(x)))
        ## There must be overlap if bootstrapping.
        if (force_ol) stopifnot(all(overlapsAny(x, y)))

        # Make sure any new ranges don't fall off the end of chromosomes
        stopifnot(is(exclude, "GRanges"))
        sq_gr <- GRanges(seqinfo(x))
        w <- width(x)
        max_w <- max(w)
        chr_ends <- c(
            resize(sq_gr, width = max_w / 2, fix = "start"),
            resize(sq_gr, width = max_w / 2, fix = "end")
        )
        # Chop the ends off so resizing won't spit an error
        exclude <- GenomicRanges::setdiff(exclude, chr_ends)
        exclude <- resize(exclude, width(exclude) + max_w, fix = "center")
        exclude <- c(granges(exclude), chr_ends)
        y <- GenomicRanges::setdiff(y, exclude)

        ## Set the iterations
        has_iter <- is.null(n_total)
        if (has_iter) {
            iter_size <- length(x)
            n_total <- as.integer(n_iter * iter_size)
        }
        n_total <- n_total[[1]]

        ## Now the random sampling
        rand_w <- sample(w, n_total, replace = TRUE)
        too_big <- sum(width(y)) > .Machine$integer.max
        if (too_big) {
            ## The human genome is ~3.1e9 & the usual max integer is 2.1e9
            ## Splitting in half should solve problems for all genomes < 4e9
            ## Sorry to wheat people...
            i <- floor(length(y) / 2)
            n1 <- floor(n_total / 2)
            y1 <- y[seq_len(i)]
            i1 <- sample(sum(width(y1)), n1, replace = replace)
            gr1 <- GPos(y1)[i1]

            y2 <- GenomicRanges::setdiff(y, y1)
            n2 <- n_total - n1
            i2 <- sample(sum(width(y2)), n2, replace = replace)
            gr2 <- GPos(y2)[i2]

            ## Randomise to make up for chunking
            i <- sample(n_total, n_total, replace = FALSE)
            gr <- GRanges(c(gr1, gr2))[i]

        } else {
            gpos <- GPos(y)
            i <- sample(seq_along(gpos), n_total, replace = replace)
            gr <- GRanges(gpos[i])
        }
        gr <- resize(gr, width = rand_w, fix = "center")
        ## Ranges will be randomised so adding iterations in order is fine
        if (has_iter) gr$iteration <- rep(seq_len(n_iter), times = iter_size)
        sort(gr)

    }
)
#' @import GenomicRanges
#' @importFrom parallel mclapply
#' @rdname makeRMRanges-methods
#' @aliases makeRMRanges
#' @export
setMethod(
    "makeRMRanges",
    signature = signature(x = "GRangesList", y = "GRangesList"),
    function(
        x, y, exclude = GRanges(), n_iter = 1, n_total = NULL, replace = TRUE,
        mc.cores = 1, ..., force_ol = TRUE, unlist = TRUE
    ){

        stopifnot(length(x) <= length(y))
        nm_x <- names(x)
        nm_y <- names(y)
        if (!is.null(nm_x) & !is.null(nm_y)) {
            ## Set GRLists in matching order if named
            stopifnot(all(nm_x %in% nm_y))
            y <- y[nm_x]
        }
        i <- seq_along(x)
        n_iter <- rep_len(n_iter, length(x))
        n_total <- lapply(i, \(j) n_total[[j]])
        out <- mclapply(
            i,
            \(i) makeRMRanges(
                x[[i]], y[[i]], exclude = exclude,  n_iter = n_iter[[i]],
                n_total = n_total[[i]], replace = replace, ...,
                force_ol = force_ol
            ), mc.cores = mc.cores
        )
        if (!is.null(nm_x)) names(out) <- nm_x
        out <- GRangesList(out)
        if (unlist) out <- sort(unlist(out))
        out
    }
)

#' @importFrom methods slot is
#' @importClassesFrom universalmotif universalmotif
#' @importFrom universalmotif create_motif convert_type
#' @keywords internal
.checkPWM <- function(motif){
    ## Convert any PPM/PCM matrices to PWM
    ## May break a few things downstream
    if (is.matrix(motif)) {
        nsites <- 100
        if (is.integer(motif)) nsites <- max(matrixStats::colSums2(motif))
        motif <- create_motif(
            motif, nsites = nsites, pseudocount = 1, type = "PPM"
        )
    }
    if (!is(motif, "universalmotif")) stop("Could not determine motif type")
    if (slot(motif, "pseudocount") == 0)
        warning("Zero pseudocounts may lead to a PWM with infinite values")
    pwm <- convert_type(motif, "PWM")
    slot(pwm, "motif")
}

#' @keywords internal
.viewMapFromXStringset <- function(stringset){
    stopifnot(is(stringset, "XStringSet"))
    # Form a map from the original sequences to a views object
    map <- data.frame(end = cumsum(width(stringset)), width = width(stringset))
    map$start <- map$end - map$width + 1
    map$names <- names(stringset)
    map
}

#' @keywords internal
#' @importClassesFrom universalmotif universalmotif
.cleanMotifList <- function(x, nt = c("A", "C", "G", "T")) {
    if (all(vapply(x, is, logical(1), "universalmotif"))) {
        ## Add the name as the motif name, swicthing to altname where required
        n <- length(x)
        nm <- vapply(x, slot, character(1), "name")
        if (length(unique(nm)) < n) {
            message("Found non-unique values in the name slot. Trying altname")
            nm <- unlist(lapply(x, slot, "altname"))
            if (length(unique(nm)) < n) {
                msg <- "Non-unique values in the altname slot. Please resolve"
                stop(msg)
            }
        }
        x <- lapply(x, slot, "motif")
        names(x) <- nm

        # nm <- unlist(lapply(x, slot, "altname")) # Less duplication than name
        # if (all(is.na(nm))) { # Handle the case where altname is empty (NA)
        #     nm <- vapply(x, slot, character(1), "name")
        # }
        # names(x) <- nm
    }
    ## Now check everything is a matrix
    all_mat <- vapply(
        x, \(mat) all(is.matrix(mat), rownames(mat) == nt), logical(1)
    )
    stopifnot(all_mat)
    x
}

#' @keywords internal
.checkMatches <- function(matches) {
    colTypes <- c(
        score = "numeric", direction = "factor",
        start = "integer", end = "integer",
        from_centre = "numeric", seq_width = "integer", match = "DNAStringSet"
    )
    reqdCols <- c("seq", names(colTypes)) # seq can be any type
    msg <- paste(reqdCols, collapse = ",")

    if (is(matches, "DataFrame")) {
        stopifnot(all(reqdCols %in% colnames(matches)))
        types <- vapply(matches, \(x) is(x)[[1]], character(1))
        stopifnot(
            identical(types[names(colTypes)], colTypes)
        )
        if (nrow(matches)) {
            wd <- unique(matches$seq_width)
            stopifnot(length(wd) == 1) # Enforce a fixed width
        }
    } else {
        stopifnot(is(matches, "list") & !is(matches, "data.frame"))
        hasCols <- vapply(
            matches, \(x) all(reqdCols %in% colnames(x)), logical(1)
        )
        if (any(!hasCols))
            stop("All objects must contain the columns: \n", msg)
        correctTypes <- vapply(
            matches,
            \(x) {
                types <- vapply(x, \(cols) is(cols)[[1]], character(1))
                identical(types[names(colTypes)], colTypes)
            }, logical(1)
        )
        if (any(!correctTypes)) stop("All columns are not of the correct type")
        wd <- vapply(matches, \(x) length(unique(x$seq_width)), integer(1))
        if (any(wd > 1)) stop("All sequences must be of identical width")
    }

    invisible(TRUE)
}

#' @keywords internal
.makeBmBins <- function(matches, binwidth, abs){

    ## Set sequence weights for multiple matches
    reps <- table(as.character(matches$seq)) # Counts reps
    matches$weight <- as.numeric(1 / reps[as.character(matches$seq)]) # 1 / reps

    ## Make the bins
    width <- max(matches$seq_width) # Now enforced to be fixed width
    stopifnot(binwidth < width)
    if (abs) {
        ## Set the dist from centre to be +ve only
        matches$from_centre <- abs(matches$from_centre)
        all_breaks <- c(seq(0, width / 2, by = binwidth), width / 2)
        seq_per_bin <- vapply(
            all_breaks, \(x) sum(matches$seq_width >= x), integer(1)
        )
    } else {
        ## To define bins, make sure the central bin is around zero and that
        ## bins extends symetrically from this point to the widest sequence
        pos_breaks <- c(
            ## Ensure the upper limit is included, even if it leaves a smaller bin
            seq(binwidth / 2, width / 2, by = binwidth), width / 2
        )
        all_breaks <- c(-1 * pos_breaks, pos_breaks)
        pos_bins <- vapply(
            pos_breaks, \(x) sum(matches$seq_width >= x), integer(1)
        )
        seq_per_bin <- c(pos_bins, pos_bins)
    }

    ## Cut into bins
    all_breaks <- unique(sort(all_breaks))
    matches$bin <- cut(matches$from_centre, breaks = all_breaks, include.lowest = TRUE)
    matches
}

#' @importFrom IRanges Views
#' @import Biostrings
#' @keywords internal
.hasPwmMatch <- function(pwm, stringset, rc = TRUE, min_score = "80%", ...) {
    ## Returns a logical vector the same length as the input stringset
    ## Checks & the map
    pwm <- .checkPWM(pwm)
    stopifnot(is(stringset, "XStringSet"))

    ## Handle empty stringsets
    n <- length(stringset)
    seq_with_hits <- NULL

    if (n > 0) {
        map <- .viewMapFromXStringset(stringset)
        # Form the entire XStringSetList into a Views object
        views <- Views(
            unlist(stringset), start = map$start, width = map$width,
            names = map$names
        )
        hits <- matchPWM(pwm, views, min.score = min_score, ...)
        if (rc) {
            rev_pwm <- reverseComplement(pwm)
            hits_rev <- matchPWM(rev_pwm, views, min.score = min_score, ...)
            hits <- c(hits, hits_rev)
        }
        seq_with_hits <- findInterval(start(hits), map$start)
    }
    seq_len(n) %in% seq_with_hits

}

#' @importFrom IRanges Views
#' @import Biostrings
#' @keywords internal
.hasClusterMatch <- function(cl, stringset, rc = TRUE, min_score = "80%", ...){

    ## Returns a logical vector the same length as the input stringset
    ## Checks & the map
    cl <- lapply(cl, .checkPWM)
    stopifnot(is(stringset, "XStringSet"))

    ## Handle empty stringsets
    n <- length(stringset)
    seq_with_hits <- NULL

    if (rc) {
        cl <- c(cl, lapply(cl, reverseComplement))
    }

    if (n > 0) {
        map <- .viewMapFromXStringset(stringset)
        # Form the entire XStringSetList into a Views object
        views <- Views(
            unlist(stringset), start = map$start, width = map$width,
            names = map$names
        )
        hits <- lapply(cl, matchPWM, subject = views, min.score = min_score, ...)
        starts <- unlist(lapply(hits, start))
        starts <- unique(sort(starts))
        seq_with_hits <- findInterval(starts, map$start)
    }
    seq_len(n) %in% seq_with_hits

}

test_that(".cleanMotifList works as expected", {
    ml <- .cleanMotifList(um_db)
    expect_equal(names(ml), vapply(um_db, slot, character(1), "name"))
    expect_true(all(vapply(ml, is, logical(1), "matrix")))
    um_db_mod <- um_db[c(1, 2, 2)]
    expect_error(suppressMessages(.cleanMotifList(um_db_mod)))
    slot(um_db_mod[[3]], "altname") <- "Test3"
    expect_message(
        ml <- .cleanMotifList(um_db_mod), "Found non-uniq.+"
    )
    expect_equal(names(ml), paste0("Test", 1:3))
})

test_that("universalmotifs are parsed to matrices", {
    mat <- .checkPWM(um_db[[1]])
    expect_true(is(mat, "matrix"))
})

test_that("checkMatches checks work", {
    expect_error(.checkMatches(data.frame()))
    expect_error(.checkMatches(DataFrame()))
    matches <- getPwmMatches(esr1, stringset)
    expect_true(.checkMatches(matches))
    expect_true(.checkMatches(list(matches)))
    expect_error(.checkMatches(matches[1:2]))
    matches$direction <- as.character(matches$direction)
    expect_error(.checkMatches(matches))
})

#' @title Find matches from a PWM cluster within an XStringSet
#'
#' @description
#' Find matches from a PWM cluster within a set of sequences
#'
#' @details
#' This function extends \link{getPwmMatches} by returning a single set of
#' results for set of clustered motifs.
#' This can help remove some of the redundancy in results returned for highly
#' similar PWMs, such as those in the GATA3 family.
#'
#' Taking a set of sequences as an XStringSet, find all matches above the
#' supplied score (i.e. threshold) for a list of Position Weight Matrices
#' (PWMs), which have been clustered together as highly-related motifs.
#' By default, matches are performed using the PWMs as provided and the reverse
#' complement, however this can easily be disabled by setting `rc = FALSE`.
#'
#' The function relies heavily on \link[Biostrings]{matchPWM} and
#' \link[IRanges]{Views} for speed.
#'
#' Where overlapping matches are found for the PWMs within a cluster, only a
#' single match is returned.
#' The motif with the highest relative score (score / maxScore(PWM)) is selected.
#'
#' When choosing to return the best match (`best_only = TRUE`), only the match
#' with the highest relative score is returned for each sequence.
#' Should there be tied scores, the best match can be chosen as either the first,
#' last, most central, all tied matches, or choosing one at random (the default).
#'
#' @param cl A list of Position Weight Matrices, universalmotifs, with each
#' element representing clusters of related matrices
#' @param stringset An XStringSet
#' @param rc logical(1) Also find matches using the reverse complement of PWMs
#' in the cluster
#' @param min_score The minimum score to return a match
#' @param best_only logical(1) Only return the best match
#' @param break_ties Method for breaking ties when only returning the best match
#' Ignored when all matches are returned (the default)
#' @param mc.cores Passed to \link[parallel]{mclapply}
#' @param ... Passed to \link[Biostrings]{matchPWM}
#'
#' @return
#' Output from getClusterMatches will be a list of DataFrames with columns:
#' `seq`, `score`, `direction`, `start`, `end`, `from_centre`, `seq_width`,
#' `motif` and `match`
#'
#' The first three columns describe the sequence with matches, the score of
#' the match and whether the match was found using the forward or reverse PWM.
#' The columns `start`, `end` and `width` describe the where the match was found
#' in the sequence, whilst `from_centre` defines the distance between the centre
#' of the match and the centre of the sequence being queried.
#' The motif column denotes which individual motif was found to match in this
#' position, again noting that when matches overlap, only the one with the
#' highest relative score is returned.
#' The final column contains the matching fragment of the sequence as an
#' `XStringSet`.
#'
#' Output from countClusterMatches will be a simple integer vector the same
#' length as the number of clusters
#'
#' @examples
#' # Load example PFMs
#' data("ex_pfm")
#' # Cluster using default settings
#' cl_ids <- clusterMotifs(ex_pfm)
#' ex_cl <- split(ex_pfm, cl_ids)
#' # Add optional names
#' names(ex_cl) <- vapply(ex_cl, \(x) paste(names(x), collapse = "/"), character(1))
#'
#' # Load example sequences
#' data("ar_er_seq")
#' # Get all matches for each cluster
#' getClusterMatches(ex_cl, ar_er_seq)
#' # Or Just count them
#' countClusterMatches(ex_cl, ar_er_seq)
#' # Compare this to individual counts
#' countPwmMatches(ex_pfm, ar_er_seq)
#'
#' @importFrom parallel mclapply
#' @importFrom methods slot is
#' @importClassesFrom universalmotif universalmotif
#' @export
#' @rdname getClusterMatches
getClusterMatches <- function(
        cl, stringset, rc = TRUE, min_score = "80%", best_only = FALSE,
        break_ties = c("all", "random", "first", "last", "central"),
        mc.cores = 1, ...
) {

    stopifnot(is.list(cl))
    break_ties <- match.arg(break_ties)
    args <- c(as.list(environment()), list(...))
    args <- args[names(args) != "mc.cores"]
    nm_type <- "integer"
    if (!is.null(names(stringset))) nm_type <- "character"
    if (!length(cl)) return(list())
    out <- NULL

    # Form the entire XStringSetList into a Views object
    map <- .viewMapFromXStringset(stringset)
    views <- Views(
        unlist(stringset), map$start, width = map$width, names = map$names
    )

    if (
        all(vapply(cl, is, logical(1), "list")) |
        all(vapply(cl, is, logical(1), "universalmotif"))
    ) {
        nm <- names(cl)
        out <- vector("list", length(cl))
        names(out) <- nm
        ret_cols <- c(
            "seq", "score", "direction", "start", "end", "from_centre",
            "seq_width", "motif", "match"
        )
        cl <- lapply(cl, .cleanMotifList)
        single_pwm <- vapply(cl, length, integer(1)) == 1
        out[single_pwm] <- mclapply(
            cl[single_pwm],
            \(x) {
                DF <- .getSinglePwmMatches(
                    x[[1]], views = views, rc = rc,
                    min_score = min_score, best_only = best_only,
                    break_ties = break_ties, nm_type = nm_type
                )
                DF$motif <- rep_len(names(x), nrow(DF))
                DF[ret_cols]
            },  mc.cores = mc.cores
        )
        out[!single_pwm] <- mclapply(
            cl[!single_pwm], .getClusterPwmMatches, views = views,
            rc = rc, min_score = min_score, best_only = best_only,
            break_ties = break_ties, nm_type = nm_type, mc.cores = mc.cores
        )
    }
    if (all(vapply(cl, is, logical(1), "matrix"))) {
        args$nm_type <- nm_type
        args$stringset <- NULL
        args$views <- views
        out <- do.call(".getClusterPwmMatches", args)
    }

    if (is.null(out)) message("Could not determine clusters")
    out

}
#' @importFrom methods is
#' @export
#' @rdname getClusterMatches
countClusterMatches <- function(
        cl, stringset, rc = TRUE, min_score = "80%", mc.cores = 1, ...
) {

    stopifnot(is.list(cl))
    out <- NULL
    if (!length(cl)) return(out)

    is_clust_list <- all(vapply(cl, is, logical(1), "list"))
    is_motif_list <- (
        all(vapply(cl, is, logical(1), "universalmotif")) |
            all(vapply(cl, is, logical(1), "matrix"))
    )
    if (!is_clust_list & is_motif_list) cl <- list(cl)
    cl <- lapply(cl, \(x) lapply(x, .checkPWM))
    cl <- lapply(cl, .cleanMotifList)
    nm <- names(cl)

    # Form the entire XStringSetList into a Views object
    map <- .viewMapFromXStringset(stringset)
    views <- Views(
        unlist(stringset), map$start, width = map$width, names = map$names
    )

    single_pwm <- vapply(cl, length, integer(1)) == 1
    out <- vector("list", length(cl))
    names(out) <- nm
    out[single_pwm] <- mclapply(
        cl[single_pwm],
        \(x) .countSinglePwmMatches(x[[1]], views, rc, min_score),
        mc.cores = mc.cores
    )
    out[!single_pwm] <- mclapply(
        cl[!single_pwm],
        .getClusterPwmMatches, views = views, rc = rc,
        min_score = min_score, counts_only = TRUE, mc.cores = mc.cores
    )
    out <- unlist(out)

    if (is.null(out)) message("Could not determine clusters")
    out

}


#' @import Biostrings
#' @importFrom IRanges Views IRangesList DataFrameList findOverlaps
#' @importFrom S4Vectors mcols<- queryHits subjectHits
#' @keywords internal
.getClusterPwmMatches <- function(
        cl, views, rc, min_score, best_only = FALSE, break_ties,
        nm_type = "integer", counts_only = FALSE, ...
){

    ## Checks
    cl <- lapply(cl, .checkPWM)
    stopifnot(is(views, "XStringViews"))
    ## Handle empty stringsets
    empty_df <- .emptyPwmDF(nm_type)
    empty_df$motif <- character()
    cols <- c(
        "seq", "score", "direction", "start", "end", "from_centre",
        "seq_width", "motif", "match"
    )
    if (!length(views)) return(empty_df[cols])

    # Find all hits
    hits <- lapply(
        cl, matchPWM, subject = views, min.score = min_score, with.score = TRUE,
        ...
    )
    .add_dir <- function(x, dir) {
        mcols(x)$direction <- rep_len(dir, length(x))
        x
    }
    hits <- lapply(hits, .add_dir, dir = "F")
    hits_rev <- list()
    if (rc) {
        rev_cl <- lapply(cl, reverseComplement)
        hits_rev <- lapply(
            rev_cl, matchPWM, subject = views, min.score = min_score,
            with.score = TRUE, ...
        )
        hits_rev <- lapply(hits_rev, .add_dir, dir = "R")
    }

    ## Combine Hits
    ir <- lapply(c(hits, hits_rev), slot, "ranges")
    all_hits <- unlist(IRangesList(ir))
    if (length(all_hits) == 0 & !counts_only) return(empty_df)

    ## Add mcols if we have hits
    mcols(all_hits) <- unlist(DataFrameList(lapply(c(hits, hits_rev), mcols)))
    mcols(all_hits)$motif <- names(all_hits)
    all_hits <- unname(all_hits)
    ## Sort by decreasing relative score to enable selecting the highest scoring
    max_scores <- vapply(cl, maxScore, numeric(1))
    prop <- mcols(all_hits)$score / max_scores[mcols(all_hits)$motif]
    mcols(all_hits)$prop <- prop
    all_hits <- all_hits[order(prop, decreasing = TRUE)]
    merged_hits <- IRanges::reduce(all_hits) # Will be in positional order
    if (counts_only) return(length(merged_hits))

    ## For overlapping hits, choose the one with highest relative score (first)
    ol <- findOverlaps(all_hits, merged_hits)
    best <- queryHits(ol)[!duplicated(subjectHits(ol))]
    final_hits <- sort(all_hits[best])

    ## Map back to the original Views
    map <- as.data.frame(views@ranges)
    hits_to_map <- findInterval(start(final_hits), map$start)
    w <- width(final_hits)

    ## Form the output
    out <- mcols(final_hits)
    out$direction <- factor(out$direction, levels = c("F", "R"))
    out$seq <- hits_to_map
    out$start <- as.integer(start(final_hits) - c(0, map$end)[hits_to_map])
    out$end <- as.integer(out$start + w - 1)
    out$seq_width <- width(views[out$seq])
    out$from_centre <- (out$start + out$end - out$seq_width) / 2

    ## The match itself
    to_rev <- out$direction == "R"
    i <- map$start[out$seq] - 1
    seq <- unlist(unname(views))
    out$match <- as(
        Views(seq, start = i + out$start, end = i + out$end), "XStringSet"
    )
    out$match[to_rev] <- reverseComplement(out$match[to_rev])

    ## Setup any named strings to appear in the same order
    if (nm_type == "character") {
        out$seq <- factor(map$names[hits_to_map], map$names)
        out$seq <- droplevels(out$seq)
    }

    ## The final object
    if (best_only) out <- .getBestMatch(out, break_ties, score_col = "prop")
    o <- order(out$seq, out$start)
    if (nm_type == "character") out$seq <- as.character(out$seq)
    out[o, cols]

}

#' @import Biostrings
#' @keywords internal
.countSinglePwmMatches <- function(
        pwm, views, rc = TRUE, min_score = "80%", ...
){
    ## Checks & the map
    pwm <- .checkPWM(pwm)

    # Form the entire XStringSetList into a Views object
    n_matches <- countPWM(pwm, views, min.score = min_score, ...)
    if (rc)
        n_matches <- c(
            n_matches,
            countPWM(reverseComplement(pwm), views, min.score = min_score, ...)
        )

    as.integer(sum(n_matches))
}

cl <- list(A = ex_pfm[1], B = ex_pfm[2:3], C = ex_pfm[4:5])

test_that("countClusterMatches returnes expected values", {
    counts <- countClusterMatches(cl, ar_er_seq)
    expect_true(is(counts, "integer"))
    expect_equal(names(counts), c("A", "B", "C"))
    expect_null(countClusterMatches(cl[0], ar_er_seq))

    counts <- countClusterMatches(ex_pfm, ar_er_seq)
    expect_equal(length(counts), 1L)

    expect_error(
        countClusterMatches(list("cat"), ar_er_seq), "Could not determine.+"
    )

})

test_that("getClusterMatches returns expected values", {
    res <- getClusterMatches(cl, ar_er_seq)
    expect_true(is(res, "list"))
    expect_true(all(vapply(res, is, integer(1), "DataFrame")))
    expect_equal(names(res), c("A", "B", "C"))
    expect_equal(getClusterMatches(cl[0], ar_er_seq), list())

    res <- getClusterMatches(ex_pfm, ar_er_seq)
    expect_true(is(res, "DataFrame"))
    nm <- c(
        "seq", "score", "direction", "start", "end", "from_centre",
        "seq_width", "motif", "match"
    )
    expect_equal(colnames(res), nm)

    names(ar_er_seq) <- paste("Seq", seq_along(ar_er_seq))
    best <- getClusterMatches(cl, ar_er_seq, best_only = TRUE)
    expect_true(is(best, "list"))
    expect_true(all(vapply(best, is, integer(1), "DataFrame")))
    expect_true(is(best$A$seq, "character"))

    expect_message(
        getClusterMatches(list("cat"), ar_er_seq),
        "Could not determine clusters"
    )

    empty <- getClusterMatches(ex_pfm, ar_er_seq[0])
    expect_equal(colnames(empty), nm)
    expect_equal(nrow(empty), 0)

})

#' motifTestR: Perform Key Analyses on Transcription Factor Binding Motifs
#'
#' The package `motifTestR` has been designed for two primary analyses of TFBMs,
#' testing for positional bias and overall enrichment.
#'
#' The package `motifTestR` provides two primary functions for testing TFBMs
#' within a set of sequences
#'
#' * [testMotifPos()] for detecting positional bias within a set of test
#' sequences
#' * [testMotifEnrich()] for testing overall enrichment of a TFBM within a set
#' of test sequences
#'
#' Motifs are also able to be clustered for analysis as a cluster, or for
#' grouping results.
#' Clusters from external approaches can also be incorporated.
#'
#' * [testClusterPos()] for detecting positional bias for matches to any motif
#' annotated to a cluster, within a set of test sequences
#' * [testClusterEnrich()] for testing overall enrichment of any TFBM annotated
#' to a cluster, within a set of test sequences
#'
#' The main functions rely on lower-level functions such as:
#'
#' * [countPwmMatches()] simply counts the number of matches within an
#' `XStringSet`
#' * [getPwmMatches()] returns the position of matches within an `XStringSet`
#' * [countClusterMatches()] simply counts the number of matches to motifs
#' annotated to a cluster within an `XStringSet`
#' * [getClusterMatches()] returns the position of matches to motifs annotated
#' to a cluster within an `XStringSet`
#' * [makeRMRanges()] which produces a set of random, matching ranges based on
#' key characteristics of the set of test sequences/ranges
#'
#' A simple utility function is provided to enable visualisation of results
#'
#' * [plotMatchPos()] enables visualisation of the matches within a set of
#' sequences using multiple strategies
#'
#' @author
#' Stevie Pederson
#'
#' @keywords internal
"_PACKAGE"

#' @title Find all PWM matches within an XStringSet
#'
#' @description
#' Find all PWM matches within a set of sequences
#'
#' @details
#' Taking a set of sequences as an XStringSet, find all matches above the
#' supplied score (i.e. threshold) for a single Position Weight Matrix (PWM),
#' generally representing a transcription factor binding motif.
#' By default, matches are performed using the PWM as provided and the reverse
#' complement, however this can easily be disabled by setting `rc = FALSE`.
#'
#' The function relies heavily on \link[Biostrings]{matchPWM} and
#' \link[IRanges]{Views} for speed.
#'
#' When choosing to return the best match (`best_only = TRUE`), only the match
#' with the highest score is returned for each sequence.
#' Should there be tied scores, the best match can be chosen as either the first,
#' last, most central, all tied matches, or choosing one at random (the default).
#'
#' @param pwm A Position Weight Matrix, list of PWMs or universalmotif list
#' @param stringset An XStringSet
#' @param rc logical(1) Also find matches using the reverse complement of pwm
#' @param min_score The minimum score to return a match
#' @param best_only logical(1) Only return the best match
#' @param break_ties Method for breaking ties when only returning the best match
#' Ignored when all matches are returned (the default)
#' @param mc.cores Passed to \link[parallel]{mclapply} if passing multiple PWMs
#' @param ... Passed to \link[Biostrings]{matchPWM}
#'
#' @return A DataFrame with columns: `seq`, `score`, `direction`, `start`,
#' `end`, `from_centre`, `seq_width`, and `match`
#'
#' The first three columns describe the sequence with matches, the score of
#' the match and whether the match was found using the forward or reverse PWM.
#' The columns `start`, `end` and `width` describe the where the match was found
#' in the sequence, whilst `from_centre` defines the distance between the centre
#' of the match and the centre of the sequence being queried.
#' The final column contains the matching fragment of the sequence as an
#' `XStringSet`.
#'
#' When passing a list of PWMs, a list of the above DataFrames will be returned.
#'
#' @examples
#' ## Load the example PWM
#' data("ex_pfm")
#' esr1 <- ex_pfm$ESR1
#'
#' ## Load the example Peaks
#' data("ar_er_seq")
#'
#' ## Return all matches
#' getPwmMatches(esr1, ar_er_seq)
#'
#' ## Just the best match
#' getPwmMatches(esr1, ar_er_seq, best_only = TRUE)
#'
#' ## Apply multiple PWMs as a list
#' getPwmMatches(ex_pfm, ar_er_seq, best_only = TRUE)
#'
#' @importFrom parallel mclapply
#' @importFrom methods slot
#' @importClassesFrom universalmotif universalmotif
#' @export
getPwmMatches <- function(
        pwm, stringset, rc = TRUE, min_score = "80%", best_only = FALSE,
        break_ties = c("all", "random", "first", "last", "central"),
        mc.cores = 1, ...
) {

    break_ties <- match.arg(break_ties)
    args <- c(as.list(environment()), list(...))
    args <- args[names(args) != "mc.cores"]
    nm_type <- "integer"
    if (!is.null(names(stringset))) nm_type <- "character"

    # Form the entire XStringSetList into a Views object
    map <- .viewMapFromXStringset(stringset)
    views <- Views(
        unlist(stringset), start = map$start, width = map$width,
        names = map$names
    )

    if (is.list(pwm)) {
        pwm <- .cleanMotifList(pwm)
        out <- mclapply(
            pwm, .getSinglePwmMatches, views = views, rc = rc,
            min_score = min_score, best_only = best_only,
            break_ties = break_ties, nm_type = nm_type, mc.cores = mc.cores
        )
    } else {
        args$nm_type <- nm_type
        args$stringset <- NULL
        args$views <- views
        out <- do.call(".getSinglePwmMatches", args)
    }
    out

}

#' @import Biostrings
#' @importFrom IRanges Views IRanges
#' @importFrom S4Vectors DataFrame mcols mcols<-
#' @importFrom methods as is
#' @importFrom stats setNames
#' @keywords internal
.getSinglePwmMatches <- function(
        pwm, views, rc, min_score, best_only = FALSE, break_ties, nm_type,
        ...
){

    ## Checks
    pwm <- .checkPWM(pwm)
    stopifnot(is(views, "XStringViews"))
    ## Handle empty stringsets
    if (!length(views)) return(.emptyPwmDF(nm_type))

    hits <- matchPWM(pwm, views, min.score = min_score, with.score = TRUE)
    mcols(hits)$direction <- rep_len("F", length(hits))
    if (rc) {
        rev_pwm <- reverseComplement(pwm)
        hits_rev <- matchPWM(
            rev_pwm, views, min.score = min_score, with.score = TRUE
        )
        mcols(hits_rev)$direction <- rep_len("R", length(hits_rev))
        hits <- c(hits, hits_rev)
        hits <- hits[!duplicated(IRanges(hits))] # Remove strict palindromic hits
    }

    ## Setup the return object as empty for any cases without matches
    if (length(hits) == 0) return(.emptyPwmDF(nm_type))

    ## Form it as a list, the eventually wrap into a DF
    cols <- c(
        "seq", "score", "direction", "start", "end", "from_centre",
        "seq_width", "match"
    )
    out <- mcols(hits)
    out$direction <- factor(out$direction, levels = c("F", "R"))

    ## Map back to the original Views
    map <- as.data.frame(views@ranges)
    hits_to_map <- findInterval(start(hits), map$start)
    w <- width(hits)

    ## Form the output
    out$seq <- hits_to_map
    out$start <- as.integer(start(hits) - c(0, map$end)[hits_to_map])
    out$end <- as.integer(out$start + w - 1)
    out$seq_width <- width(views[out$seq])
    out$from_centre <- (out$start + out$end - out$seq_width) / 2

    ## The match itself
    to_rev <- out$direction == "R"
    out$match <- as(hits, "XStringSet")
    out$match[to_rev] <- reverseComplement(out$match[to_rev])

    ## Setup any named strings to appear in the same order
    if (nm_type == "character") {
        out$seq <- factor(map$names[hits_to_map], map$names)
        out$seq <- droplevels(out$seq)
    }

    ## The final object
    if (best_only) out <- .getBestMatch(out, break_ties)
    o <- order(out$seq, out$start)
    if (nm_type == "character") out$seq <- as.character(out$seq)
    out[o, cols]

}

#' @importFrom S4Vectors splitAsList
#' @keywords internal
.getBestMatch <- function(df, ties, score_col = "score"){

    if (nrow(df) == 0) return(df)

    ## Decide how to break ties
    pos <- sample.int(nrow(df), nrow(df)) # Default to random
    if (ties == "first") pos <- df$start
    if (ties == "last") pos <- (-1) * df$end
    if (ties == "central") pos <- abs(df$from_centre)

    ## Pick the best match using the values we have
    if (ties != "all") {
        o <- order(df$seq, -df[[score_col]], pos)
        df <- df[o,]
        df <- df[!duplicated(df$seq), ]
    } else {
        split_scores <- split(df[[score_col]], cumsum(!duplicated(df$seq)))
        keep <- as.logical(unlist(lapply(split_scores, \(x) x == max(x))))
        df <- df[keep,]
    }
    df

}

#' @importFrom S4Vectors DataFrame
#' @keywords internal
.emptyPwmDF <- function(nm_type){
    DataFrame(
        seq = vector(nm_type, 0),
        score = numeric(),
        direction = factor(NULL, levels = c("F", "R")),
        start = integer(),
        end = integer(),
        from_centre = numeric(),
        seq_width = integer(),
        match = DNAStringSet()
    )
}

bm <- getPwmMatches(ex_pfm[1:2], seq, best_only = TRUE, break_ties = "all")

test_that("Defaults work", {
    p <- plotMatchPos(bm, se = FALSE)
    expect_true(is(p@layers[[1]]$geom, "GeomSmooth"))
    expect_equal(levels(p@data$name), names(bm))
    expect_equal(
        vapply(p@mapping, as_label, character(1)),
        c(x = "bin_centre", y = "p", colour = "name")
    )
    expect_true(!p@layers[[1]]$geom_params$se)
})

test_that("cdf plots work", {
    p <- plotMatchPos(bm, abs = TRUE, type = "cdf", geom = "line")
    expect_true(is(p@layers[[1]]$geom, "GeomLine"))
    expect_equal(
        vapply(p@mapping, as_label, character(1)),
        c(x = "bin_start", y = "cdf", colour = "name")
    )
})

test_that("col plots work", {
    p <- plotMatchPos(bm, abs = TRUE, geom = "col", position = "dodge")
    expect_true(is(p@layers[[1]]$geom, "GeomCol"))
    expect_true(is(p@layers[[1]]$position, "PositionDodge"))
    expect_equal(
        vapply(p@mapping, as_label, character(1)),
        c(x = "bin_start", y = "p", fill = "name")
    )
})

test_that("heatmaps work", {
    p <- plotMatchPos(bm, type = "heat")
    expect_true(is(p@layers[[1]]$geom, "GeomTile"))
})
test_that("clusters work", {
    p <- plotMatchPos(bm, type = "heat", cluster = TRUE)
    expect_true(is(p, "patchwork"))
    expect_true(is(p[[1]]@layers[[1]]$geom, "GeomSegment"))
    expect_equal(colnames(p[[1]]$data), c("x", "y", "xend", "yend"))
    expect_true(is(p[[2]]@layers[[1]]$geom, "GeomTile"))
    expect_equal(
        lapply(p[[2]]@mapping, as_label),
        list(x = "bin_centre", y = "name", fill = "p")
    )
    expect_null(p[[2]]@labels$y)
})

test_that("single input works", {
    p <- plotMatchPos(bm[[1]])
    expect_true(is(p@layers[[1]]$geom, "GeomSmooth"))
    expect_equal(
        lapply(p@mapping, as_label),
        list(x = "bin_centre", y = "p", colour = "NULL")
    )
})

test_that("errors are as expected",{
    expect_error(plotMatchPos(bm[[1]], type = "heat"))
    expect_error(plotMatchPos(bm, type = "heat", heat_fill = scale_fill_discrete()))
    expect_error(plotMatchPos(bm[c(1, 1)]))
})

exp_cols <- c(
    "seq", "score", "direction", "start", "end", "from_centre", "seq_width", "match"
)

test_that("getPwmMatches works as expected",{
    all <- getPwmMatches(esr1, stringset)
    fwd_only <- getPwmMatches(esr1, stringset, rc = FALSE)
    expect_true(is(all, "DataFrame"))
    expect_true(nrow(all) == 1)
    expect_true(nrow(fwd_only) == 0)
    expect_equal(colnames(all), colnames(fwd_only))
    expect_equal(colnames(all), exp_cols)
})

test_that("No matches are returned correctly", {
    no_hit <- DNAStringSet(c(no_hit = stringset[[1]][41:100]))
    empty <- getPwmMatches(esr1, no_hit, rc = FALSE)
    expect_true(nrow(empty) == 0)
    expect_equal(colnames(empty), exp_cols)
})

test_that("Choosing only the best matches works", {
    ## Test the default first
    stringset <- ar_er_seq[2]
    best <- getPwmMatches(esr1, stringset, best_only = TRUE)
    expect_true(nrow(best) == 1)
    expect_true(best$start == 99)
    ## Now modify & check other options in the private function
    all <- getPwmMatches(esr1, stringset)
    all$score <- 5 # Bodge up a tie
    expect_true(.getBestMatch(all, "first")$start == 99)
    expect_true(.getBestMatch(all, "central")$start == 186)
    expect_true(.getBestMatch(all, "last")$start == 186)
    expect_equal(nrow(.getBestMatch(all, "all")), nrow(all))

})

test_that("Passing a list also produces a list",{
    multi <- getPwmMatches(ex_pfm[1:2], stringset)
    expect_true(is(multi, "list"))
    expect_true(length(multi) == 2)
    expect_true(all(vapply(multi, is, logical(1), "DataFrame")))
})

test_that("Fails are correct",{
    expect_error(getPwmMatches(letters, stringset))
    expect_error(getPwmMatches(esr1, stringset[[1]]))
    expect_true(nrow(.getBestMatch(data.frame(start = integer()))) == 0)
})

#' Form a set of masked regions for hg19 using telomeres & centromeres
library(AnnotationHub)
ah <- AnnotationHub()
centro <- granges(ah[["AH107360"]])
centro$type <- "centromere"
telo <- ah[["AH107361"]]
telo <- trim(telo)
telo <- granges(telo)
telo$type <- "telomere"
hg19_mask <- sort(c(centro, telo))
names(hg19_mask) <- c()

#' Now load the N.bed regions from the 2bit file.
#' These were produced on an HPC by taking the hard-masked fasta file from UCSC
#' converting to 2bit and extracting the regions with Ns
#'
#' faToTwoBit hg19.fa.masked hg19.fa.masked.2bit
#' twoBitInfo -nBed hg19.fa.masked.2bit N.bed
#' gzip N.bed
library(rtracklayer)
masks <- import.bed("inst/scripts/N.bed.gz") # Hidden in .gitignore & .Rbuildignore
masks <- subset(masks, seqnames %in% seqlevels(hg19_mask))
masks <- sortSeqlevels(masks)
masks <- sort(masks)
seqlevels(masks) <- seqnames(seqinfo(hg19_mask))
seqinfo(masks) <- seqinfo(hg19_mask)
masks$type <- 'N-content'

#' Get all the sequences corresponding to these masks from the actual BSgenome
library(BSgenome.Hsapiens.UCSC.hg19)
seq <- getSeq(BSgenome.Hsapiens.UCSC.hg19, masks)
n_probs <- letterFrequency(seq, letters = "N", as.prob = TRUE)
#" Only retain those with any Ns at all
hg19_mask <- c(hg19_mask, masks[n_probs > 0]) |>
  sort() |>
  reduce()
save(hg19_mask, file = "data/hg19_mask.RData")

#' @title Example Position Frequency Matrices
#'
#' @description Example Position Frequency Matrices
#'
#' @details
#' This object contains 5 PFMs taken from HOCOMOCOv11-coreA for examples and testing
#'
#' Generation of this motif list is documented in
#' `system.file("scripts/ex_pfm.R", package = "motifTestR")`
#'
#' @usage data("ex_pfm")
#'
#' @examples
#' data("ex_pfm")
#' ex_pfm$ESR1
#'
#' @name ex_pfm
#' @rdname ex_pfm
"ex_pfm"

#' @title A set of peaks with AR and ER detected
#'
#' @description
#' A set of ChIP-Seq peaks where AR and ER were both detected
#'
#' @details
#' The subset of peaks found on chr1 which contained signal from at least
#' two of AR, ER and H3K27ac, taken from GSE123767.
#' Peaks were resized to a uniform width of 400bp after downloading
#'
#' Generation of these ranges is documented in
#' `system.file("scripts/ar_er_peaks.R", package = "motifTestR")`
#'
#' @source \url{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE123767}
#'
#' @usage data("ar_er_peaks")
#'
#' @examples
#' data("ar_er_peaks")
#' ar_er_peaks
#' @name ar_er_peaks
#' @rdname ar_er_peaks
"ar_er_peaks"

#' @title Sequences from peaks with AR and ER detected
#'
#' @description
#' The genomic sequences obtained from the ar_er_peaks
#'
#' @details
#' These sequences represent the sequences obtained from
#' BSgenome.Hsapiens.UCSC.hg19 for thw peaks supplied as `ar_er_peaks`
#'
#' Generation of these sequences is documented in
#' `system.file("scripts/ar_er_peaks.R", package = "motifTestR")`
#'
#' @usage data("ar_er_seq")
#'
#' @examples
#' data("ar_er_seq")
#' ar_er_seq
#' @name ar_er_seq
#' @rdname ar_er_seq
"ar_er_seq"

#' @title Regions from hg19 with high N content
#'
#' @description
#' A GRanges object with regions annotated as telomeres or centromeres
#'
#' @details
#' The regions defined as centromeres or telomeres in hg19, taken from
#' AnnotationHub objects "AH107360" and "AH107361". These were combined with
#' regions containing Ns from the UCSC 2bit file, and regions with Ns in the
#' BSgenome.Hsapiens.UCSC.hg19 were retained.
#'
#' Generation of these ranges is documented in
#' `system.file("scripts/hg19_mask.R", package = "motifTestR")`
#'
#' @source The package AnnotationHub and
#' \url{https://hgdownload.cse.ucsc.edu/goldenpath/hg19/bigZips/hg19.fa.masked.gz}
#'
#' @usage data("hg19_mask")
#' @examples
#' data("hg19_mask")
#' hg19_mask
#' @name hg19_mask
#' @rdname hg19_mask
"hg19_mask"

#' @title Candidate Enhancer Regions from ZR-75-1 Cells
#'
#' @description
#' The chr1 subset of candidate enhancers for ZR-75-1 cells
#'
#' @details
#' These enhancers are the chr1 subset of enhancer regions for ZR-75-1 cells as
#' identified by EnhancerAtlas 2.0
#'
#' #' Generation of these ranges is documented in
#' `system.file("scripts/zr75_enh.R", package = "motifTestR")`
#'
#'
#' @source \url{http://www.enhanceratlas.org/index.php}
#'
#' @usage data("zr75_enh")
#' @examples
#' data("zr75_enh")
#' zr75_enh
#' @name zr75_enh
#' @rdname zr75_enh
"zr75_enh"

library(rtracklayer)
library(extraChIPs)
sq <- defineSeqinfo("GRCh37", chr = TRUE)
genome(sq) <- "hg19"
enh <- import.bed("http://www.enhanceratlas.org/data/download/enhancer/hs/ZR75-1.bed")
enh <- granges(enh)
seqlevels(enh) <- seqnames(sq)
seqinfo(enh) <- sq
zr75_enh <- subset(enh, seqnames == "chr1")

save(zr75_enh, file = "data/zr75_enh.RData")

#' @title Simulate sequences using optional TFBMs
#'
#' @description
#' Simulate a set of fixed-width sequences using optional TFBMs
#'
#' @details
#' Using the nucleotide and probabilities provided as set of sequences can be
#' simulated. By default, this will effectively be a set of 'background'
#' sequences, with letters effectively chosen at random.
#'
#' If a PWM/PFM is supplied, the shape parameters are first passed to
#' \link[VGAM]{rbetabinom.ab} to determine the random positions the motif will
#' be placed, with the default parameters representing a discrete uniform
#' distribution.
#'
#' The sequences to have a motif inserted will be selected, along with the
#' number of motifs, using the rate and theta parameters.
#' If both are NULL, every sequence will have a single motif inserted.
#' If the rate is > 0 and theta is NULL, sequences will be selected to have
#' motifs inserted using a poisson distribution.
#' If theta is also provided, sequences will be selected to contain motifs
#' using a negative binomial distribution, noting that smaller values of theta
#' lead to higher over-dispersion
#'
#' Once positions and sequences for the TFBM have been selected, nucleotides
#' will be randomly sampled using the probabilities provided in the PWM and
#' these motifs will be placed at the randomly sampled positions
#'
#' @return
#' By default a DNAStringSet will be returned.
#' If possible, the position of any randomly sampled motifs will be included
#' in the mcols element of the returned object.
#'
#' @param n The number of sequences to simulate
#' @param width Width of sequences to simulate
#' @param pfm Probability Weight/Frequency Matrix
#' @param nt Nucleotides to include
#' @param prob Sampling probabilities for each nucleotide
#' @param shape1,shape2 Passed to \link[VGAM]{rbetabinom.ab}
#' @param rate The expected rate of motifs per sequence. Is equivalent to
#' \eqn{ \lambda } in \link[stats]{rpois}. If set to NULL or NA, all sequences
#' will be simulated with a single motif, otherwise a Poisson distribution will
#' be used
#' @param theta Overdispersion parameter passed to \link[MASS]{rnegbin}.
#' If set to NULL or NA, the rate parameter will be passed to \link[stats]{rpois}.
#' However if this value is set, the rate and theta parameters are passed to
#' \link[MASS]{rnegbin} to simulate overdispersed counts
#' @param as ObjectClass to return objects as. Defaults to DNAStringSet, but
#' other viable options may include 'character', 'CharacterList' or any
#' other class from which a character vector may be coerced.
#' @param ... Not used
#'
#' @examples
#' ## Randomly generate 10x50nt sequences without any TFBMs present
#' simSeq(10, 50)
#'
#' ## Now place a motif at random positions
#' data('ex_pfm')
#' sim_seq <- simSeq(10, width = 20, pfm = ex_pfm$ESR1)
#' sim_seq
#' ## The position of the motif within each sequence is included in the mcols
#' mcols(sim_seq)
#'
#' ## Use this to extract the random motifs from the random sequences
#' library(IRanges)
#' i <- mcols(sim_seq)$pos + cumsum(width(sim_seq)) - width(sim_seq)
#' Views(unlist(sim_seq), start = i, width = 10)
#'
#'
#' @importFrom S4Vectors mcols<- DataFrame endoapply
#' @importFrom IRanges IntegerList
#' @importFrom methods slot is
#' @export
simSeq <- function(
        n, width, pfm = NULL, nt = c("A", "C", "G", "T"), prob = rep(0.25, 4),
        shape1 = 1, shape2 = shape1, rate = NULL, theta = NULL, as = "DNAStringSet",
        ...
){

    ## Assuming an even frequency, create a vector randomly
    prob <- rep_len(prob, length(nt))
    bg <- sample(nt, n * width, replace = TRUE, prob = prob)
    seq_starts <- seq(1, n*width, by = width) # Where each sequence starts

    ## If a PWM is provided, now sample using the motifs
    if (!is.null(pfm)) {

        ## Check we have PFMs, not PWMs or any other format
        if (is(pfm, "universalmotif")) pfm <- slot(pfm, "motif")
        col_sums <- colSums(pfm)
        is_pfm <- isTRUE(
            all.equal(unname(col_sums), rep(1, length(col_sums)))
        )
        if (!is_pfm) {
            ## Convert to a PWM, then to PFM using existing checks
            pwm <- .checkPWM(pfm)
            pfm <- slot(create_motif(pwm, type = "PPM"), "motif")
        }
        stopifnot(all(rownames(pfm) %in% nt)) # Same alphabet

        pfm_width <- ncol(pfm)
        stopifnot(pfm_width <= width)
        ## Determine the valid positions given the width of the PFM.
        max_start <- width - pfm_width

        ## We really need to just choose the positions using the
        ## different distributions. Everything else can follow.
        pos <- .samplePos(n, rate, theta, shape1, shape2, max_start, seq_starts)
        vec_pos <- vapply(
            pos, \(i) seq(i, length.out = pfm_width), numeric(pfm_width)
        )
        vec_pos <- as.integer(vec_pos)

        ## Sample the random motifs as a matrix, then coerce to a vector
        rnd_mot <- replicate(
            length(pos),
            apply(pfm, MARGIN = 2, FUN = \(p) sample(nt, 1, prob = p))
        )

        ## Handle any overlapping positions by treating them as duplicates
        ## This will result in some partial motifs being placed
        dups <- duplicated(vec_pos)
        bg[vec_pos[!dups]] <- as.character(rnd_mot)[!dups]

    }

    seq <- apply(matrix(bg, ncol = n), MARGIN = 2, paste, collapse = "")
    seq <- as(seq, as)
    if (is(seq, "Vector") & !is.null(pfm)) {

        which_seq <- ceiling(pos / width)
        pos_list <- IntegerList(vector("list", n))
        pos_list[unique(which_seq)] <- split(pos %% width, f = which_seq)
        pos_list <- endoapply(endoapply(pos_list, unique), sort)
        ## Now form any mcols
        n_motifs <- vapply(pos_list, length, integer(1))
        if (all(n_motifs == 1)) pos_list <- unlist(pos_list)
        mcols(seq) <- DataFrame(pos = pos_list, n_motifs = n_motifs)
    }
    seq

}

#' @keywords internal
.samplePos <- function(n, rate, theta, shape1, shape2, max_start, seq_starts) {

    if (!requireNamespace('VGAM', quietly = TRUE))
        stop("Please install 'VGAM' to insert TFBMs into the sequences.")

    if (is.null(rate) | isTRUE(is.na(rate))) {
        pos <- VGAM::rbetabinom.ab(n, max_start, shape1, shape2) + seq_starts
    } else {
        stopifnot(rate > 0)
        if (is.null(theta) | isTRUE(is.na(theta))) {
            pos <- .simPoisSeq(n, rate, shape1, shape2, max_start, seq_starts)
        } else {
            stopifnot(theta > 0)
            pos <- .simNBSeq(n, rate, theta, shape1, shape2, max_start, seq_starts)
        }
    }
}

#' @keywords internal
.simPoisSeq <- function(n, rate, shape1, shape2, max_start, seq_starts) {

    ## Unlike simulating for all sequences, we need a list here.
    ## First get the number of motifs per sequences
    n_pois <- stats::rpois(n, rate)
    which_seq <- which(n_pois > 0)
    pos <- lapply(
        n_pois[which_seq],
        \(x) VGAM::rbetabinom.ab(x, max_start, shape1, shape2)
    )

    ## Expand to the global position with the random sequence
    seq_starts <- seq_starts[which_seq]
    motif_starts <- lapply(
        seq_along(seq_starts), \(i) seq_starts[i] + pos[[i]] ## may be variable length
    )
    unlist(motif_starts)

}

#' @keywords internal
.simNBSeq <- function(n, rate, theta, shape1, shape2, max_start, seq_starts) {

    if (!requireNamespace('MASS', quietly = TRUE))
        stop("Please install 'MASS' to inject over-dispersed TFBMs into the sequences.")

    ## Unlike simulating for all sequences, we need a list here.
    ## First get the number of motifs per sequences
    n_negbin <- MASS::rnegbin(n, rate, theta)
    which_seq <- which(n_negbin > 0)
    pos <- lapply(
        n_negbin[which_seq],
        \(x) VGAM::rbetabinom.ab(x, max_start, shape1, shape2)
    )

    ## Expand to the global position with the random sequence
    seq_starts <- seq_starts[which_seq]
    motif_starts <- lapply(
        seq_along(seq_starts), \(i) seq_starts[i] + pos[[i]] ## may be variable length
    )
    unlist(motif_starts)

}

pfm <- matrix(0, ncol = 4, nrow = 4)
diag(pfm) <- 1
rownames(pfm) <- c("A", "C", "G", "T")
um <- universalmotif::create_motif(pfm, type = "PWM", pseudocount = 0.01)

test_that("Basic sequence simulation works", {
    sim_seq <- simSeq(10, 4, pfm)
    expect_true(is(sim_seq, "DNAStringSet"))
    expect_equal(mcols(sim_seq)$pos, rep_len(1, 10))
    expect_equal(unique(as.character(sim_seq)), paste(rownames(pfm), collapse = ""))

    sim_seq <- simSeq(10, 4, um, as = "character")
    expect_true(mean(sim_seq ==  paste(rownames(pfm), collapse = "")) > 0.5)
    expect_true(is(sim_seq, "character"))
})

test_that("Poisson simulation works", {
    set.seed(1000)
    sim_seq <- simSeq(10, 5, pfm, rate = 0.5)
    expect_true(is(mcols(sim_seq)$pos, "IntegerList"))
    expect_true(is(mcols(sim_seq)$n_motifs, "integer"))
    ## For a width of 5, can only be pos 1 or 2
    expect_true(max(unlist(mcols(sim_seq)$pos)) <= 2)

})

test_that("NegBin simulation works", {
    set.seed(1000)
    sim_seq <- simSeq(10, 5, pfm, rate = 1, theta = 1)
    expect_true(is(mcols(sim_seq)$pos, "IntegerList"))
})

test_that("simSeq errors where expected", {
    expect_error(simSeq(1, 3, pfm))
    expect_error(simSeq(1, 4, pfm, rate = -1))
    expect_error(simSeq(1, 4, pfm, rate = 1, theta = -1))
})

#' @title Plot Motif Match Positions
#'
#' @description
#' Plot the distribution of motif matches across sequences
#'
#' @details
#' Multiple options are provided for showing the distribution of PWM matches
#' within a set of sequences, using either the smoothed probability density,
#' the probability CDF or as a heatmap.
#' Distances can be shown as symmetrical around the centre or using absolute
#' distances from the central position within the sequences.
#'
#' Heatmaps are only enabled for comparisons across multiple PWMs, with
#' optional clustering enabled. If adding a dendrogram for clustering,
#' the returned plot object will be a patchwork object.
#'
#' @return A ggplot2 object
#'
#'
#' @param matches Output from \link{getPwmMatches}
#' @param binwidth Width of bins to use when plotting
#' @param abs logical(1) Plot absolute distances from centre
#' @param use_totals logical(1). If TRUE, plots will use total counts. The
#' default (FALSE) plots probabilities.
#' @param type Plot match density, the CDF or a binned heatmap
#' @param geom Type of geom to be used for line plots. Ignored for heatmaps
#' @param cluster logical(1) Cluster motifs when drawing a heatmap. If TRUE a
#' dendrogram will be added to the LHS of the plot
#' @param w Relative width of the dendrogram on (0, 1)
#' @param heat_fill scale_fill_continuous object for heatmaps. If not provided,
#' scale_fill_viridis_c() will be added to the heatmap.
#' @param ... Passed to individual geom* functions
#'
#' @examples
#' ## Load the example PWM
#' data("ex_pfm")
#' esr1 <- ex_pfm$ESR1
#'
#' ## Load the example sequences from the peaks
#' data("ar_er_seq")
#'
#' ## Just the best match
#' bm <- getPwmMatches(esr1, ar_er_seq, best_only = TRUE)
#' plotMatchPos(bm, se = FALSE)
#'
#' ## Matches can also be shown by distance from centre
#' plotMatchPos(bm, abs = TRUE)
#'
#' ## Cumulative Probability plots are also implemented
#' plotMatchPos(bm, type = "cdf", geom = "line", colour = "red") +
#'   geom_abline(intercept = 0.5, slope = 1/ 400)
#'
#' @importFrom S4Vectors splitAsList
#' @importFrom rlang !! sym
#' @import ggplot2
#' @export
plotMatchPos <- function(
        matches, binwidth = 10, abs = FALSE, use_totals = FALSE,
        type = c("density", "cdf", "heatmap"),
        geom = c("smooth", "line", "point", "col"),
        cluster = FALSE, w = 0.1, heat_fill = NULL, ...
){
    .checkMatches(matches)
    type <- match.arg(type)
    if (!is(matches, "DataFrame")) {
        if (is.null(names(matches))) names(matches) <- seq_along(matches)
        nm <- names(matches)
        if (any(duplicated(nm)))
            stop("List elements in matches must have unique names")
        binned <- lapply(matches, .makeBmBins, binwidth = binwidth, abs = abs)
        bins <- lapply(binned, \(x){
            vapply(splitAsList(x, x$bin), \(df) sum(df$weight), numeric(1))
        })
        df_list <- lapply(nm, \(x){
            df <- data.frame(
                name = x, bin = names(bins[[x]]), total = as.numeric(bins[[x]])
            )
            df$p <- df$total / sum(df$total)
            df$cdf <- cumsum(df$p)
            df
        })
        df <- do.call("rbind", df_list)
        df$name <- factor(df$name, levels = nm)
        name <- sym("name")

    } else {
        binned <- .makeBmBins(matches, binwidth, abs)
        bins <- vapply(
            splitAsList(binned, binned$bin), \(x) sum(x$weight), numeric(1)
        )
        df <- data.frame(bin = names(bins), total = as.numeric(bins))
        df$p <- df$total / sum(df$total)
        df$cdf <- cumsum(df$p)
        name <- NULL
    }
    df$bin_start <- as.numeric(gsub("^.(.+),.+", "\\1", df$bin))
    df$bin_end <- as.numeric(gsub("^.+,(.+)\\]", "\\1", df$bin))
    df$bin_centre <- (df$bin_end + df$bin_start) / 2
    df$bin <- factor(df$bin, unique(df$bin))
    x <- sym("bin_centre")
    if (abs) x <- sym("bin_start")
    yval <- ifelse(use_totals, "total", "p")

    geom <- paste0("geom_", match.arg(geom))
    geom_fun <- match.fun(geom)
    if (type == "density") {
        plot_aes <- aes({{ x }}, !!sym(yval), colour = {{ name }})
        if (geom == "geom_col") names(plot_aes)[[3]] <- "fill"
        plot <- ggplot(df, plot_aes) + geom_fun(...)
    }
    if (type == "cdf") {
        plot <- ggplot(df, aes({{ x }}, !!sym("cdf"), colour = {{ name }})) +
            geom_fun(...)
    }
    if (type == "heatmap") {
        if (is.null(name)) stop("Heatmaps not inplemented for a single PWM")
        plot <- .createHeatmap(
            df, cluster, yval, w, heat_fill, width = binwidth, abs = abs
        )
    }

    plot

}

#' @importFrom stats hclust dist xtabs as.dendrogram as.formula
#' @importFrom patchwork plot_layout
#' @importFrom rlang !! sym
#' @keywords internal
.createHeatmap <- function(df, cluster, yval, w, fill_scale, abs, ...) {

    levels <- unique(df$name)
    if (is.null(fill_scale)) fill_scale <- scale_fill_viridis_c()
    stopifnot(is(fill_scale, "ScaleContinuous"))
    stopifnot(w > 0 & w < 1)

    ## Create the dendrogram if needed
    if (cluster) {

        if (!requireNamespace('ggdendro', quietly = TRUE))
            stop("Please install 'ggdendro' to use this function.")

        fm <- as.formula(paste(yval, "~ name + bin"))
        mat <- as.matrix(xtabs(fm, df[c("name", "bin", yval)]))
        clust <- hclust(dist(mat), method = "ward.D2")
        dend <- ggdendro::dendro_data(as.dendrogram(clust))
        levels <- dend$labels$label
        ym <- max(dend$segments$xend)
        dend_aes <- aes(
            x = !!sym("yend"), y = !!sym("xend"),
            xend = !!sym("y"), yend = !!sym("x")
        )
        dend_plot <- ggplot(dend$segments) +
            geom_segment(dend_aes) +
            scale_x_reverse(expand = expansion(0)) +
            scale_y_continuous(limits = c(0, ym) + 0.5, expand = expansion(0)) +
            labs(x = "", y = c()) +
            theme_minimal() +
            theme(
                panel.grid = element_blank(),
                axis.text = element_blank(), axis.ticks = element_blank(),
                plot.margin = unit(c(5.5, 0, 5.5, 5.5), "points")
            )
        ylab <- NULL
    }

    xlab <- "bin_centre"
    ylab <- "name"
    if (cluster) ylab <- NULL
    if (abs) xlab <- "bin_start"
    # Always create the heatmap...
    df$name <- factor(df$name, levels = levels)
    plot <- ggplot(
        df, aes(!!sym("bin_centre"), !!sym("name"), fill = !!sym(yval))
    ) +
        geom_tile(...) +
        labs(x = xlab, y = ylab) +
        scale_x_continuous(expand = rep_len(0, 4)) +
        scale_y_discrete(expand = rep_len(0, 4)) +
        fill_scale
    if (!cluster) return(plot)
    dend_plot + plot + plot_layout(widths = c(w, 1 - w))

}

sq <- Seqinfo(seqnames = "chr1", seqlengths = 500, FALSE, "test")
rng <- GRanges(c("chr1:101-150", "chr1:201-250"), seqinfo = sq)
chr <- GRanges("chr1:51-500", seqinfo = sq)
grl <- splitAsList(rng, c("promoter", "other"))
y <- GRangesList(
    promoter = GRanges("chr1:1-200"), other = GRanges("chr1:201-500")
)
seqinfo(y) <- sq

test_that("Basic functions of makeRMRanges work", {
    bg <- makeRMRanges(rng, chr, n_iter = 2)
    expect_true(length(bg) == length(rng) * 2)
    expect_true("iteration" %in% colnames(mcols(bg)))
    expect_true(all(width(bg) == 50))

    bg <- makeRMRanges(rng, chr, n_total = 3)
    expect_true(length(bg) == 3)
    expect_true(length(colnames(mcols(bg))) == 0)

})

test_that("GRangesList behaves as expected", {

    bg <- makeRMRanges(grl, y, n_iter = 2)
    expect_true("iteration" %in% colnames(mcols(bg)))
    nm <- structure(
        c(other = 2L, promoter = 2L), dim = 2L,
        dimnames = structure(list(c("other", "promoter")), names = ""),
        class = "table"
    )
    expect_equal(table(names(bg)), nm)

    bg <- makeRMRanges(grl, y, n_total = c(2, 1))
    nm <- structure(
        c(other = 2L, promoter = 1L), dim = 2L,
        dimnames = structure(list(c("other", "promoter")), names = ""),
        class = "table"
    )
    expect_equal(table(names(bg)), nm)

    bg <- makeRMRanges(grl, y, unlist = FALSE)
    expect_equal(names(bg), c("other", "promoter"))
    expect_true(all(vapply(bg, length, integer(1)) == 1))

})

test_that("makeRMRanges errors as expected", {
    expect_error(makeRMRanges(rng, chr, "chr1:1"))
    expect_error(makeRMRanges(rng, GRanges("chr1:150-120"))) # BG too short
    expect_error(makeRMRanges(rng, GRanges("chr1:1-200"))) # Range 2 excluded
    expect_error(makeRMRanges(grl, y[1]))
})

test_that("splitting entire genomes works", {
    hg19 <- defineSeqinfo("GRCh37")
    genome(hg19) <- "hg19"
    rng <- GRanges(c("chr1:101-150", "chr1:201-250"), seqinfo = hg19)
    rm_ranges <- makeRMRanges(rng, GRanges(hg19), n_total = 2)
    expect_true(length(rm_ranges) == 2)
    expect_equal(dim(mcols(rm_ranges)), c(2, 0))
})

---
title: "Motif Analysis Using motifTestR"
author:
- name: Stevie Pederson
  affiliation: Black Ochre Data Labs, The Kids Research Institute Australia
  email: stephen.pederson.au@gmail.com
package: motifTestR
bibliography: '`r system.file("references.bib", package = "motifTestR")`'
output:
  BiocStyle::html_document:
    toc: true
    toc_depth: 2
abstract: |
  Analysis of transcription factor binding motifs using Position Weight Matrices
  (PWMs) is a common task in analysis of genomic data.
  Two key tests for analysis of TFBMs using morifTestR are demonstrated below
vignette: |
  %\VignetteIndexEntry{Motif Analysis Using motifTestR}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, echo = FALSE}
knitr::opts_chunk$set(message = FALSE, crop = NULL)
```


# Introduction

Bioinformatic analysis of data from ChIP-Seq or ATAC-Seq, along with multiple other approaches, can involve the analysis of sequences within the regions identified is being of interest.
Whilst these analyses are not restricted to transcription factors, this can often form an important component of this type of analysis.
Analysis of Transcription Factor Binding Motifs (TFBMs) is often performed using Position Weight Matrices (PWMs) to encode the flexibility in which exact sequence is bound by the particular transcription factor, and is a computationally demanding task with many popular tools enabling analysis outside of R.

The tools within `motifTestR` aim to build on and expand the existing resources available to the Bioconductor community, performing all analyses inside the R environment.
The package offers two complementary approaches to TFBM analysis within `XStringSet` objects containing multiple sequences.
The function `testMotifPos()` identifies motifs showing *positional bias* within a set of sequences, whilst *overall enrichment* within a set of sequences is enabled by `testMotifEnrich()`.
These are then extended to analyse motifs grouped into "clusters" using `testClusterPos()` and `testClusterEnrich()`.
Additional functions aid in the visualisation and preparation of these two key 
approaches.


# Setup

## Installation

In order to perform the operations in this vignette, first install `motifTestR`.

```r 
if (!"BiocManager" %in% rownames(installed.packages()))
  install.packages("BiocManager")
BiocManager::install("motifTestR")
```

Once installed, we can load all required packages, set a default plotting theme and setup how many threads to use during the analysis.

```{r load-packages}
library(motifTestR)
library(rtracklayer)
library(BSgenome.Hsapiens.UCSC.hg19)
library(parallel)
library(ggplot2)
library(patchwork)
library(universalmotif)
library(extraChIPs)
theme_set(theme_bw())
cores <- 1
```


## Defining a Set of Peaks

```{r load-ar-er-peaks, echo = FALSE}
data("ar_er_peaks")
```


The peaks used in this workflow were obtained from the bed files denoting binding sites of the Androgen Receptor and Estrogen Receptor along with H3K27ac marks, in ZR-75-1 cells under DHT treatment [@Hickey2021-mz].
The object `ar_er_peaks` contains a subset of `r length(ar_er_peaks)` peaks found within chromosome 1, with all peaks resized to 400bp 

```{r load-example-peaks}
data("ar_er_peaks")
ar_er_peaks
sq <- seqinfo(ar_er_peaks)
```

Whilst the example dataset is small for the convenience of an R package, those wishing to work on the complete set of peaks (i.e. not just chromosome 1) may run the code provided in the final section to obtain all peaks.
This will produce a greater number of significant results in subsequent analyses but will also increase running times for all functions.


## Obtaining a Set of Sequences for Testing

Now that we have genomic co-ordinates as a set of peaks, we can obtain the sequences that are associated with each peak.
The source ranges can optionally be added to the sequences as names by coercing the ranges to a character vector.

```{r test-seq}
test_seq <- getSeq(BSgenome.Hsapiens.UCSC.hg19, ar_er_peaks)
names(test_seq) <- as.character(ar_er_peaks)
```

## Obtaining a List of PWMs for Testing

A small list of Position Frequency Matrices (PFMs), obtained from `r Biocpkg("MotifDb")` are provided with the package and these will be suitable for all downstream analysis.
All functions will convert PFMs to PWMs internally.

```{r ex-pwm}
data("ex_pfm")
names(ex_pfm)
ex_pfm$ESR1
```

Again, a larger set of motifs may be obtained using or modifying the example code at the end of the vignette

# Searching Sequences

## Finding PWM Matches

All PWM matches within the test sequences can be returned for any of the PWMs, with `getPwmMatches()` searching using the PWM and it's reverse complement by default.
Matches are returned showing their position within the sequence, as well as the distance from the centre of the sequence and the matching section within the larger sequence.
Whilst there is no strict requirement for sequences of the same width, generally this is good practice for this type of analysis and is commonly a requirement for downstream statistical analysis.

```{r match-esr1}
score_thresh <- "70%"
getPwmMatches(ex_pfm$ESR1, test_seq, min_score = score_thresh)
```

Many sequences will contain multiple matches, and we can subset our results to only the 'best match' by setting `best_only = TRUE`.
The best match is chosen by the highest score returned for each match.
If multiple matches return identical scores, all tied matches are returned by default and will be equally down-weighted during positional analysis.
This can be further controlled by setting `break_ties` to any of c("random", "first", "last", "central"), which will choose randomly, by sequence order or proximity to centre.

```{r best-match-esr1}
getPwmMatches(ex_pfm$ESR1, test_seq, min_score = score_thresh, best_only = TRUE)
```

We can return all matches for a complete list of PWMs, as a list of DataFrame objects.
This strategy allows for visualisation of results as well as testing for positional bias.

```{r bm-all}
bm_all <- getPwmMatches(
  ex_pfm, test_seq, min_score = score_thresh, best_only = TRUE, break_ties = "all",
  mc.cores = cores
)
```

This same strategy of passing a single, or multiple PWMs can be applied even when simply wishing to count the total matches for each PWM.
Counting may be useful for restricting downstream analysis to the set of motifs with more than a given number of matches.

```{r count-pwm-matches}
countPwmMatches(ex_pfm, test_seq, min_score = score_thresh, mc.cores = cores)
```


# Analysis of Positional Bias

## Testing for Positional Bias

A commonly used tool within the MEME-Suite is `centrimo` [@Bailey2012-qz] and `motifTestR` provides a simple, easily interpretable alternative approach using `testMotifPos()`.
This function bins the distances from the centre of each sequence with the central bin being symmetrical around zero, and if no positional bias is expected (i.e. H~0~), matches should be equally distributed between bins.
Unlike `centrimo`, *no assumption of centrality* is made and any notable deviations from a discrete uniform distribution may be considered as significant.

A test within each bin is performed using `binom.test()` and a single, summarised p-value across all bins is returned using the asymptotically exact harmonic mean p-value (HMP) [@Wilson2019-ln].
By default, the binomial test is applied for the null hypothesis to detect matches in each bin which are *greater* than expected, however, this can also be set by the user.
When using the harmonic-mean p-value however, choosing the alternate hypothesis as "greater" tends return a more conservative p-value across the entire set of bins.


```{r test-motif-pos}
res_pos <- testMotifPos(bm_all, mc.cores = cores)
head(res_pos)
```

The bins returned by the function represent the widest range of bins where the raw p-values were below the HMP.
Wide ranges tend to be associated with lower significance for a specific PWM.
This is a further point of divergence from `centrimo` in that results are dependent on the pre-determined bin-size and the region of enrichment is formed using p-values, instead of the adaptive methods of `centrimo`  [@Bailey2012-qz].

Due to the two-stranded nature of DNA, the distance from zero can also be assessed by setting `abs = TRUE` and in this case the first bin *begins* at zero.

```{r test-motif-pos-abs}
res_abs <- testMotifPos(bm_all, abs = TRUE, mc.cores = cores) 
head(res_abs)
```

This approach is particularly helpful for detecting co-located transcription factors which can be any distance from the TF which was used to obtain and centre the test set of sequences.

## Viewing Matches

The complete set of matches returned as a list above can be simply passed to `ggplot2` [@Wickham2016] for visualisation, in order to asses whether any PWM appears to have a positional bias.
By default, smoothed values across all motifs will be overlaid (Figure 1A), however, tailoring using ggplot is simple to produce a wide variety of outputs (Figure 1B)

```{r plot-pos, fig.cap = "Distribution of motif matches around the centres of the set of peaks", fig.alt="Two panel plot with the left panel showing the probability distribution of motif matches for ESR1, ANDR and FOXA1 within the set of test sequences as coloured lines. The right panel shows the counts within each 10bp bin around the centre of the sequences with a loess curve overlaid"}
topMotifs <- res_pos |>
    subset(fdr < 0.05) |>
    rownames()
A <- plotMatchPos(bm_all[topMotifs], binwidth = 10, se = FALSE)
B <- plotMatchPos(
    bm_all[topMotifs], binwidth = 10, geom = "col", use_totals = TRUE
) +
  geom_smooth(se = FALSE, show.legend = FALSE) +
  facet_wrap(~name)
A + B + plot_annotation(tag_levels = "A") & theme(legend.position = "bottom")
```

Whilst the above will produce figures showing the symmetrical distribution around the peak centres, the distance from the peak centre can also be shown as an absolute distance.
In Figure 2 distances shown as a heatmap (A) or as a CDF (B).
The latter makes it easy to see that 50% of ESR1 matches occur within a short distance of the centre (~25bp), whilst for ANDR and FOXA1, this distance is roughly doubled.
Changing the binwidth argument can either smooth data or increase the fine resolution.


```{r plot-abs-pos, fig.cap = "Distribution of motif matches shown as a distance from the centre of each sequence", fig.alt="Two panel plot with the left panel showing the distances from the sequence centres of the best matches for FOXA1, ANDR and ESR1. The right panel shows the CDF of these matches as a line plot"}
topMotifs <- res_abs |>
    subset(fdr < 0.05) |>
    rownames()
A <- plotMatchPos(bm_all[topMotifs], abs = TRUE, type = "heatmap") +
  scale_fill_viridis_c()
B <- plotMatchPos(
  bm_all[topMotifs], abs = TRUE, type = "cdf", geom = "line", binwidth = 5
)
A + B + plot_annotation(tag_levels = "A") & theme(legend.position = "bottom")
```


# Testing For Motif Enrichment

As well as providing methods for analysing positional bias within a set of PWM matches, methods to test for enrichment are also implemented in `motifTestR`.
A common approach when testing for motif enrichment is to obtain a set of random or background sequences which represent a suitable control set to define the null hypothesis (H~0~).
In `motifTestR`, two alternatives are offered utilising this approach, which both return similar results but involve different levels of computational effort.

The first approach is to sample multiple sets of background sequences and by 'iterating' through to obtain a null distribution for PWM matches and comparing our observed counts against this distribution.
It has been noticed that this approach commonly produces a set of counts for H~0~ which closely resemble a Poisson distribution, and a second approach offered in `motifTestR` is to sample a suitable large set of background sequences and estimate the parameters for the Poisson distribution for each PWM, and testing against these.

## Defining a Set of Control Sequences

Choosing a suitable set of control sequences can be undertaken by any number of methods.
`motifTestR` enables a strategy of matching sequences by any number of given features.
The data object `zr75_enh` contains the candidate enhancers for ZR-75-1 cells defined by v2.0 of the Enhancer Atlas [@gao2019], for chromosome 1 only.
A high proportion of our peaks are associated with these regions and choosing control sequences drawn from the same proportion of these regions may be a viable strategy.

```{r zr75-enh}
data("zr75_enh")
mean(overlapsAny(ar_er_peaks, zr75_enh))
```

First we can annotate each peak by whether there is any overlap with an enhancer, or whether the peak belongs to any other region.
Next we can define two sets of GenomicRanges, one representing the enhancers and the other being the remainder of the genome, here restricted to chromosome 1 for consistency.
Control regions can be drawn from each with proportions that match the test set of sequences.

```{r define-bg-ranges}
ar_er_peaks$feature <- ifelse(
  overlapsAny(ar_er_peaks, zr75_enh), "enhancer", "other"
)
chr1 <- GRanges(sq)[1]
bg_ranges <- GRangesList(
  enhancer = zr75_enh,
  other = GenomicRanges::setdiff(chr1, zr75_enh)
)
```

The provided object `hg19_mask` contains regions of the genome which are rich in Ns, such as centromeres and telomeres.
Sequences containing Ns produce warning messages when matching PWMs and avoiding these regions may be wise, without introducing any sequence bias.
These are then passed to `makeRMRanges()` as ranges to be excluded, whilst sampling multiple  random, size-matched ranges corresponding to our test set of ranges with sequences being analysed, and drawn proportionally from matching genomic regions.
Whilst our example only used candidate enhancers, any type and number of genomic regions can be used, with a limitless number of classification strategies being possible.

```{r rm-ranges}
data("hg19_mask")
set.seed(305)
rm_ranges <- makeRMRanges(
  splitAsList(ar_er_peaks, ar_er_peaks$feature),
  bg_ranges, exclude = hg19_mask,
  n_iter = 100
)
```

This has now returned a set of control ranges which are randomly-selected (R) size-matched (M) to our peaks and are drawn from a similar distribution of genomic features.
By setting `n_iter = 100`, this set will be 100 times larger than our test set and typically this value can be set to 1000 or even 5000 for better estimates of parameters under the null distribution.
However, this will increase the computational burden for analysis.

If not choosing an iterative strategy, a total number of sampled ranges can also be specified.
In this case the column `iteration` will not be added to the returned ranges.

In order to perform the analysis, we can now extract the genomic sequences corresponding to our randomly selected control ranges.
Passing the `mcols` element ensure the iteration numbers are passed to the sequences, as these are required for this approach.

```{r rm-seq}
rm_seq <- getSeq(BSgenome.Hsapiens.UCSC.hg19, rm_ranges)
mcols(rm_seq) <- mcols(rm_ranges)
```

If choosing strategies for enrichment testing outside of `motifTestR`, these sequences can be exported as a fasta file using `writeXStringSet` from the `r Biocpkg("Biostrings")` package.

## Testing For Enrichment

Testing for overall motif enrichment is implemented using multiple strategies, using Poisson, QuasiPoisson or pure Iterative approaches.
Whilst some PWMs may closely follow a Poisson distribution under H~0~, others may be over-dispersed and more suited to a Quasi-Poisson approach.
Each approach has unique advantages and weaknesses as summarised below:

- *Poisson*
    + Can use any number of BG Sequences, unrelated to the test set
    + Modelling is performed based on the expected number of matches *per sequence*
    + The fastest approach
    + Anti-conservative p-values where counts are over-dispersed
    + Appropriate for "quick & dirty" checks for expected discoveries
- *QuasiPoisson*
    + Modelling is performed per *set of sequences* (of identical size to the test set)
    + Requires BG Sequences to be in 'iterative' blocks
    + Fewer 'iterative blocks' can still model over-dispersion reasonably well
- *Iterative*
    + No model assumptions
    + Requires BG Sequences to be in 'iterative' blocks
    + P-Values derived from Z-scores (using the Central Limit Theorem)
    + Sampled p-values from iterations can be used if preferred
    + Requires the largest number of iterative blocks (>1000) 
    + Slowest, but most reliable approach

From a per iteration perspective there is little difference between the Iterative and the modelled QuasiPoisson approaches, however the modelled approaches can still return reliable results from a lower number of iterative blocks, lending a clear speed advantage.
Z-scores returned are only used for statistical testing under the iterative approach and are for indicative purposes only under all other models model.

Whilst no guidelines have been developed for an optimal number of sequences, a control set which is orders of magnitude larger than the test set may be prudent.
A larger set of control sequences clearly leads to longer analytic time-frames and larger computational resources, so this is left to what is considered appropriate by the researcher, nothing that here, we chose a control set which is 100x larger than our test sequences.
If choosing an iterative approach and using the iteration-derived p-values, setting a number of iterations based on the resolution required for these values may be important, noting that the lowest possible p-value is 1/n_iterations.

```{r enrich-res}
enrich_res <- testMotifEnrich(
  ex_pfm, test_seq, rm_seq, min_score = score_thresh, model = "quasi", mc.cores = cores
)
head(enrich_res)
```

Setting the model to "iteration" instead uses a classical iterative approach to define the null distributions of counts and Z-scores are calculated from these values.
The returned p-values from this test are taken from the Z-scores directly, with p-values derived from the sampled iterations also returned if preferred for use in results by the researcher.
Whilst requiring greater computational effort, fewer statistical assumptions are made and results may be more conservative than under modelling approaches.

```{r iter-res}
iter_res <- testMotifEnrich(
  ex_pfm, test_seq, rm_seq, min_score = score_thresh, mc.cores = cores, model = "iteration"
)
head(iter_res)
```

Once we have selected our motifs of interest, sequences with matches can be compared to easily assess co-occurrence, using `plotOverlaps()` from `r Biocpkg("extraChIPs")`.
In our test set, peaks were selected based on co-detection of ESR1 and ANDR, however the rate of co-occurrence is low, revealing key insights into the binding dynamics of these two TFs.

```{r plot-overlaps, fig.cap = "Distribution of select PWM matches within sequences. Each sequence is only considered once and as such, match numbers may be below those returned during testing, which includes multiple matches within a sequence.", fig.alt = "UpSet plot showing the sequences which match one or more of the PWMs in the example dataset."}
ex_pfm |>
  getPwmMatches(test_seq, min_score = score_thresh, mc.cores = cores) |>
  lapply(\(x) x$seq) |>
  plotOverlaps(type = "upset")
```

# Working with Clustered Motifs

TFBMs often contain a high level of similarity to other TFBMs, especially those within large but closely related families, such as the GATA or STAT families.
It can be difficult to ascertain which member of a family is truly bound to a site from inspecting the sequence data alone.

As a relevant example, when looking at the above UpSet plot, about half of the sequences where a match to the ANDR motif was found also contain a FOXA1 motif.
Does this mean that both are bound to the sequence, or have the same sequences simply been matched to both PWMs?
As we can see there is quite some similarity in the core regions of the two binding motifs

```{r compare-andr-foxa1, fig.cap = "Binding Motifs for ANDR and FOXA1 overlaid for the best alignment, showing the similarity in the core regions.", fig.alt="Binding Motifs for ANDR and FOXA1 overlaid for the best alignment, showing the similarity in the core regions."}
c("ANDR", "FOXA1") |>
    lapply(
        \(x) create_motif(ex_pfm[[x]], name = x, type = "PPM")
    ) |>
    view_motifs()
```


`motifTestR` offers a simple but helpful strategy for reducing the level of redundancy within a set of results, by grouping highly similar motifs into a *cluster* and testing for enrichment, or positional bias for any TFBM within that cluster.
The clustering methodology enabled within `motifTestR` allows for the use of any of the comparison methods provided in `compare_motifs()` [@Tremblay2024].

Whilst the example set of TFBMs is slightly artificial motifs can still be grouped into clusters.
If using a larger database of TFBMs a carefully selected threshold may be more appropriate, however for this dataset, four clusters are able to be formed using the default settings. 
There is *no ideal method* to group PWMs into clusters and manual inspection of any clusters produced is the best strategy for this process.
Names are not strictly required for downstream analysis, but can help with interpretability

```{r ex-cl, fig.cap = "Plot produced when requested by clusterMotifs showing the relationship between motifs. The red horizontal line indicates the threshold below which motifs are grouped together into a cluster. Clustering is performed using hclust on the distance matrices produced by `universalmotif::compare_motifs()`", fig.alt = "Dendrogram showing clusters formed by `clusterMotifs()` with default settings."}
cl <- clusterMotifs(ex_pfm, plot = TRUE, labels = NULL)
ex_cl <- split(ex_pfm, cl)
names(ex_cl) <- vapply(split(names(cl), cl), paste, character(1), collapse = "/")
```

These clusters can now be tested for positional bias using `testClusterPos()`.
Matches are the unique sites with a match to *any* PWM within the cluster, and any overlapping sites from matches to multiple PWMs are counted as a single match.
When overlapping matches are found, the one with the highest relative score (score / maxScore) is chosen given that raw scores for each PWM will be on different scales.
A list of *best matches* can be produced in an analogous way to for individual PWMs, using `getClusterMatches()` instead of `getPwmMatches()`, and the motif assigned to each match is also provided.

```{r cl-matches}
cl_matches <- getClusterMatches(
    ex_cl, test_seq, min_score = score_thresh, best_only = TRUE
)
cl_matches
```

These matches can then be passed to `testClusterPos()`, which works in a near-identical manner to `testMotifPos()`

```{r test-cluster-pos}
testClusterPos(cl_matches, test_seq, abs = TRUE)
```

All methods implemented for testing enrichment against a set of background sequences can also be used for clusters of motifs.

```{r test-cluster-enrich}
testClusterEnrich(
  ex_cl, test_seq, rm_seq, min_score = score_thresh, model = "quasi", mc.cores = cores
)
```


# Working With Larger Datasets

Vignettes are commonly prepared for compiling with limited resources and as such example datasets and analyses may reveal less information than realistically sized data.
Motif analysis is particularly well-known for taking many minutes when working with large datasets.
For more comprehensive analysis and realistically sized data, the following code snippets will allow analysis of the above dataset, but without being restricted to chromosome 1.

To obtain the full set of peaks, simply run the following and use these peaks repeating the steps above.

```r
## Not run
base_url <- "https://ftp.ncbi.nlm.nih.gov/geo/samples/GSM3511nnn"
bed_url <- list(
  AR = file.path(
    base_url, "GSM3511083/suppl/GSM3511083%5FAR%5Fpeaks%5FED.bed.gz"
  ),
  ER = file.path(
    base_url, "GSM3511085/suppl/GSM3511085%5FER%5Fpeaks%5FED.bed.gz"
  ),
  H3K27ac = file.path(
    base_url, "GSM3511087/suppl/GSM3511087%5FH3K27ac%5Fpeaks%5FED.bed.gz"
  )
)
all_peaks <- GRangesList(lapply(bed_url, import.bed))
seqlevels(all_peaks) <- seqnames(sq)
seqinfo(all_peaks) <- sq
## Return the ranges with coverage from 2 or more targets
ar_er_peaks <- makeConsensus(
  all_peaks, p = 2/3, method = "coverage", min_width = 200
) |>
  ## Now subset to the ranges which overlap a peak from every target
  subset(n == 3) |> 
  resize(width = 400, fix = 'center')
```

The full set of PWMs for HOCOMOCOv11 (core-A) provided in `MotifDb` can be obtained using the following.
Alternatively, query fields can be customised as preferred.

```r
## Not run
library(MotifDb)
ex_pfm <- MotifDb |>
  subset(organism == "Hsapiens") |>
  query("HOCOMOCOv11-core-A") |>
  as.list() 
names(ex_pfm) <- gsub(".+HOCOMOCOv11-core-A-(.+)_.+", "\\1", names(ex_pfm))
```

Similarly, a set of candidate enhancers found on all chromosomes can be obtained here.
If choosing this dataset, note that `bg_ranges` will need to be drawn from the entire genome, not just chromosome 1.

``` r
## Not run
zr75_url <- "http://www.enhanceratlas.org/data/download/enhancer/hs/ZR75-1.bed"
zr75_enh <- import.bed(zr75_url)
zr75_enh <- granges(zr75_enh)
seqlevels(zr75_enh) <- seqnames(sq)
seqinfo(zr75_enh) <- sq
mean(overlapsAny(ar_er_peaks, zr75_enh))
```

# References {.unnumbered}


# Session info {.unnumbered}

```{r sessionInfo, echo=FALSE}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotMatchPos.R
\name{plotMatchPos}
\alias{plotMatchPos}
\title{Plot Motif Match Positions}
\usage{
plotMatchPos(
  matches,
  binwidth = 10,
  abs = FALSE,
  use_totals = FALSE,
  type = c("density", "cdf", "heatmap"),
  geom = c("smooth", "line", "point", "col"),
  cluster = FALSE,
  w = 0.1,
  heat_fill = NULL,
  ...
)
}
\arguments{
\item{matches}{Output from \link{getPwmMatches}}

\item{binwidth}{Width of bins to use when plotting}

\item{abs}{logical(1) Plot absolute distances from centre}

\item{use_totals}{logical(1). If TRUE, plots will use total counts. The
default (FALSE) plots probabilities.}

\item{type}{Plot match density, the CDF or a binned heatmap}

\item{geom}{Type of geom to be used for line plots. Ignored for heatmaps}

\item{cluster}{logical(1) Cluster motifs when drawing a heatmap. If TRUE a
dendrogram will be added to the LHS of the plot}

\item{w}{Relative width of the dendrogram on (0, 1)}

\item{heat_fill}{scale_fill_continuous object for heatmaps. If not provided,
scale_fill_viridis_c() will be added to the heatmap.}

\item{...}{Passed to individual geom* functions}
}
\value{
A ggplot2 object
}
\description{
Plot the distribution of motif matches across sequences
}
\details{
Multiple options are provided for showing the distribution of PWM matches
within a set of sequences, using either the smoothed probability density,
the probability CDF or as a heatmap.
Distances can be shown as symmetrical around the centre or using absolute
distances from the central position within the sequences.

Heatmaps are only enabled for comparisons across multiple PWMs, with
optional clustering enabled. If adding a dendrogram for clustering,
the returned plot object will be a patchwork object.
}
\examples{
## Load the example PWM
data("ex_pfm")
esr1 <- ex_pfm$ESR1

## Load the example sequences from the peaks
data("ar_er_seq")

## Just the best match
bm <- getPwmMatches(esr1, ar_er_seq, best_only = TRUE)
plotMatchPos(bm, se = FALSE)

## Matches can also be shown by distance from centre
plotMatchPos(bm, abs = TRUE)

## Cumulative Probability plots are also implemented
plotMatchPos(bm, type = "cdf", geom = "line", colour = "red") +
  geom_abline(intercept = 0.5, slope = 1/ 400)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/countPwmMatches.R
\name{countPwmMatches}
\alias{countPwmMatches}
\title{Count the matches to a PWM within an XStringSet}
\usage{
countPwmMatches(
  pwm,
  stringset,
  rc = TRUE,
  min_score = "80\%",
  mc.cores = 1,
  ...
)
}
\arguments{
\item{pwm}{A Position Weight Matrix}

\item{stringset}{An XStringSet}

\item{rc}{logical(1) Also find matches using the reverse complement of pwm}

\item{min_score}{The minimum score to return a match}

\item{mc.cores}{Passed to \link[parallel]{mclapply} when analysing a list of
PWMs}

\item{...}{Passed to \link[Biostrings]{countPWM}}
}
\value{
An integer vector
}
\description{
Count the matches to a PWM within an XStringSet
}
\details{
Will simply count the matches within an XStringSet and return an integer.
All matches are included.
}
\examples{
## Load the example PWM
data("ex_pfm")
esr1 <- ex_pfm$ESR1

## Load the example Peaks
data("ar_er_seq")
countPwmMatches(esr1, ar_er_seq)

## Count all PWMs
countPwmMatches(ex_pfm, ar_er_seq)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testMotifPos.R
\name{testMotifPos}
\alias{testMotifPos}
\title{Test for a Uniform Distribution across a set of best matches}
\usage{
testMotifPos(
  x,
  stringset,
  binwidth = 10,
  abs = FALSE,
  rc = TRUE,
  min_score = "80\%",
  break_ties = "all",
  alt = c("greater", "less", "two.sided"),
  sort_by = c("p", "none"),
  mc.cores = 1,
  ...
)
}
\arguments{
\item{x}{A Position Weight Matrix, universalmotif object or list thereof.
Alternatively can be a single DataFrame or list of DataFrames as returned
by \link{getPwmMatches} with \code{best_only = TRUE}}

\item{stringset}{An XStringSet. Not required if matches are supplied as x}

\item{binwidth}{Width of bins across the range to group data into}

\item{abs}{Use absolute positions around zero to find symmetrical enrichment}

\item{rc}{logical(1) Also find matches using the reverse complement of each
PWM}

\item{min_score}{The minimum score to return a match}

\item{break_ties}{Choose how to resolve matches with tied scores}

\item{alt}{Alternative hypothesis for the binomial test}

\item{sort_by}{Column to sort results by}

\item{mc.cores}{Passed to \link[parallel]{mclapply}}

\item{...}{Passed to \link[Biostrings]{matchPWM}}
}
\value{
A data.frame with columns \code{start}, \code{end}, \code{centre}, \code{width}, \code{total_matches},
\code{matches_in_region}, \code{expected}, \code{enrichment}, \code{prop_total}, \code{p}
and \code{consensus_motif}
The total matches represent the total number of matches within the set of
sequences, whilst the number observed in the final region are also given,
along with the proportion of the total this represents.
Enrichment is simply the ratio of observed to expected based on the
expectation of the null hypothesis

The consensus motif across all matches is returned as a Position Frequency
Matrix (PFM) using \link[Biostrings]{consensusMatrix}.
}
\description{
Test for a Uniform Distribution across a set of best matches
}
\details{
This function tests for an even positional spread of motif matches across a
set of sequences, using the assumption (i.e. H~0~) that if there is no
positional bias, matches will be evenly distributed across all positions
within a set of sequences.
Conversely, if there is positional bias, typically but not necessarily near
the centre of a range, this function intends to detect this signal, as a
rejection of the null hypothesis.

Input can be provided as the output from \link{getPwmMatches} setting
\code{best_only = TRUE} if these matches have already been identified.
If choosing to provide this object to the argument \code{matches}, nothing is required
for the arguments \code{pwm}, \code{stringset}, \code{rc}, \code{min_score} or \code{break_ties}
Otherwise, a Position Weight Matrix (PWM) and an \code{XStringSet} are required,
along with the relevant arguments, with best matches identified within the
function.

The set of best matches are then grouped into bins along the range, with the
central bin containing zero, and tallied.
Setting \code{abs} to \code{TRUE} will set all positions from the centre as
\emph{absolute values}, returning counts purely as bins with distances from zero,
marking this as an inclusive lower bound.
Motif alignments are assigned into bins based on the central position of the
match, as provided in the column \code{from_centre} when calling
\link{getPwmMatches}.

The \link[stats]{binom.test} is performed on each bin using the alternative
hypothesis, with the returned p-values across all bins combined using the
Harmonic Mean p-value (HMP) (See \link[harmonicmeanp]{p.hmp}).
All bins with raw p-values below the HMP are identified and the returned
values for start, end, centre, width, matches in region, expected and
enrichment are across this set of bins.
The expectation is that where a positional bias is evident, this will be a
narrow range containing a non-trivial proportion of the total matches.

It should also be noted that \code{binom.test()} can return p-values of zero, as
beyond machine precision. In these instances, zero p-values are excluded from
calculation of the HMP. This will give a very slight conservative bias, and
assumes that for these extreme cases, neighbouring bins are highly likely to
also return extremely low p-values and no significance will be lost.
}
\examples{
## Load the example PWM
data("ex_pfm")
esr1 <- ex_pfm$ESR1

## Load the example sequences
data("ar_er_seq")

## Get the best match and use this data
matches <- getPwmMatches(esr1, ar_er_seq, best_only = TRUE)
## Test for enrichment in any position
testMotifPos(matches)

## Provide a list of PWMs, testing for distance from zero
testMotifPos(ex_pfm, ar_er_seq, abs = TRUE, binwidth = 10)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clusterMotifs.R
\name{clusterMotifs}
\alias{clusterMotifs}
\title{Assign each motif to a cluster}
\usage{
clusterMotifs(
  motifs,
  type = c("PPM", "ICM"),
  method = c("PCC", "EUCL", "SW", "KL", "ALLR", "BHAT", "HELL", "SEUCL", "MAN",
    "ALLR_LL", "WEUCL", "WPCC"),
  power = 1,
  agglom = "complete",
  thresh = 0.2,
  return_d = FALSE,
  plot = FALSE,
  labels = FALSE,
  cex = 1,
  linecol = "red",
  ...
)
}
\arguments{
\item{motifs}{A list of universalmotifs or a list of PWMs}

\item{type}{Can be ICM or PPM}

\item{method}{The method to be used for determining similarity/distances}

\item{power}{Raise correlation matrices to this power before converting to a
distance matrix. Only applied if method is either "PCC" or "WPCC"}

\item{agglom}{Method to be used for agglomeration by \link[stats]{hclust}}

\item{thresh}{Tree heights below which motifs are formed into a cluster}

\item{return_d}{logical(1) Return the distance matrices for each cluster}

\item{plot}{Show tree produced by \link[stats]{hclust}. If requested the
value set by thresh will be shown as a horizontal line}

\item{labels, cex}{Passed to \link[stats]{plot.hclust}}

\item{linecol}{Passed to \link[graphics]{abline} as the argument \code{col}}

\item{...}{passed to \link[universalmotif]{compare_motifs}}
}
\value{
Named vector with numeric values representing which cluster each
motif has been assigned to.

If setting return_d = TRUE, a named list will be returned with the clusters
as the element \code{cl} and a list with distance matrices for each cluster as the
element \code{d}
}
\description{
Cluster related motifs for testing as a group
}
\details{
This builds on \link[universalmotif]{compare_motifs}, enabling the
assignment of each PWM to a cluster, and subsequent testing of motifs as
a cluster, rather than returning individual results.

Internally all matrices are converted to distance matrices and
\link[stats]{hclust} is used to form clusters. By default, options such as
"EUCL", "MAN" produce distances, whilst similarity matrices are produced
when choosing "PCC" and other correlation based metrics. In these cases, the
distance matrix is obtained by taking 1 - similarity.

By default PWM labels are hidden (labels = FALSE), however these can be shown
using labels = NULL as explained in \link[stats]{plot.hclust}.

Raising the threshold will lead to fewer, larger clusters whilst leaving
this value low will return a more conservative approach, with more smaller
clusters.
The final decision as the best clustering strategy is highly subjective and
left to the user.
Manual inspection of motifs within a cluster can be performed using
\link[universalmotif]{view_motifs}, as shown in the vignette.
}
\examples{
# Load the example motifs
data("ex_pfm")

# Return a vector with each motif assigned a cluster
# The default uses Pearson's Correlation Coefficient
clusterMotifs(ex_pfm)

# Preview the settings noting that showing labels can clutter the plot
# with large numbers of motifs. The defaults for Euclidean distance
# show the threshold may need raising
clusterMotifs(ex_pfm, plot = TRUE, labels = NULL, method = "EUCL")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{ar_er_peaks}
\alias{ar_er_peaks}
\title{A set of peaks with AR and ER detected}
\format{
An object of class \code{GRanges} of length 849.
}
\source{
\url{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE123767}
}
\usage{
data("ar_er_peaks")
}
\description{
A set of ChIP-Seq peaks where AR and ER were both detected
}
\details{
The subset of peaks found on chr1 which contained signal from at least
two of AR, ER and H3K27ac, taken from GSE123767.
Peaks were resized to a uniform width of 400bp after downloading

Generation of these ranges is documented in
\code{system.file("scripts/ar_er_peaks.R", package = "motifTestR")}
}
\examples{
data("ar_er_peaks")
ar_er_peaks
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testMotifEnrich.R
\name{testMotifEnrich}
\alias{testMotifEnrich}
\title{Test motif enrichment using a background set of sequences}
\usage{
testMotifEnrich(
  pwm,
  stringset,
  bg,
  var = "iteration",
  model = c("poisson", "hypergeometric", "quasipoisson", "glm_poisson", "iteration"),
  sort_by = c("p", "none"),
  mc.cores = 1,
  prior.count = 1,
  seed = 100,
  ...
)
}
\arguments{
\item{pwm}{A Position Weight Matrix or list of PWMs}

\item{stringset}{An XStringSet with equal sequence widths}

\item{bg}{An XStringSet with the same sequence widths as the test XStringset}

\item{var}{A column in the mcols element of bg, usually denoting an iteration
number}

\item{model}{The model used for analysis}

\item{sort_by}{Column to sort results by}

\item{mc.cores}{Passed to \link[parallel]{mclapply}}

\item{prior.count}{Added to all counts to better manage zero counts in
background sequences. For analysis under QuasiPoisson models prior counts
are added as Poisson noise using this value as expected counts}

\item{seed}{Used for reproducibility when adding Poisson noise}

\item{...}{Passed to \link{getPwmMatches} or \link{countPwmMatches}}
}
\value{
A data.frame with columns: \code{sequences}, \code{matches}, \code{expected}, \code{enrichment},
and \code{p}, with additional columns \code{Z}, \code{est_bg_rate} (Poisson), \code{odds_ratio}
(Hypergeometric) or \code{Z}, \code{sd_bg}, \code{n_iter} and \code{iter_p} (Iterations).
The numbers of sequences and matches refer to the test set of sequences,
whilst expected is the expected number of matches under the Poisson or
iterative null distribution. The ratio of matches to expected is given as
\code{enrichment}, along with the Z score and p-value. Whilst the Z score is only
representative under the Poisson model, it is used to directly estimate
p-values under the iterative approach.
Under this latter approach, the sd of the matches found in the background
sequences is also given, along with the number of iterations and the p-values
from permutations testing the one-sided hypothesis hypothesis for enrichment.

It may also be worth noting that if producing background sequences using
\link{makeRMRanges} with \code{replace = TRUE} and \code{force_ol = TRUE}, the
iterative model corresponds to a bootstrap, given that the test sequences
will overlap the background sequences and background ranges are able to be
sampled with replacement.
}
\description{
Test for motif enrichment within a set of sequences using a background set to
derive a NULL hypothesis
}
\details{
This function offers four alternative models for assessing the enrichment of
a motif within a set of sequences, in comparison to a background set of
sequences.
Selection of the BG sequences plays an important role and, in conjunction
with the question being addressed, determines the most appropriate model to
use for testing, as described below.

It should also be noted that the larger the BG set of sequences, the larger
the computational burden, and results can take far longer to return. For
many millions of background sequences, this may run beyond an hour
\subsection{Descriptions of Models and Use Cases}{
\subsection{Hypergeometric Tests}{

Hypergeometric tests are best suited to the use case where the test set of
sequences represents a subset of a larger set, with a specific feature or
behaviour, whilst the BG set may be the remainder of the set without that
feature. For example, the test set may represent ChIP-Seq binding sites
where signal changes in response to treatment, whilst the BG set represents
the sites where no changed signal was observed. Testing is one-sided, for
enrichment of motifs within the test set.

Due to these relatively smaller sized datasets, setting
model = "hypergeometric", will generally return results quickly
}

\subsection{Poisson Tests}{

This approach requires a set of background sequences which should be much
larger than the test set of sequences.
The parameters for a Poisson model are estimated in a per-sequence manner on
the set of BG sequences, and the observed rate of motif-matches within the
test set is then tested using \link[stats]{poisson.test}.
Testing is two-sided.

This approach assumes that all matches follow a Poisson distribution, which
is often true, but data can also be over-dispersed. Given that this model can
also return results relatively quickly, is it primarily suitable for data
exploration, such as quickly checking for expected behaviours, but not for
final results.
}

\subsection{Quasi-Poisson Test}{

The quasipoisson model allows for over-dispersion and will return more
conservative results than using the standard Poisson model.
Under the method currently implemented here, BG sequences should be divided
into blocks (i.e. iterations), identical in size to the test set of sequences.
Model parameters are estimated per iteration across the BG set of sequences,
with the rate of matches in the test set being compared against these blocks.
This ensures more conservative results that if analysing test and bg
sequences as collections of individual sequences.

It is expected that the BG set will matched for the features of interest and
chosen using \link{makeRMRanges} with a large number of iterations, e.g.
n_iter = 1000.
Due to this parameterisation, quasipoisson approaches can be computationally
time-consuming, as this is effectively an iterative approach.
Testing is two-sided.
}

\subsection{GLM-Poisson Test}{

This follows the same approach as the Quasi-Poisson model, relying on fitting
iterations using \code{\link[=glm]{glm()}}.
For this model however, no over-dispersions are estimated and the underlying
family is simply the \code{poisson()} family
}

\subsection{Iteration}{

Setting the model as "iteration" performs a non-parametric analysis, with
the exception of returning Z-scores under the Central Limit Theorem.
Mean and SD of matches is found for each iteration, and used to return Z
scores, with p-values returned from both a Z-test and from comparing
observed values directly to sampled values obtained from the BG sequences.
Sampled values are calculated directly and as such, are limited in precision.

As for the QuasiPoisson model, a very large number of iterations is expected
to be used, to ensure the CLT holds, again making this a computationally
demanding test.
Each iteration/block is expected to be identically-sized to the test set,
and matched for any features as appropriate using \code{\link[=makeRMRanges]{makeRMRanges()}}.
}

}
}
\examples{
## Load the example peaks & the sequences
data("ar_er_peaks")
data("ar_er_seq")
sq <- seqinfo(ar_er_peaks)
## Now sample size-matched ranges 10 times larger. In real-world analyses,
## this set should be sampled as at least 1000x larger, ensuring features
## are matched to your requirements. This example masks regions with known N
## content, including centromeres & telomeres
data("hg19_mask")
set.seed(305)
bg_ranges <- makeRMRanges(
  ar_er_peaks, GRanges(sq)[1], exclude = hg19_mask, n_iter = 10
)

## Convert ranges to DNAStringSets
library(BSgenome.Hsapiens.UCSC.hg19)
genome <- BSgenome.Hsapiens.UCSC.hg19
bg_seq <- getSeq(genome, bg_ranges)
mcols(bg_seq)$iteraton <-bg_ranges$iteration

## Test for enrichment of the ESR1 motif
data("ex_pfm")
esr1 <- ex_pfm$ESR1
testMotifEnrich(esr1, ar_er_seq, bg_seq, model = "poisson")

## Test all motifs
testMotifEnrich(ex_pfm, ar_er_seq, bg_seq, model = "poisson")


}
\seealso{
\code{\link[=makeRMRanges]{makeRMRanges()}}, \code{\link[=getPwmMatches]{getPwmMatches()}}, \code{\link[=countPwmMatches]{countPwmMatches()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{zr75_enh}
\alias{zr75_enh}
\title{Candidate Enhancer Regions from ZR-75-1 Cells}
\format{
An object of class \code{GRanges} of length 5237.
}
\source{
\url{http://www.enhanceratlas.org/index.php}
}
\usage{
data("zr75_enh")
}
\description{
The chr1 subset of candidate enhancers for ZR-75-1 cells
}
\details{
These enhancers are the chr1 subset of enhancer regions for ZR-75-1 cells as
identified by EnhancerAtlas 2.0

#' Generation of these ranges is documented in
\code{system.file("scripts/zr75_enh.R", package = "motifTestR")}
}
\examples{
data("zr75_enh")
zr75_enh
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testClusterEnrich.R
\name{testClusterEnrich}
\alias{testClusterEnrich}
\title{Test enrichment across a cluster of motifs using a background set of sequences}
\usage{
testClusterEnrich(
  cl,
  stringset,
  bg,
  var = "iteration",
  model = c("poisson", "hypergeometric", "quasipoisson", "glm_poisson", "iteration"),
  sort_by = c("p", "none"),
  mc.cores = 1,
  prior.count = 1,
  seed = 100,
  ...
)
}
\arguments{
\item{cl}{A list of Position Weight Matrices, universalmotifs, with each
element representing clusters of related matrices}

\item{stringset}{An XStringSet with equal sequence widths}

\item{bg}{An XStringSet with the same sequence widths as the test XStringset}

\item{var}{A column in the mcols element of bg, usually denoting an iteration
number}

\item{model}{The model used for analysis}

\item{sort_by}{Column to sort results by}

\item{mc.cores}{Passed to \link[parallel]{mclapply}}

\item{prior.count}{Added to all counts to better manage zero counts in
background sequences. For analysis under QuasiPoisson models prior counts
are added as Poisson noise using this value as expected counts}

\item{seed}{Used for reproducibility when adding Poisson noise}

\item{...}{Passed to \link{getPwmMatches} or \link{countPwmMatches}}
}
\value{
See \link{testMotifEnrich}
}
\description{
Test for enrichment of any motif within a cluster across a set of sequences
using a background set to
derive a NULL hypothesis
}
\details{
This extends the analytic methods offered by \link{testMotifEnrich} using
PWMs grouped into a set of clusters.
As with all cluster-level approaches, hits from multiple PWMs which overlap
are counted as a single hit ensuring that duplicated matches are not
double-counted, and that only individual positions within the sequences are.
}
\examples{
## Load the example peaks & the sequences
data("ar_er_peaks")
data("ar_er_seq")
sq <- seqinfo(ar_er_peaks)
## Now sample size-matched ranges 10 times larger. In real-world analyses,
## this set should be sampled as at least 1000x larger, ensuring features
## are matched to your requirements. This example masks regions with known N
## content, including centromeres & telomeres
data("hg19_mask")
set.seed(305)
bg_ranges <- makeRMRanges(
  ar_er_peaks, GRanges(sq)[1], exclude = hg19_mask, n_iter = 10
)

## Convert ranges to DNAStringSets
library(BSgenome.Hsapiens.UCSC.hg19)
genome <- BSgenome.Hsapiens.UCSC.hg19
bg_seq <- getSeq(genome, bg_ranges)

## Test for enrichment of clustered motifs
data("ex_pfm")
cl <- list(A = ex_pfm[1], B = ex_pfm[2:3])
testClusterEnrich(cl, ar_er_seq, bg_seq, model = "poisson")


}
\seealso{
\code{\link[=makeRMRanges]{makeRMRanges()}}, \code{\link[=getClusterMatches]{getClusterMatches()}}, \code{\link[=countClusterMatches]{countClusterMatches()}}, \code{\link[=testMotifEnrich]{testMotifEnrich()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simMultiMotifs.R
\name{simMultiMotifs}
\alias{simMultiMotifs}
\title{Simulate sequences with multiple motifs}
\usage{
simMultiMotifs(
  n,
  width,
  pfm = NULL,
  bg = NULL,
  nt = c("A", "C", "G", "T"),
  prob = rep(0.25, 4),
  shape1 = 1,
  shape2 = shape1,
  rate = NA,
  theta = NA,
  as = "DNAStringSet",
  ol = c("random", "first", "last"),
  ...
)
}
\arguments{
\item{n}{The number of sequences to simulate}

\item{width}{Width of sequences to simulate}

\item{pfm}{List of Probability Weight/Frequency Matrices}

\item{bg}{Optional, pre-defined set of background sequences. Can be passed as
an XStringSet or character vector. All sequences must be the same width}

\item{nt}{Nucleotides to include}

\item{prob}{Sampling probabilities for each nucleotide}

\item{shape1, shape2}{Passed to \link[VGAM]{rbetabinom.ab}}

\item{rate}{The expected rate of motifs per sequence. Is equivalent to
\eqn{ \lambda } in \link[stats]{rpois}. If set to NULL or NA, all sequences will
be simulated with a single motif, otherwise a Poisson distribution will be used}

\item{theta}{Overdispersion parameter passed to \link[MASS]{rnegbin}.
If set to NULL or NA the rate parameter will be passed to \link[stats]{rpois}.
However if this value is set, the rate and theta parameters are passed to
\link[MASS]{rnegbin} to simulate overdispersed counts}

\item{as}{ObjectClass to return objects as. Defaults to DNAStringSet, but
other viable options may include 'character', 'CharacterList' or any
other class from which a character vector may be coerced.}

\item{ol}{When randomly simulated positions overlap, choose one either at
random, by the first occurring PFM in the list of PFMs, or by the last.}

\item{...}{Not used}
}
\value{
A DNAStringSet with mcols denoting the positions of all inserted motifs
}
\description{
Simulate a set of sequences incorporating multiple motifs
}
\details{
Simulate a set of sequences with multiple motifs inserted using different
rates and distributions, as specified.
All shape, rate and theta parameters are recycled to match the length of the
supplied motif list, and can be supplied as vectors to tailor these
parameters to each provided element of the list of matrices
}
\examples{
data("ex_pfm")
## Simulate sequences including both ESR1 and ANDR, but with
## ESR1 being included at a higher rate
seq <- simMultiMotifs(10, 100, ex_pfm[1:2], rate = c(2, 1))
seq
## The positions of the motifs are included in the mcols
mcols(seq)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getPwmMatches.R
\name{getPwmMatches}
\alias{getPwmMatches}
\title{Find all PWM matches within an XStringSet}
\usage{
getPwmMatches(
  pwm,
  stringset,
  rc = TRUE,
  min_score = "80\%",
  best_only = FALSE,
  break_ties = c("all", "random", "first", "last", "central"),
  mc.cores = 1,
  ...
)
}
\arguments{
\item{pwm}{A Position Weight Matrix, list of PWMs or universalmotif list}

\item{stringset}{An XStringSet}

\item{rc}{logical(1) Also find matches using the reverse complement of pwm}

\item{min_score}{The minimum score to return a match}

\item{best_only}{logical(1) Only return the best match}

\item{break_ties}{Method for breaking ties when only returning the best match
Ignored when all matches are returned (the default)}

\item{mc.cores}{Passed to \link[parallel]{mclapply} if passing multiple PWMs}

\item{...}{Passed to \link[Biostrings]{matchPWM}}
}
\value{
A DataFrame with columns: \code{seq}, \code{score}, \code{direction}, \code{start},
\code{end}, \code{from_centre}, \code{seq_width}, and \code{match}

The first three columns describe the sequence with matches, the score of
the match and whether the match was found using the forward or reverse PWM.
The columns \code{start}, \code{end} and \code{width} describe the where the match was found
in the sequence, whilst \code{from_centre} defines the distance between the centre
of the match and the centre of the sequence being queried.
The final column contains the matching fragment of the sequence as an
\code{XStringSet}.

When passing a list of PWMs, a list of the above DataFrames will be returned.
}
\description{
Find all PWM matches within a set of sequences
}
\details{
Taking a set of sequences as an XStringSet, find all matches above the
supplied score (i.e. threshold) for a single Position Weight Matrix (PWM),
generally representing a transcription factor binding motif.
By default, matches are performed using the PWM as provided and the reverse
complement, however this can easily be disabled by setting \code{rc = FALSE}.

The function relies heavily on \link[Biostrings]{matchPWM} and
\link[IRanges]{Views} for speed.

When choosing to return the best match (\code{best_only = TRUE}), only the match
with the highest score is returned for each sequence.
Should there be tied scores, the best match can be chosen as either the first,
last, most central, all tied matches, or choosing one at random (the default).
}
\examples{
## Load the example PWM
data("ex_pfm")
esr1 <- ex_pfm$ESR1

## Load the example Peaks
data("ar_er_seq")

## Return all matches
getPwmMatches(esr1, ar_er_seq)

## Just the best match
getPwmMatches(esr1, ar_er_seq, best_only = TRUE)

## Apply multiple PWMs as a list
getPwmMatches(ex_pfm, ar_er_seq, best_only = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{ar_er_seq}
\alias{ar_er_seq}
\title{Sequences from peaks with AR and ER detected}
\format{
An object of class \code{DNAStringSet} of length 849.
}
\usage{
data("ar_er_seq")
}
\description{
The genomic sequences obtained from the ar_er_peaks
}
\details{
These sequences represent the sequences obtained from
BSgenome.Hsapiens.UCSC.hg19 for thw peaks supplied as \code{ar_er_peaks}

Generation of these sequences is documented in
\code{system.file("scripts/ar_er_peaks.R", package = "motifTestR")}
}
\examples{
data("ar_er_seq")
ar_er_seq
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/makeRMRanges.R
\name{makeRMRanges}
\alias{makeRMRanges}
\alias{makeRMRanges,GRanges,GRanges-method}
\alias{makeRMRanges,GRangesList,GRangesList-method}
\title{Form a set of random, matching ranges for bootstrapping or permuting}
\usage{
makeRMRanges(x, y, ...)

\S4method{makeRMRanges}{GRanges,GRanges}(
  x,
  y,
  exclude = GRanges(),
  n_iter = 1,
  n_total = NULL,
  replace = TRUE,
  ...,
  force_ol = TRUE
)

\S4method{makeRMRanges}{GRangesList,GRangesList}(
  x,
  y,
  exclude = GRanges(),
  n_iter = 1,
  n_total = NULL,
  replace = TRUE,
  mc.cores = 1,
  ...,
  force_ol = TRUE,
  unlist = TRUE
)
}
\arguments{
\item{x}{GRanges/GRangesList with ranges to be matched}

\item{y}{GRanges/GRangesList with ranges to select random matching ranges from}

\item{...}{Not used}

\item{exclude}{GRanges of ranges to omit from testing}

\item{n_iter}{The number of times to repeat the random selection process}

\item{n_total}{Setting this value will over-ride anything set using n_iter.
Can be vector of any length, corresponding to the length of x, when x is a
GRangesList}

\item{replace}{logical(1) Sample with our without replacement when creating
the set of random ranges.}

\item{force_ol}{logical(1) Enforce an overlap between every site in x and y}

\item{mc.cores}{Passsed to \link[parallel]{mclapply}}

\item{unlist}{logical(1) Return as a sorted GRanges object, or leave as a
GRangesList}
}
\value{
A GRanges or GRangesList object
}
\description{
Form a set of ranges from y which (near) exactly match those in x for use as
a background set requiring matching
}
\details{
This function uses the width distribution of the 'test' ranges (i.e. \code{x}) to
randomly sample a set of ranges with matching width from the ranges provided
in \code{y}. The width distribution will clearly be exact when a set of
fixed-width ranges is passed to \code{x}, whilst random sampling may yield some
variability when matching ranges of variable width.

When both x and y are GRanges objects, they are implicitly assumed to both
represent similar ranges, such as those overlapping a promoter or enhancer.
When passing two GRangesList objects, both objects are expected to contain
ranges annotated as belonging to key features, such that the list elements in
y must encompass all elements in x.
For example if \code{x} contains two elements named 'promoter' and 'intron', \code{y}
should also contain elements named 'promoter' and 'intron' and these will
be sampled as matching ranges for the same element in \code{x}. If elements of
\code{x} and \code{y} are not named, they are assumed to be in matching order.

The default behaviour is to assume that randomly-generated ranges are for
iteration, and as such, ranges are randomly formed in multiples of the number
of 'test' ranges provided in \code{x}. The column \code{iteration} will be added to the
returned ranges.
Placing any number into the \code{n_total} argument will instead select a total
number of ranges as specified here. In this case, no \code{iteration} column will
be included in the returned ranges.

Sampling is assumed to be with replacement as this is most suitable for
bootstrapping and related procedures, although this can be disabled by
setting \code{replace = FALSE}
}
\examples{
## Load the example peaks
data("ar_er_peaks")
sq <- seqinfo(ar_er_peaks)
## Now sample size-matched ranges for two iterations from chr1
makeRMRanges(ar_er_peaks, GRanges(sq)[1], n_iter = 2)

## Or simply sample 100 ranges if not planning any iterative analyses
makeRMRanges(ar_er_peaks, GRanges(sq)[1], n_total = 100)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testClusterPos.R
\name{testClusterPos}
\alias{testClusterPos}
\title{Test positional bias motifs within a cluster}
\usage{
testClusterPos(
  x,
  stringset,
  binwidth = 10,
  abs = FALSE,
  rc = TRUE,
  min_score = "80\%",
  break_ties = "all",
  alt = c("greater", "less", "two.sided"),
  sort_by = c("p", "none"),
  mc.cores = 1,
  ...
)
}
\arguments{
\item{x}{A Position Weight Matrix, universalmotif object or list thereof.
Alternatively can be a single DataFrame or list of DataFrames as returned
by \link{getClusterMatches} with \code{best_only = TRUE}}

\item{stringset}{An XStringSet. Not required if matches are supplied as x}

\item{binwidth}{Width of bins across the range to group data into}

\item{abs}{Use absolute positions around zero to find symmetrical enrichment}

\item{rc}{logical(1) Also find matches using the reverse complement of each
PWM}

\item{min_score}{The minimum score to return a match}

\item{break_ties}{Choose how to resolve matches with tied scores}

\item{alt}{Alternative hypothesis for the binomial test}

\item{sort_by}{Column to sort results by}

\item{mc.cores}{Passed to \link[parallel]{mclapply}}

\item{...}{Passed to \link[Biostrings]{matchPWM}}
}
\value{
A data.frame with columns \code{start}, \code{end}, \code{centre}, \code{width}, \code{total_matches},
\code{matches_in_region}, \code{expected}, \code{enrichment}, \code{prop_total}, \code{p}
and \code{consensus_motif}
The total matches represent the total number of matches within the set of
sequences, whilst the number observed in the final region are also given,
along with the proportion of the total this represents.
Enrichment is simply the ratio of observed to expected based on the
expectation of the null hypothesis

The consensus motif across all matches is returned as a Position Frequency
Matrix (PFM) using \link[Biostrings]{consensusMatrix}.
}
\description{
Test positional bias for all motifs within a given cluster
}
\details{
This is a reimplementation of \link{testMotifPos} for sets of motifs which
have been clustered for similarity.
The positions test the bias of any motifs within the cluster given that
overlapping matches are only counted once, and with the match retained being
the one with the highest relative score.

It should also be noted that some motif clusters will contain PWMs of
varying length. When finding positional bias, the widest motif is taken as
the width for all, and any matches from narrower motifs outside of the range
allowed by wider motifs are discarded. This reduction in signal will make a
small difference in the outer bins, but is not considered to be problematic
for the larger analysis.
}
\examples{
## Load the example PWM
data("ex_pfm")
## Load the example sequences
data("ar_er_seq")

## Cluster the motifs
cl <- list(A = ex_pfm[1], B = ex_pfm[2:3])

## Get the best match and use this data
matches <- getClusterMatches(cl, ar_er_seq, best_only = TRUE)
## Test for enrichment in any position
testClusterPos(matches)

## Or just pass the clustered matrices
## Here we've set abs = TRUE to test absolute distance from the centre
testClusterPos(cl, ar_er_seq, abs = TRUE, binwidth = 10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/motifTestR.R
\docType{package}
\name{motifTestR-package}
\alias{motifTestR}
\alias{motifTestR-package}
\title{motifTestR: Perform Key Analyses on Transcription Factor Binding Motifs}
\description{
The package \code{motifTestR} has been designed for two primary analyses of TFBMs,
testing for positional bias and overall enrichment.
}
\details{
The package \code{motifTestR} provides two primary functions for testing TFBMs
within a set of sequences
\itemize{
\item \code{\link[=testMotifPos]{testMotifPos()}} for detecting positional bias within a set of test
sequences
\item \code{\link[=testMotifEnrich]{testMotifEnrich()}} for testing overall enrichment of a TFBM within a set
of test sequences
}

Motifs are also able to be clustered for analysis as a cluster, or for
grouping results.
Clusters from external approaches can also be incorporated.
\itemize{
\item \code{\link[=testClusterPos]{testClusterPos()}} for detecting positional bias for matches to any motif
annotated to a cluster, within a set of test sequences
\item \code{\link[=testClusterEnrich]{testClusterEnrich()}} for testing overall enrichment of any TFBM annotated
to a cluster, within a set of test sequences
}

The main functions rely on lower-level functions such as:
\itemize{
\item \code{\link[=countPwmMatches]{countPwmMatches()}} simply counts the number of matches within an
\code{XStringSet}
\item \code{\link[=getPwmMatches]{getPwmMatches()}} returns the position of matches within an \code{XStringSet}
\item \code{\link[=countClusterMatches]{countClusterMatches()}} simply counts the number of matches to motifs
annotated to a cluster within an \code{XStringSet}
\item \code{\link[=getClusterMatches]{getClusterMatches()}} returns the position of matches to motifs annotated
to a cluster within an \code{XStringSet}
\item \code{\link[=makeRMRanges]{makeRMRanges()}} which produces a set of random, matching ranges based on
key characteristics of the set of test sequences/ranges
}

A simple utility function is provided to enable visualisation of results
\itemize{
\item \code{\link[=plotMatchPos]{plotMatchPos()}} enables visualisation of the matches within a set of
sequences using multiple strategies
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/smped/motifTestR}
  \item Report bugs at \url{https://github.com/smped/motifTestR/issues}
}

}
\author{
Stevie Pederson
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simSeq.R
\name{simSeq}
\alias{simSeq}
\title{Simulate sequences using optional TFBMs}
\usage{
simSeq(
  n,
  width,
  pfm = NULL,
  nt = c("A", "C", "G", "T"),
  prob = rep(0.25, 4),
  shape1 = 1,
  shape2 = shape1,
  rate = NULL,
  theta = NULL,
  as = "DNAStringSet",
  ...
)
}
\arguments{
\item{n}{The number of sequences to simulate}

\item{width}{Width of sequences to simulate}

\item{pfm}{Probability Weight/Frequency Matrix}

\item{nt}{Nucleotides to include}

\item{prob}{Sampling probabilities for each nucleotide}

\item{shape1, shape2}{Passed to \link[VGAM]{rbetabinom.ab}}

\item{rate}{The expected rate of motifs per sequence. Is equivalent to
\eqn{ \lambda } in \link[stats]{rpois}. If set to NULL or NA, all sequences
will be simulated with a single motif, otherwise a Poisson distribution will
be used}

\item{theta}{Overdispersion parameter passed to \link[MASS]{rnegbin}.
If set to NULL or NA, the rate parameter will be passed to \link[stats]{rpois}.
However if this value is set, the rate and theta parameters are passed to
\link[MASS]{rnegbin} to simulate overdispersed counts}

\item{as}{ObjectClass to return objects as. Defaults to DNAStringSet, but
other viable options may include 'character', 'CharacterList' or any
other class from which a character vector may be coerced.}

\item{...}{Not used}
}
\value{
By default a DNAStringSet will be returned.
If possible, the position of any randomly sampled motifs will be included
in the mcols element of the returned object.
}
\description{
Simulate a set of fixed-width sequences using optional TFBMs
}
\details{
Using the nucleotide and probabilities provided as set of sequences can be
simulated. By default, this will effectively be a set of 'background'
sequences, with letters effectively chosen at random.

If a PWM/PFM is supplied, the shape parameters are first passed to
\link[VGAM]{rbetabinom.ab} to determine the random positions the motif will
be placed, with the default parameters representing a discrete uniform
distribution.

The sequences to have a motif inserted will be selected, along with the
number of motifs, using the rate and theta parameters.
If both are NULL, every sequence will have a single motif inserted.
If the rate is > 0 and theta is NULL, sequences will be selected to have
motifs inserted using a poisson distribution.
If theta is also provided, sequences will be selected to contain motifs
using a negative binomial distribution, noting that smaller values of theta
lead to higher over-dispersion

Once positions and sequences for the TFBM have been selected, nucleotides
will be randomly sampled using the probabilities provided in the PWM and
these motifs will be placed at the randomly sampled positions
}
\examples{
## Randomly generate 10x50nt sequences without any TFBMs present
simSeq(10, 50)

## Now place a motif at random positions
data('ex_pfm')
sim_seq <- simSeq(10, width = 20, pfm = ex_pfm$ESR1)
sim_seq
## The position of the motif within each sequence is included in the mcols
mcols(sim_seq)

## Use this to extract the random motifs from the random sequences
library(IRanges)
i <- mcols(sim_seq)$pos + cumsum(width(sim_seq)) - width(sim_seq)
Views(unlist(sim_seq), start = i, width = 10)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getClusterMatches.R
\name{getClusterMatches}
\alias{getClusterMatches}
\alias{countClusterMatches}
\title{Find matches from a PWM cluster within an XStringSet}
\usage{
getClusterMatches(
  cl,
  stringset,
  rc = TRUE,
  min_score = "80\%",
  best_only = FALSE,
  break_ties = c("all", "random", "first", "last", "central"),
  mc.cores = 1,
  ...
)

countClusterMatches(
  cl,
  stringset,
  rc = TRUE,
  min_score = "80\%",
  mc.cores = 1,
  ...
)
}
\arguments{
\item{cl}{A list of Position Weight Matrices, universalmotifs, with each
element representing clusters of related matrices}

\item{stringset}{An XStringSet}

\item{rc}{logical(1) Also find matches using the reverse complement of PWMs
in the cluster}

\item{min_score}{The minimum score to return a match}

\item{best_only}{logical(1) Only return the best match}

\item{break_ties}{Method for breaking ties when only returning the best match
Ignored when all matches are returned (the default)}

\item{mc.cores}{Passed to \link[parallel]{mclapply}}

\item{...}{Passed to \link[Biostrings]{matchPWM}}
}
\value{
Output from getClusterMatches will be a list of DataFrames with columns:
\code{seq}, \code{score}, \code{direction}, \code{start}, \code{end}, \code{from_centre}, \code{seq_width},
\code{motif} and \code{match}

The first three columns describe the sequence with matches, the score of
the match and whether the match was found using the forward or reverse PWM.
The columns \code{start}, \code{end} and \code{width} describe the where the match was found
in the sequence, whilst \code{from_centre} defines the distance between the centre
of the match and the centre of the sequence being queried.
The motif column denotes which individual motif was found to match in this
position, again noting that when matches overlap, only the one with the
highest relative score is returned.
The final column contains the matching fragment of the sequence as an
\code{XStringSet}.

Output from countClusterMatches will be a simple integer vector the same
length as the number of clusters
}
\description{
Find matches from a PWM cluster within a set of sequences
}
\details{
This function extends \link{getPwmMatches} by returning a single set of
results for set of clustered motifs.
This can help remove some of the redundancy in results returned for highly
similar PWMs, such as those in the GATA3 family.

Taking a set of sequences as an XStringSet, find all matches above the
supplied score (i.e. threshold) for a list of Position Weight Matrices
(PWMs), which have been clustered together as highly-related motifs.
By default, matches are performed using the PWMs as provided and the reverse
complement, however this can easily be disabled by setting \code{rc = FALSE}.

The function relies heavily on \link[Biostrings]{matchPWM} and
\link[IRanges]{Views} for speed.

Where overlapping matches are found for the PWMs within a cluster, only a
single match is returned.
The motif with the highest relative score (score / maxScore(PWM)) is selected.

When choosing to return the best match (\code{best_only = TRUE}), only the match
with the highest relative score is returned for each sequence.
Should there be tied scores, the best match can be chosen as either the first,
last, most central, all tied matches, or choosing one at random (the default).
}
\examples{
# Load example PFMs
data("ex_pfm")
# Cluster using default settings
cl_ids <- clusterMotifs(ex_pfm)
ex_cl <- split(ex_pfm, cl_ids)
# Add optional names
names(ex_cl) <- vapply(ex_cl, \(x) paste(names(x), collapse = "/"), character(1))

# Load example sequences
data("ar_er_seq")
# Get all matches for each cluster
getClusterMatches(ex_cl, ar_er_seq)
# Or Just count them
countClusterMatches(ex_cl, ar_er_seq)
# Compare this to individual counts
countPwmMatches(ex_pfm, ar_er_seq)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{hg19_mask}
\alias{hg19_mask}
\title{Regions from hg19 with high N content}
\format{
An object of class \code{GRanges} of length 345.
}
\source{
The package AnnotationHub and
\url{https://hgdownload.cse.ucsc.edu/goldenpath/hg19/bigZips/hg19.fa.masked.gz}
}
\usage{
data("hg19_mask")
}
\description{
A GRanges object with regions annotated as telomeres or centromeres
}
\details{
The regions defined as centromeres or telomeres in hg19, taken from
AnnotationHub objects "AH107360" and "AH107361". These were combined with
regions containing Ns from the UCSC 2bit file, and regions with Ns in the
BSgenome.Hsapiens.UCSC.hg19 were retained.

Generation of these ranges is documented in
\code{system.file("scripts/hg19_mask.R", package = "motifTestR")}
}
\examples{
data("hg19_mask")
hg19_mask
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{ex_pfm}
\alias{ex_pfm}
\title{Example Position Frequency Matrices}
\format{
An object of class \code{list} of length 5.
}
\usage{
data("ex_pfm")
}
\description{
Example Position Frequency Matrices
}
\details{
This object contains 5 PFMs taken from HOCOMOCOv11-coreA for examples and testing

Generation of this motif list is documented in
\code{system.file("scripts/ex_pfm.R", package = "motifTestR")}
}
\examples{
data("ex_pfm")
ex_pfm$ESR1

}
\keyword{datasets}
